//'WWW''''''''''''''''''''''''''''''WW''''WW''''''''WW''''''''''WWW'''''''
//WW  W                             WW              WW           WW       
//WW     WWW WWWW W   WW  W WW  WWW WW    WW  WWWW WWWW     WWW  WW   WWW 
//WW    W WW WW WW W WWWW WW   W WW WW    WW WWW    WW     WW W  WW  W WW 
//WW  W W WW WW WW W WW   WW   W WW WW    WW   WWW  WW     WW    WW  W WW 
// WWW   WWW WW WW W  WWW WW    WWW WWWWW WW WWWW   WW      WWW WWWW  WWW 
//                                                                        
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#include "CameraList.h"
#include "debug.h"

// NOTE:
//
// the following tridiagonal matrix solver and cubic interpolation code are
// translated from a java program written by T. von Petersdorff
// 
//        http://www.wam.umd.edu/~petersd/interp.html

  
// solve linear system with tridiagonal n by n matrix a
// using Gaussian elimination *without* pivoting
// where   a(i,i-1) = sub[i]  for 2<=i<=n
//         a(i,i)   = diag[i] for 1<=i<=n
//         a(i,i+1) = sup[i]  for 1<=i<=n-1
// (the values sub[1], sup[n] are ignored)
//  right hand side vector b[1:n] is overwritten with solution 
//  NOTE: 1...n is used in all arrays, 0 is unused */
void CameraList::solveTridiag(float sub[], float diag[], float sup[],
			      float b[], int n){
    
  int i;
  // factorization and forward substitution
  for(i=2; i<=n; i++){
    sub[i] = sub[i]/diag[i-1];
    diag[i] = diag[i] - sub[i]*sup[i-1];
    b[i] = b[i] - sub[i]*b[i-1];
  }
  b[n] = b[n]/diag[n];
  for(i=n-1;i>=1;i--){
    b[i] = (b[i] - sup[i]*b[i+1])/diag[i];
  }
}       
  
// x[i] are assumed evenly spaced from 0..1
// y[i] are from 0..np-1
float CameraList::cubicInterp(float y[], int np, float t){
//  return 1.0;

  float yi = 0.0;
  
  if (np>1){  
    float a[np];
    float h[np];        // distance between points along x axis
    float t1;           // fraction of h[i] from first point
    float t2;           // fraction of h[i] from second point
    for (int i=1; i<=np-1; i++){
      h[i] = 1.0/(np-1); // x[i] - x[i-1] : evenly spaced, in our case.
    }
    if (np>2){
      float sub[np-1];
      float diag[np-1];
      float sup[np-1];

      // initialize tridiagonal matrix
      for (int i=1; i<=np-2; i++){
	diag[i] = (h[i] + h[i+1])/3.0;
	sup[i] = h[i+1]/6.0;
	sub[i] = h[i]/6.0;
	a[i] = (y[i+1]-y[i])/h[i+1]-(y[i]-y[i-1])/h[i];
      }
      a[0] = a[np-1] = 0.0;
	
      solveTridiag(sub,diag,sup,a,np-2);
    }
      
    int i;
    float j;
    i = (int)( (np-1)*t ) + 1;    // interval
    j = t*(np-1) - (i-1);         // percent along interval
      
    t1 = h[i]*j;
    t2 = h[i] - t1;
      
    yi = ( (-a[i-1]/6.0*(t2+h[i])*t1 + y[i-1]) * t2 +
	   (-a[i]/6.0*(t1+h[i])*t2 + y[i]) * t1 ) / h[i];

  }

  return(yi);
}


int CameraList::readList(char *camFile){
  struct stat file_stat;
  
  if (camFile == NULL)
    camFile = "camlist.iv";

  // test for presence of file
  if (stat(camFile, &file_stat) == 0){

    // open the file
    SoInput *camIn = new SoInput();
    if (camIn->openFile("camlist.iv")){

      // read contents of file
      SoSeparator *camScene = SoDB::readAll(camIn);
      camScene->ref();

      // cam scene is assumed to be only SoPerspectiveCameras
      int numCams = camScene->getNumChildren();
      cerr << "Adding " << numCams << " cameras to camera list.\n";

      // loop through cameras
      for (int i=0; i<numCams; i++){
	SoNode *node = camScene->getChild(i);

	// add cameras to camList
	if (node->isOfType(SoPerspectiveCamera::getClassTypeId()))
	  addCamera((SoPerspectiveCamera *)node);
      }

      // delete list
      camScene->unref();
    }

    // delete SoInput
    delete camIn;

    return(1);
  }
  else {
    cerr << "No camera list file camlist.iv found. Not using any camera paths.\n";
    return(0);
  }
}


void CameraList::positionCameraLinear(SoPerspectiveCamera *cam, float time){

  int numcams = camlist.size();
  if (numcams<1)
    return;
    
  int first, second;
  first = (int)( (numcams-1)*time );
  second = MIN2(first+1, numcams-1);

  // adjust time
  if (numcams==1){
    time = 0.0;
    second = first;
  }
  else
    time = (time-first*1.0/(numcams-1)) / (1.0/(numcams-1));

  SoPerspectiveCamera *cam1 = camlist[first];
  SoPerspectiveCamera *cam2 = camlist[second];

  // interpolate values

  // position
  SbVec3f pos1 = cam1->position.getValue();
  SbVec3f pos2 = cam2->position.getValue();
  cam->position.setValue( LERP(pos1, pos2, time) );

  // orientation
  // interpolate quaternions for smooth rotation
  const float *rot1 = cam1->orientation.getValue().getValue();
  const float *rot2 = cam2->orientation.getValue().getValue();

  cam->orientation.setValue(
    LERP(rot1[0], rot2[0], time),
    LERP(rot1[1], rot2[1], time),
    LERP(rot1[2], rot2[2], time),
    LERP(rot1[3], rot2[3], time)
    );

  // focalDistance
  float fd1 = cam1->focalDistance.getValue();
  float fd2 = cam2->focalDistance.getValue();
  cam->focalDistance.setValue( LERP(fd1, fd2, time) );

  // heightAngle
  float ha1 = cam1->heightAngle.getValue();
  float ha2 = cam2->heightAngle.getValue();
  cam->heightAngle.setValue( LERP(ha1, ha2, time) );

    
}

void CameraList::positionCameraCubic(SoPerspectiveCamera *cam, float time){

  int numcams = camlist.size();
  if (numcams<1)
    return;

  // do linear interpolation if only two cams
  if (numcams<=2){
    positionCameraLinear(cam, time);
    return;
  }
    
  // interpolate values
  SbVec3f pos, newPos = SbVec3f(0.0, 400.0, -800.0);
  SoPerspectiveCamera *cami;
  float x[numcams];
  float interp;
    
  // position
  for (int c=0; c<3; c++){
      
    // populate x[]
    for (int i=0; i<numcams; i++){
      cami = camlist[i];
      pos = cami->position.getValue();
      x[i] = pos[c];
    }

    // get interpolated value
    interp = cubicInterp(x, numcams, time);
    newPos[c] = interp;
  }
  cam->position.setValue( newPos );

  /*
  // debugging
  static int pnts = 0;
  if (pnts++<200)
    showPoint(newPos, 5.0);
  */

  // orientation
  // interpolate quaternions for smooth rotation
  float newRot[4];
  for (int c=0; c<4; c++){
      
    // populate x[]
    for (int i=0; i<numcams; i++){
      cami = camlist[i];
      x[i] = cami->orientation.getValue().getValue()[c];
    }

    // get interpolated value
    interp = cubicInterp(x, numcams, time);
    newRot[c] = interp;
  }
  cam->orientation.setValue( newRot );

  // focalDistance
  // populate x[]
  for (int i=0; i<numcams; i++){
    cami = camlist[i];
    x[i] = cami->focalDistance.getValue();
  }
    
  // get interpolated value
  interp = cubicInterp(x, numcams, time);
  cam->focalDistance.setValue( interp );

  // heightAngle
  // populate x[]
  for (int i=0; i<numcams; i++){
    cami = camlist[i];
    x[i] = cami->heightAngle.getValue();
  }
    
  // get interpolated value
  interp = cubicInterp(x, numcams, time);
  cam->heightAngle.setValue( interp );
    
}
