////////////////////////////////////////////////////////////////////////////
//
//WWWW''''''''''''''''''WW'WW''''''''''''''''''''''WWW''''''''''''WW'''''''
// WW                   WW                        WW  W           WW       
// WW  W WW W WW  WWW  WWW WW  WWW WWW   WWW  WW  WW     WWW  WWW WWW   WW 
// WW  WW   WW   W WW W WW WW W WW WW W WW W WWWW WW    W WW WW W WW W WWWW
// WW  WW   WW   W WW W WW WW W WW WW W WW   WW   WW  W W WW WW   WW W WW  
//WWWW WW   WW    WWW  WWW WW  WWW WW W  WWW  WWW  WWW   WWW  WWW WW W  WWW
//                                                                         
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// File: IrradianceCache.h
// Description: IrradianceCache class header
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Irradiance Cache for the Pane ray tracer
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// This class is an implementation of an Irradiance Cache,
// which is described in the following paper:
//
// A Ray Tracing Solution for Diffuse Interreflection
// by Greg Ward.
//
// Briefly, irradiance samples are stored along
// with thier position in an octtree in such a way as to quickly
// find nearby stored samples for a given point in space.
//
// On querying, nearby samples are found (if they exist) and their
// values are interpolated to produce values between samples.
// 
// $Id: IrradianceCache.h,v 1.14 2004/10/24 22:21:46 beason Exp $
//
////////////////////////////////////////////////////////////////////////////

#ifndef IRRADIANCECACHECLASS
#define IRRADIANCECACHECLASS

#include "Main.h"
#include "GlobalVars.h"

class Ray; // forward declaration

// irradiance point sample
typedef struct pnt_st {
  float ri;
  SbVec3f irrad;
  SbVec3f p, n, ng;
  SbVec3f Er, Et;
} irpoint;

// octtree cell
typedef struct ic_node_st {
  struct ic_node_st *child[8];
  SbVec3f mid;
  float half_size;
  short terminal;
  list<irpoint *> pntlist;
} ic_node;

// used for passing several values when calling findPoints()
typedef struct cacheParam {
  SbVec3f *e;
  SbVec3f *n;
  SbVec3f *ng;
  SbVec3f *p;
  thread_context *ctx;
  float denom;
  short num_vals;
  short exactFound;
  int prepass;
} cacheParam;

// Irradiance cache class
class IrradianceCache {
  
  private:
  
  float a;                              // accuracy control
  const static int delta[8][3];         // cube vertex indices
  float first_cell_size;                // spatial size of cache
  float resolution;                     // resolution (# in one dim) of samples
  float minRi;                          // minimum spacing of samples
  int num_pnts;                         // number of samples
  int treeDebug;                        // bool for debugging
  ic_node *root;                        // root node of octtree
  GlobalVars *gv;                       // pointer to global variables

  public:
  
  SbVec3f min;                          // minimum spatial corner of cache

  IrradianceCache();                    // constructor
  ~IrradianceCache();                   // deconstructor

  float size(){ return(first_cell_size); };  // returns spatial size of cache
  void setA(float newa){ a = newa; }         // returns accuracy
  void setResolution(float newr){            // sets resolution of sample pts
    resolution = newr;
    minRi = first_cell_size/resolution;
  }
  void setGV(GlobalVars *gv_ptr){            // sets globa variables pointer
    gv = gv_ptr;
  }
  void setDims(SbVec3f min, float size, int debug);   // sets size of cache

  // returns interpolated irradiance value, or 
  int getIrradiance(SbVec3f *e, Ray &r, SbVec3f *n, SbVec3f *ng,
		    int prepass=0);
  void storeIrradiance(SbVec3f &e, SbVec3f &p, SbVec3f &n, SbVec3f &ng,
		       float r, SbVec3f &Er, SbVec3f &Et);
  void findPoints(cacheParam &icp, ic_node *currnode);
  int validPoint(SbVec3f &n, SbVec3f &x, float r, SbColor &e);
  void closestPoint(ic_node *currnode, SbVec3f &n, SbVec3f &x, float &r,
		    irpoint *&irr);
  void deleteNode(ic_node *n);
  void resetCache();
  int numPoints(){ return(num_pnts); }
};

#endif /* IRRADIANCECACHECLASS */
