/////////////////////////////////////////////////////////////////////////////
//                                                                           
//  WWWWW                                                                     
//  WW  WW                                                                    
//  WW  WW                                                                    
//  WWWWW    WWWW   WWWW    WWW                                               
//  WW          WW  WW WW  WW WW                                              
//  WW       WWWWW  WW WW  WWWWW                                              
//  WW      WW  WW  WW WW  WW                                                 
//  WW       WWWWW  WW WW   WWWW                                              
//                                                                           
//                                                                           
//
// File: Main.cxx
// Description: Main ray tracer routine (start up)
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Main routine for the Pane ray tracer.
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// Usage: pane [OPTION]... [IV FILE]
//
// Synopsis: pane is a ray tracer.
//           Input is a scene in the Open Inventor format (scene.iv)
//           Output is an image (scene.png).
//
// $Id: Main.cxx,v 1.39 2005/09/20 22:21:37 beason Exp beason $
//
////////////////////////////////////////////////////////////////////////////

#include "Main.h"
#include "debug.h"
#include "Ray.h"
#include "Texture.h"
#include "Triangle.h"
#include "TriangleTree.h"
#include "GlobalVars.h"
#include "Pane.h"
#include "sobol.h"
#include "CameraList.h"

int DEBUG_STATE=0;                         // debugging mode if true
int STOP_RENDERING=0;                      // stop rendering if true
const int mutexNum = 16;                   // max number of threads
pthread_mutex_t mutexes[mutexNum];         // mutual exclusion locks for threads
SoSeparator *scnRoot;                      // overall scene graph root

// vector finite
int finite(SbVec3f &c){
  return( finite(c[X]) && finite(c[Y]) && finite(c[Z]) );
}

// vector mult
SbVec3f mul(SbVec3f &x, SbVec3f &y, float c){
  return(SbVec3f(x[X]*y[X]*c,
		 x[Y]*y[Y]*c,
		 x[Z]*y[Z]*c));
}

//''WW'WW''''''''WW''WWW'''''''''''''''''''''''WWW''''''''''''''
// WW            WW WW  W                     WW  W             
//WWWW WW WWW   WWW  WW    WWW  WW  WWW   WW  WW     WWW WWWW W 
// WW  WW WW W W WW   WW  WW W WWWW WW W WWWW WW    W WW WW WW W
// WW  WW WW W W WW W  WW WW   WW   WW W WW   WW  W W WW WW WW W
// WW  WW WW W  WWW  WWW   WWW  WWW WW W  WWW  WWW   WWW WW WW W
//                                                              
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Finds the first SoCamera in a scene
//
void findSceneCam(GlobalVars *gv){
  // Search for camera in scene graph
  SoSearchAction sa;
  sa.setType(SoCamera::getClassTypeId());
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.apply(gv->root);
  if (sa.getPath())
    gv->cam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();
}

// forward declaration
void animateCamera(void *data, SoSensor *);

//WW''''''''''''''WWWWW'''''''''''''''''WW''WW''W'''''''''''''WW'WWW'''''''
//WW              WW                    WW  WW  W             WW  WW       
//WW W   WW  W WW WWWW  WW W  WW  WWW  WWWW WW  W  WWW WWW   WWW  WW   WW  
//WWW   WWWW W WW WW    WW W WWWW WW W  WW  WWWWW W WW WW W W WW  WW  WWWW 
//WW W  WW   W WW WW    WW W WW   WW W  WW  WW  W W WW WW W W WW  WW  WW   
//WW  W  WWW  WWW WWWWW  WW   WWW WW W  WW  WW  W  WWW WW W  WWW WWWW  WWW 
//             WW                                                          
//'''''''''''WWW'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Processes viewer keypresses
//
void keyEventHandler(void *gvp, SoEventCallback *callBack){
  const SoEvent *event = callBack->getEvent();
  GlobalVars *gv = (GlobalVars *)gvp;

  // reorder nodes
  if (SO_KEY_PRESS_EVENT(event, E)){
    gv->pane->tritree.reorderNodes();
  }
  
  // Render image when 'R' is pressed
  if (SO_KEY_PRESS_EVENT(event, R)){
    gv->pane->render();
  }

  // Write rendered image to file when 'W' is pressed
  if (SO_KEY_PRESS_EVENT(event, W)){
    gv->pane->writeImg();
  }

  // Toggle headlight when 'H' is pressed
  if (SO_KEY_PRESS_EVENT(event, H)){
    int i = gv->headLightPos;
    i = (i+1)%3;
    gv->headLightPos = i;
    cerr << "Headlight ";
    switch(i){
    case HL_OFF:
      cerr << "OFF";
      gv->sceneHasLights--;
      break;
    case HL_CENTER:
      cerr << "CENTER";
      gv->sceneHasLights++;
      break;
    case HL_OFFCENTER:
      cerr << "OFFCENTER";
    }
    cerr << endl;
  }

  // turn up headlight
  if (SO_KEY_PRESS_EVENT(event, EQUAL)){ // up
    double i = gv->pane->headLight->intensity.getValue();
    cerr << "Increasing headLight intensity from "
	 << setiosflags(ios::showpoint | ios::fixed) << setprecision(8)
	 << i;
    i = pow(10.0, log10(i)+0.25);
    cerr << " to " << i << endl;
    gv->pane->headLight->intensity.setValue(i);
  }

  // turn down headlight
  if (SO_KEY_PRESS_EVENT(event, MINUS)){ // down
    double i = gv->pane->headLight->intensity.getValue();
    cerr << "Decreasing headLight intensity from "
	 << setiosflags(ios::showpoint | ios::fixed) << setprecision(8)
	 << i;
    i = pow(10.0, log10(i)-0.25);
    cerr << " to " << i << endl;
    gv->pane->headLight->intensity.setValue(i);
  }

  // Print camera info
  if (SO_KEY_PRESS_EVENT(event, C)){
    cerr << "Camera info:" << endl;
    SoOutput *out = new SoOutput();
    out->setFilePointer(stderr);
    SoWriteAction *wa = new SoWriteAction(out);

    // Search for camera in scene graph
    SoSearchAction sa;
    sa.setType(SoCamera::getClassTypeId());
    sa.setSearchingAll(FALSE); // don't look under off switches
    sa.apply(gv->root);
    if (sa.getPath())
      gv->cam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();
    
    wa->apply(gv->cam); //gv->cam);
  }

  // Save scene to file
  if (SO_KEY_PRESS_EVENT(event, S)){
    //cerr << "Camera info:" << endl;
    SoOutput *out = new SoOutput();
    //out->setFilePointer(stderr);
    char *fn = "pane_dump.iv";
    out->openFile(fn);
    SoWriteAction *wa = new SoWriteAction(out);
    wa->apply(gv->root); //gv->cam);
    out->closeFile();
    cerr << "Dumped scene graph to " << fn << ".\n";
    delete out;
    delete wa;
  }
  
  // Align Camera to axes
  if (SO_KEY_PRESS_EVENT(event, Z)){
    static int rev=1;
    rev = !rev;

    findSceneCam(gv);
    gv->cam->orientation.setValue(SbVec3f(0, 1, 0), M_PI*rev);
  }
  if (SO_KEY_PRESS_EVENT(event, X)){
    static int orient = 1;
    orient *= -1;

    findSceneCam(gv);
    gv->cam->orientation.setValue(SbVec3f(0, 1, 0),
				  orient*1.57079632679489661922);
  }
  if (SO_KEY_PRESS_EVENT(event, Y)){
    static int orient = 1;
    orient *= -1;

    findSceneCam(gv);
    gv->cam->orientation.setValue(SbVec3f(1, 0, 0),
				  orient*1.57079632679489661922);
  }

  // position camera
  if (SO_KEY_PRESS_EVENT(event, V)){
    cerr << "Positioned camera" << endl;
    gv->cam->orientation.setValue(
      SbVec3f(-0.0014208, 0.999947, 0.0102222),  5.10545);
    gv->cam->position.setValue(-2298.35, 602.14, -1020.66);
  }

  // animate camera
  if (SO_KEY_PRESS_EVENT(event, A)){
    static SoTimerSensor *sensor;
    static int animating = 0;

    if (sensor == NULL)
      sensor = new SoTimerSensor(animateCamera, gv);
    
    // fixes default value (set to signify not to position camera)
    if (gv->timeVal<0.0)
      gv->timeVal = 0.0;

    if (animating){
      cerr << "Stopping animation.\n";
      sensor->unschedule();
      animating = 0;
    }
    else {
      cerr << "Animating.\n";
      sensor->setInterval(SbTime(0.015)); // 70 frames per second
      sensor->schedule();
      animating = 1;
    }
  }

  // rotate scene
  if (SO_KEY_PRESS_EVENT(event, O)){

    SbVec3f axis;
    float angle;
    gv->trans->rotation.getValue(axis, angle);
    gv->trans->rotation.setValue(axis, angle+M_PI/4.0);

    // recompute octree
    gv->pane->rescanScene();
  }

  // trace pixel
  if (SO_KEY_PRESS_EVENT(event, T)){
    gv->pane->render(1);
  }

  // show normals
  if (SO_KEY_PRESS_EVENT(event, N)){
    gv->showNormals = !gv->showNormals;
    if (gv->showNormals)
      cerr << "Showing normals.\n";
    else
      cerr << "Disabled showing normals.\n";
  }
  
  // shoot photons
  if (SO_KEY_PRESS_EVENT(event, P)){
    TIME(gv->pane->shootPhotons());
  }

  // debugging
  // irradiance estimate
  if (SO_KEY_PRESS_EVENT(event, D)){
    SbVec3f hit_x = SbVec3f(371.333801, 144.431351, 331.016296);
    SbVec3f hit_n = SbVec3f(-0.0148499319, -0.999874473, -0.00552253844);
    float irrad[3];

    int temp = DEBUG_STATE;
    DEBUG_STATE = 1;
    gv->pane->gPhotonMap->irradiance_estimate(
      irrad, hit_x.getValue(), (-hit_n).getValue(),
      gv->gPhotonEstimate);
    DEBUG_STATE = temp;    
    cerr << "irrad = "; printVec(SbVec3f(irrad)); cerr << endl;
  }

  // toggle geometry
  if (SO_KEY_PRESS_EVENT(event, G)){

    int val = gv->sceneSwitch->whichChild.getValue();
    int newVal = (val==0)?-1:0;
    gv->sceneSwitch->whichChild.setValue(newVal);
  }

  // toggle crosshair
  if (SO_KEY_PRESS_EVENT(event, PERIOD)){

    int val = gv->crossSwitch->whichChild.getValue();
    int newVal = (val==0)?-1:0;
    gv->crossSwitch->whichChild.setValue(newVal);
  }
  
  // turns surfaces black, for photon viewing
  if (SO_KEY_PRESS_EVENT(event, B)){

    SoCallbackAction cb;
    cb.addPreCallback(SoMaterial::getClassTypeId(),
		      Pane::matBlackCBHandler, (void *)NULL);
    cb.apply(gv->scene);
  }

  // cycle window background color
  if (SO_KEY_PRESS_EVENT(event, K)){
    static int col = 0;
    // 0 = windowBG, 1 = black, 2 = gray, 3 = white, 4 = purple
    col = (col+1)%5;
    switch(col){
    case 0:
      gv->xtViewer->setBackgroundColor(gv->windowBG);
      break;
    case 1:
      gv->xtViewer->setBackgroundColor(SbColor(0.0, 0.0, 0.0));
      break;
    case 2:
      gv->xtViewer->setBackgroundColor(SbColor(0.25, 0.25, 0.25));
      break;
    case 3:
      gv->xtViewer->setBackgroundColor(SbColor(1.0, 1.0, 1.0));
      break;
    case 4:
      gv->xtViewer->setBackgroundColor(SbColor(0.5, 0.0, 0.5));
      break;
      
    default:
      break;
    }
  }
  
}

// Animate scene
void rotateScene(void *data, SoSensor *){
  GlobalVars *gv = (GlobalVars *)data;

  // rotate scene
  SbVec3f axis;
  float angle;
  gv->trans->rotation.getValue(axis, angle);
  angle += 0.02;
  if (angle>M_PI*2.0)
    angle = 0.0;
  gv->trans->rotation.setValue(axis, angle);

  // update octree if not set to do it automatically
  if (!gv->automaticUpdate)
    gv->pane->rescanScene();
}


//''''''''''WW'''''''''''''''WW''''''''WWW'''''''''''''''''''''''''''''
//                           WW       WW  W                            
// WWW WWW  WW WWWW W   WWW WWWW  WW  WW     WWW WWWW W   WW  W WW  WWW
//W WW WW W WW WW WW W W WW  WW  WWWW WW    W WW WW WW W WWWW WW   W WW
//W WW WW W WW WW WW W W WW  WW  WW   WW  W W WW WW WW W WW   WW   W WW
// WWW WW W WW WW WW W  WWW  WW   WWW  WWW   WWW WW WW W  WWW WW    WWW
//                                                                     
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Reposition the camera
//
void animateCamera(void *data, SoSensor *){
  GlobalVars *gv = (GlobalVars *)data;

  SoPerspectiveCamera *cam = (SoPerspectiveCamera *)gv->xtViewer->getCamera();

  if (gv->camList)
    gv->camList->positionCameraCubic(cam, (1.0-cos(gv->timeVal*M_PI))/2.0);
  
  // increment time

  //float secs = gv->frames * 0.15;
  gv->timeVal += 1.0/(gv->frames-1);
  if (gv->timeVal>1.0)
    gv->timeVal = 0.0;
}



//'''''''''''''''''''''''''''WWW''WW''''''''''''''''''''''''WWW''WWWW'
//                          WW  W WW                       WW  W WW  W
// WWWW  WWW  WW  WWW   WW  WW    WWW   WWW WWW   WWW  WW  WW    WWWW 
//WWW   WW W WWWW WW W WWWW WW    WW W W WW WW W W WW WWWW WW    WW  W
//  WWW WW   WW   WW W WW   WW  W WW W W WW WW W W WW WW   WW  W WW  W
//WWWW   WWW  WWW WW W  WWW  WWW  WW W  WWW WW W  WWW  WWW  WWW  WWWW 
//                                                 WW                 
//'''''''''''''''''''''''''''''''''''''''''''''''WWW''''''''''''''''''
//
// Callback to update octree if scene changes
//
void sceneChangeCB(void *data, SoSensor *sensor){
  ((GlobalVars *)data)->pane->rescanScene();
}


//'''''''''''''''''''''WW''''WW''''''WW'''''
//                     WWW  WWW             
//WWW  W WW  WW    WWW W WWW WW  WWW WW WWW 
//WW W WW   WW W  W WW W  W  WW W WW WW WW W
//WW W WW   WW W  W WW W     WW W WW WW WW W
//WWW  WW    WW    WWW W     WW  WWW WW WW W
//WW                WW                      
//WW''''''''''''''WWW'''''''''''''''''''''''
//
// Main subroutine for the ray tracer
//
int progMain(int argc, char* argv[]){

  struct timeval totalStart, totalEnd;
  gettimeofday(&totalStart, NULL); // start delay measure
  
  // Global variables object
  GlobalVars gv;
  Pane pane(gv);
  gv.pane = &pane;
  gv.readOpts(argc, argv);

  // seed random number generator
  //srand48(time(NULL));             // random
  srand48(gv.srandSeed);            // deterministic (default 0)
  sobinit(20);                       // initialize sobol point generator

  // Initialize Inventor
  if (gv.renderQuit){
    SoDB::init();
  }
  else {
    SoXt::init(argv[0]);     // works better than SoDB::init(), thanks Chuck.
  }
  
  cerr << "Reading scene file...";
  struct timeval start, end;
  gettimeofday(&start, NULL); // start delay measure

  // Read scene
  if (gv.inImgFile){
    cerr << "(null scene)";
    gv.scene = new SoSeparator();
  }
  else {
    
    // read scene data
    SoInput *sceneIn = new SoInput();
    if (!sceneIn->openFile(gv.sceneFile)){
      cerr << "Unable to open scene file: " << gv.sceneFile << endl;
      exit(1);
    }
    
    gv.scene = SoDB::readAll(sceneIn);
    delete sceneIn;
  }
  gv.scene->ref();
  if (!gv.scene){
    cerr << "Error: scene separator is null!" << endl;
    exit(1);
  }

  // Does scene have a camera?
  SoSearchAction sa;
  sa.setType(SoCamera::getClassTypeId());
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.apply(gv.scene);
  SoCamera *sceneCam = NULL;
  SoFullPath *fullPath = NULL;  
  if (sa.getPath()){
    sceneCam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();
    fullPath = (SoFullPath *) sa.getPath();
  }
  
  // If camera exists, and is of a known type, move it to top of
  // scene root. If it's of unknown type, remove it and create
  // a perspective camera at scene root.
  //
  // We do this because the ExaminerViewer attaches to and modifies
  // the first camera (if it exists), modifying the scenegraph.
  //
  // Since we want to detect changes to the scenegraph, not including
  // view/camera changes, this is bad. Therefore we move the camera outside
  // of the scenegraph.
  if (sceneCam){

    // find parent node of camera
    SoGroup *parent = (SoGroup *)fullPath->getNodeFromTail(1);

    // identify camera type and if it's unknown,
    // create a new camara.
    if (sceneCam->isOfType(SoOrthographicCamera::getClassTypeId())){
      cerr << "Orthographic camera found...";
      gv.cam = sceneCam;
    }
    else if (sceneCam->isOfType(SoPerspectiveCamera::getClassTypeId())){
      cerr << "Perspective camera found...";
      gv.cam = sceneCam;
    }
    else {
      cerr << "Unknown camera type found, creating perspective camera...";

      // copy common stuff
      gv.cam = new SoPerspectiveCamera();
      gv.cam->viewportMapping = sceneCam->viewportMapping.getValue();
      gv.cam->position = sceneCam->position.getValue();
      gv.cam->orientation = sceneCam->orientation.getValue();
      gv.cam->aspectRatio = sceneCam->aspectRatio.getValue();
      gv.cam->focalDistance = sceneCam->focalDistance.getValue();
    }

    // remove the original camera
    // gv.cam->ref();
    // parent->removeChild(sceneCam);
    // gv.cam->unrefNoDelete();
  }

  // otherwise create a camera
  else { 
    gv.cam = new SoPerspectiveCamera();
    gv.cam->viewAll(gv.scene, SbViewportRegion());
  }


  //                   scnRoot
  //                (aka gv->root)
  //                       |
  //            -----------------------------------...
  //            |          |         |
  //      crosshairSep  camera   sceneSwitch
  //            |                    |
  //           ...                sceneSep
  //                           (aka gv->scene)
  //                                 |
  //                         --------------... (rest of the scene)
  //                         |         
  //                     transform    
  
  // create scene root
  gv.root = new SoSeparator();
  gv.root->ref();
  scnRoot = gv.root;
  
  // add crosshair
  SoSeparator *chSep = new SoSeparator();
  SoMaterial *chMat = new SoMaterial();
//  chMat->emissiveColor.setValue(0.9, 0.6, 0.9);
  chMat->emissiveColor.setValue(0.2, 0.8, 0.2);
  chSep->addChild(chMat);
  SoTransform *chXform = new SoTransform();
  chXform->scaleFactor.setValue(0.003, 0.003, 0.003);
  chXform->translation.setValue(0.0, 0.0, -0.01);
  chSep->addChild(new SoResetTransform());
  chSep->addChild(chXform);
  chSep->addChild(new SoCube());

  // put crosshair in switch so it can be toggled
  gv.crossSwitch = new SoSwitch();
  gv.crossSwitch->addChild(chSep);
  if (!gv.figureMode)
    gv.crossSwitch->whichChild.setValue(0);
  gv.root->addChild(gv.crossSwitch);

  // add camera
  gv.root->addChild(gv.cam);

  // add scene after adding transform (for rotation) to it
  gv.trans = new SoTransform();
  gv.scene->insertChild(gv.trans, 0);

  // put scene in switch so it can be toggled
  gv.sceneSwitch = new SoSwitch();
  gv.sceneSwitch->addChild(gv.scene);
  gv.sceneSwitch->whichChild.setValue(0);
  
  gv.root->addChild(gv.sceneSwitch);
  gv.scene->unref(); // now that addChild ref'd, can unref earlier ref()
  gv.root->addChild(new SoResetTransform());

  gettimeofday(&end, NULL);   // end delay measure
  cerr << "done. (" << delay(start, end) << "ms)" << endl;

  // compute normals
  if (gv.computeNormals){
    cerr << "\nComputing normals...\n";
    pane.computeNormals();
  }
  
  cerr << "Initializing ray tracing engine...";
  gettimeofday(&start, NULL); // start delay measure
  int numTris = pane.init();  // initialize pane
  gettimeofday(&end, NULL);   // end delay measure
  cerr << numTris << " triangles read ("
       << delay(start, end) << "ms)" << endl;
  gv.times.octreeCreation = delay(start, end);

  // make camera list from file
  gv.camList = new CameraList();
  gv.camList->readList();

  // timeVal of <0 (the default) signifies not to set the camera position
  if (gv.cam->isOfType(SoPerspectiveCamera::getClassTypeId()) &&
      gv.timeVal>=0.0){
    gv.camList->positionCameraCubic(
      (SoPerspectiveCamera *)gv.cam,
      gv.timeVal);
  }

  // fixes default value (set to signify not to position camera)
  if (gv.timeVal<0.0)
    gv.timeVal = 0.0;
  
  // Create Examiner Viewer
  if (! gv.renderQuit){

    // This returns a main window to use.
    // If unsuccessful, exit.
    Widget myWindow = SoXt::init(argv[0]); // pass the app name
    if (myWindow == NULL) exit(1);

    // animate scene
    if (gv.animateScene){      
      SoTimerSensor *sensor = new SoTimerSensor(rotateScene, &gv);
//      SoTimerSensor *sensor = new SoTimerSensor(animateCamera, &gv);
      sensor->setInterval(SbTime(0.015)); // 70 frames per second
      sensor->schedule();
    }
    
    // Set up keyboard
    SoEventCallback *callBack = new SoEventCallback();
    callBack->addEventCallback(SoKeyboardEvent::getClassTypeId(),
			       keyEventHandler, &gv);
    gv.root->addChild(callBack);

    cerr <<
      "\n"
      "Press ESC, then R to render.\n"
      "Then press W to write image.\n"
      "\n"
      "Keys:\n"
  "  P  shoot photons                R  render scene\n"      
  "  H  toggle headlight             W  write image to disk\n"
  "\n"
      "More Keys:\n"
  "  +  increase headlight power     X  point camera along negative X axis\n"
  "  -  decrease headlight power     Y  point camera along negative Y axis\n"
  "  C  print camera node            Z  point camera along negative Z axis\n"
  "  S  save scene graph to file     O  rotate scene and recalculate octree\n"
  "  T  trace ray and show           G  toggle geometry\n"
  "  N  show normals for traced ray  A  animate camera using camlist.iv\n"
  "  .  Turn off crosshair           B  turn surfaces black\n"
  "                                  K  cycle window background color\n"
	 << endl;
    
    // Set up viewer
    SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
    gv.xtViewer = myViewer;  // store viewer for use by animateCamera
   
    myViewer->setSceneGraph(gv.root);
    myViewer->setTitle("Pane Scene Viewer");
    if (!gv.fastTransparency){
      myViewer->setTransparencyType(SoGLRenderAction::DELAYED_BLEND);
      //myViewer->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_BLEND);
    }
    //myViewer->setTransparencyType(SoGLRenderAction::SCREEN_DOOR);
    myViewer->setBackgroundColor(gv.windowBG);

    // FIGURE: remove me
    if (gv.figureMode)
      pane.shootPhotons();
    
    // Show scene
    myViewer->show();

    // automatically update octree on scene change
    if (gv.automaticUpdate){
      SoNodeSensor *nodeSensor = new SoNodeSensor(sceneChangeCB, &gv);
      nodeSensor->attach(gv.scene);
    }
    
    SoXt::show(myWindow); // Display main window
    SoXt::mainLoop();	// Main Inventor event loop
  }

  // Or Render and Quit (no display)
  else {

    if (
      gv.algorithm != 2 && gv.algorithm !=3 && gv.inImgFile==NULL && 
      ( (gv.renderFlags == 0 || gv.renderFlags & Pane::C_CAUSTIC) &&
	gv.numCPhotons>0) ||
      ( (gv.algorithm==1 || gv.renderFlags & Pane::C_MDIFFUSE) &&
	gv.numPhotons>0)
	)
      pane.shootPhotons();

    // render a series of frames
    if (gv.animateScene){

      float time;
      cerr << "Rendering " << gv.frames << " frames.\n";
//      for (int i=0; i<gv.frames; i++){
      // start at frame determined by time
      for (int i=int(gv.timeVal*(gv.frames-1)); i<gv.frames; i++){
	time = (float)i/(gv.frames-1);
	cerr << "Rendering frame " << i+1 << " of " << gv.frames << "\n";
	cerr << "Time = " << time << endl;

	// position camera
	gv.camList->positionCameraCubic(
	  (SoPerspectiveCamera *) gv.cam,
	  (1.0-cos(time*M_PI))/2.0);
	
	// render
	pane.render();
	pane.writeImg();
      }

      // write last frame twice to make a nice series of file names
      pane.writeImg();

    }
    
    // render IV file
    else if (gv.renderIv){
      pane.render();          // calls renderIv() after initial setup
      pane.writeIv();
    }

    // do image processing only
    else if (gv.inImgFile){
      pane.writeImg();
    }
    
    // render one frame
    else {
      pane.render();
      pane.writeImg();
    }

    // erase scene graph (for memory debugging)
    gv.root->unref();

    // print out total rendering time

    // user time
    struct tms processTime;
    times(&processTime);

//   cerr << "utime = " << processTime.tms_utime << "\n";
//   cerr << "stime = " << processTime.tms_stime << "\n";
//   cerr << "cutime = " << processTime.tms_cutime << "\n";
//   cerr << "cstime = " << processTime.tms_cstime << "\n";
//   cerr << "totalTicks = " << gv.times.totalTicks << "\n";
//   cerr << "renderTicks = " << gv.times.renderTicks << "\n";
  
    gv.times.totalTicks += processTime.tms_utime + processTime.tms_cutime;
    clock_t totalTicks = gv.times.totalTicks;
    long ticksPerSec = sysconf(_SC_CLK_TCK);
    float ticksMs =  totalTicks/(double)ticksPerSec*1000.0;
    print ("\nTotal cpu ticks: %d, Ticks/s: %d\n", totalTicks, ticksPerSec);
    print("Total cpu time   : (%f ms, %f s, %f m, %f h)\n",
	  ticksMs, ticksMs/1000.0, ticksMs/1000.0/60.0,
	  ticksMs/1000.0/60.0/60.0);

    // actual time
    gettimeofday(&totalEnd, NULL);   // end delay measure
    int ms = delay(totalStart, totalEnd);
    print("Total actual time: (%d ms, %f s, %f m, %f h)\n",
	  ms, ms/1000.0, ms/1000.0/60.0,
	  ms/1000.0/60.0/60.0);
  }
    
  return(0);
}


//'''''''''''''WW'''''
//                    
//WWWW W   WWW WW WWW 
//WW WW W W WW WW WW W
//WW WW W W WW WW WW W
//WW WW W  WWW WW WW W
//                    
//''''''''''''''''''''
//
// Entry point for pane executable. Calls progMain()
//
int main(int argc, char *argv[]){
  int retVal = progMain(argc, argv);
#ifdef DEBUG_MEM  
  DumpUnfreed();
#endif // DEBUG_MEM
  return(retVal);
}
