////////////////////////////////////////////////////////////////////////////
//
//WWWW'''''''''''''''''''''''''WWW''''''''''''''''''
//WW  W                         WW                  
//WW  W  WWW WWW   WW      WWW  WW   WWW  WWWW  WWWW
//WWWW  W WW WW W WWWW    WW W  WW  W WW WWW   WWW  
//WW    W WW WW W WW      WW    WW  W WW   WWW   WWW
//WW     WWW WW W  WWW     WWW WWWW  WWW WWWW  WWWW 
//                                                  
//''''''''''''''''''''''''''''''''''''''''''''''''''
//
// File: Pane.h
// Description: Pane class (Core Ray Tracing Engine) header
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Rendering engine for the Pane ray tracer
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// class Pane implements core ray tracing routines
// such as computing pixel colors and writing
// images to files. These routines are called
// from progMain() in Main.cxx.
// 
// $Id: Pane.h,v 1.50 2005/07/02 12:30:46 beason Exp beason $
//
////////////////////////////////////////////////////////////////////////////

#ifndef PANECLASS
#define PANECLASS

#include "Main.h"
#include "rgbe.h"
#include "GlobalVars.h"
#include "Ray.h"
#include "TriangleList.h"
#include "TriangleTree.h"
#include "Photon_map.h"
#include "IrradianceCache.h"
#include "EnvironMap.h"
#include "Material.h"

class Triangle; // forward declaration of Triangle class
class EnvironMap; // forward declaration of EnvironMap class
enum {FRONTSIDE, BACKSIDE};

// for containing image plane information
typedef struct CanvasParameters {
  double s, height, width;
  SbRotation sbr;
  SbVec3f e, g, vup, w, u, v, x;
  int prepass;
  int sqrt_samples;
  int trace;
  int flags;
  int orthographic;
} CanvasParameters;

typedef enum {
  GLOBAL, CAUSTIC, IMPORTANCE
} photon_t;

// for passing object functions to threads
typedef struct MemberFunction {
  Pane *object;
  void (Pane::*pmf)(thread_context*);
  void (Pane::*pmf2)(thread_context*, float, photon_t);
  float numToFire;
  thread_context *ctx;
  photon_t phot_type;
} MemberFunction;


// Core ray tracing engine class
class Pane {
  
  private:
  
  TriangleList litTris;                 // list of polygonal luminaires
  vector<SoPointLight *> litPnts;       // list of point luminaires
  int numLights;                        // number of lights
  float maxEmisAvg;                     // maximum emissive average
  float totalEmisAvg;                   // total emissive average for all lights (hW)
  SbColor *pix;                         // image ptr
                // (0,0) = (X,Y) is bottom left of pixmap, X is horiz
  float *pixDepth;                      // image depth map
  SbColor *filtpix;                     // filter pixels for rendering glare
  double totalPower;                    // power emitted from lights in W
  int firedPhotons;                     // number of emitted photons
  int max_photons;                      // maximum number of photons in each map
  IrradianceCache irradCache;           // irradiance cache
  CanvasParameters canvas;              // image plane information
  float completion[16];                 // thread rendering completion meters
  char completion_stage[16];            // thread rendering stage indicators
  float minApproxDistance;              // minimum length of ray to use approx. sol.
  int rejectedICPoints;                 // number of points rejected for cache
  int toneMapped;                       // true if img tone mapped, false otherwise
  EnvironMap *envMap;                    // environment map for image based lighting
  int sampMapHeight;                    // sample map height
  int sampMapWidth;                     // sample map width
  float *sampMap;                       // sample map pixel values
  static const int Mmax=100, Nmax=300;  // maximum (phi,theta) indirect samples
  int Mmdc, Nmdc;                       // (phi, theta) indirect samples to use
  float sinTxcos2T[Mmax], sinTmsinT[Mmax]; // holds precompute values for mdc
  float vk_X[Nmax], vk_Z[Nmax];            // holds ... for irr. grad.
  int renderRaysFired;                     // counter for rays fired for rendering
  int photonRaysFired;                     // counter for rays fired for photon tracing
  float lastImageYi;

  vector <Material> materialList;
  
  static SoCallbackAction::Response
    materialTrackerCB(void *data, SoCallbackAction *action,
			    const SoNode *node);
  static SoCallbackAction::Response
    groupPreCBHandler(void *data, SoCallbackAction *action,
			    const SoNode *node);
  static SoCallbackAction::Response
    groupPostCBHandler(void *data, SoCallbackAction *action,
			     const SoNode *node);
  

  float sampMapValue(float x, float y); // returns value of sample map

  public:

  // FIXME
  TriangleTree tritree;                 // octtree acceleration structure

  GlobalVars* gv;                           // global variables
  SoPerspectiveCamera *perspectiveCamera;   // camera
  SoPointLight *headLight;                  // camera mounted point light
  Photon_map *importanceMap;                // visual importance map
  Photon_map *gPhotonMap;                   // global photon map 
  Photon_map *cPhotonMap;                   // caustic photon map
  
  Pane();                                  // constructor
  Pane(GlobalVars &globVar);               // constructor
  ~Pane();                                 // deconstuctor
  
  // flags
  static const int C_APPROX   = 1;
  static const int C_SPECULAR = 2;
  static const int C_DIRECT   = 4;
  static const int C_EMITTED  = 8;
  static const int C_CAUSTIC  = 16;
  static const int C_MDIFFUSE = 32;
  static const int C_SIMPLE   = 64;
  static const int C_PREPASS  = 128;
  static const int C_RUSSIANR = 256;
  static const int C_EXPLICITL = 512;
  static const int C_NORMAL   = C_EMITTED | C_DIRECT | C_SPECULAR | C_CAUSTIC;
  static const int C_FULL     = C_NORMAL | C_MDIFFUSE; //62   ^^30

  // refraction constants
  // From http://www.is.kiruna.se/~cjo/d2i/REFRACTION.INDEX3.html
  static const double N_AIR;// = 1.0;   // refr. index for air, 1.00029;
  static const double N_GLASS;// = 1.7; // refr. index for glass, 1.5-1.7

  // returns maximum xyz coordinate of triangle vertex in tree
  double getMaxDist(){
    return(tritree.max_dist_of_vert);
  }

  // Sets a pixel in the image
  void setPix(int x, int y, const SbColor &c, float depth=0.0){
    pix[gv->imgDim[X]*y+x] = SbColor(c);
    pixDepth[gv->imgDim[X]*y+x] = depth;
  }
  
  int init();                          // initalizes rendering engine
  int readTris();                      // reads scene graph into octtree
  void rescanScene();                  // re-reads scene graph
  static const char* getVersion();     // returns source code version

  // adds triangle to octtree
  static void triCBHandler(void *data, SoCallbackAction *action,
			   const SoPrimitiveVertex *v0,
			   const SoPrimitiveVertex *v1,
			   const SoPrimitiveVertex *v2);

  // adds point light into point light list
  static SoCallbackAction::Response
    pntLtCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node);

  // decreases transparency of objects for interactive viewing
  static SoCallbackAction::Response
    transMatCBHandler(void *data, SoCallbackAction *action,
		      const SoNode *node);

  // stores pointer to most recent SoMaterial node
  static SoCallbackAction::Response
    matBlackCBHandler(void *data, SoCallbackAction *action,
		 const SoNode *node);

  
  // computes refractive direction of ray or signals total internal refl.
  void refract(SbVec3f &ret, int &tir, SbVec3f &d, SbVec3f n,
	       double nnt); // n/nt

  // computes reflected radiance due to point lights
  SbColor pointLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng);
  SbColor pointLightContrib2(Ray &r, SbVec3f &n, SbVec3f &ng);

  // experimental volume scattering effect
  SbColor volPointLightContrib(Ray &r, SbColor &c);
  
  // computes reflected radiance due to area lights, in 2D
  SbColor areaLightContrib2D(Ray &r, SbVec3f n, SbVec3f ng);

  // computes reflected radiance due to area lights, sampling every light
  SbColor areaLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng);

  // computes reflected radiance due to area lights, importance sampling the lights
  SbColor areaImpLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng, int simple=0);

  // precomputes tables needed for multipleDiffuseContrib()
  void precomputeMDCTables();

  // computes reflected radiance due to indirect light
  SbColor multipleDiffuseContrib(Ray &r, SbVec3f &n, SbVec3f &ngl,
				 int simpleCalc, int prepass, int inOrOut);

  // computes (refracted+reflected) radiance due to transparent dielectric
  SbColor refractionContrib(Ray &r, SbVec3f &n, SbVec3f &nl, int flags);

  // function pointer to preferred radiance computing member
  SbColor (Pane::*getRadiance)(Ray &r, int flags, Triangle *ignoreTri);

  // photon mapping radiance computation
  SbColor getRadiance1(Ray &r, int flags=C_NORMAL, Triangle *ignoreTri = NULL);

  // path tracing radiance computatoin
  SbColor getRadiance2(Ray &r, int flags=C_NORMAL, Triangle *ignoreTri = NULL);

  void findCamera();                // set camera pointer
  void configureHeadlight();        // update camera-mounted light in lit pts. list
  void canvasSetup();               // sets up canvas camera settings
  void render(int trace=0);         // compute image pixel radiance values
  void renderTile(thread_context *ctx); // compute image region's pixel values
  void renderVerts(thread_context *ctx);// computes vertex colors
  void renderIv(thread_context *ctx);   // compute per-vertex radiance values
  void toneMap();                   // tone maps the image
  void writeHDR();
  void readHDR();
  int readGlareFilter(char *fn, int *fdim);
  void glareFilter(float Yi, float Yw);
  void checkForOverwrite(char *filename);
  string createComment();
  void computeNormals();
  void writeImg();                  // writes image as PNG file
  void writeIv();                   // outputs scene to inventor file
  void shootPhotons();              // distribute photons throughout scene
  
  // shoot photons from lights
  void emitPhotons(thread_context *ctx, float numToFire,
		   photon_t phot_type);
  
  // propogate photons through scene
  void bouncePhoton(thread_context *ctx, Ray photon, SbColor emission,
		    photon_t phot_type=GLOBAL, float maxPhotonComponent = 1.0,
		    int light_id=1,
		    Triangle *ignoreTri = NULL);

  // wrapper function, glue for threading
  static void *renderTileWrapper(void *arg){
    
    MemberFunction *memberFunc = static_cast <MemberFunction*> (arg);
    Pane *pane = memberFunc->object;
    void (Pane::*pmf)(thread_context*) = memberFunc->pmf;
    (pane->*pmf)(memberFunc->ctx);
    return(NULL);
  }

  // wrapper function, glue for threading
  static void *emitPhotonsWrapper(void *arg){
    
    MemberFunction *memberFunc = static_cast <MemberFunction*> (arg);
    Pane *pane = memberFunc->object;
    void (Pane::*pmf)(thread_context*, float numToFire, photon_t) =
      memberFunc->pmf2;
    (pane->*pmf)(memberFunc->ctx, memberFunc->numToFire, memberFunc->phot_type);
    return(NULL);
  }

};

#endif /* PANECLASS */
