//WWWWWW'''''''WW''''''''''''''''WWW'''''''WWWWWW''''''''''''''''
//  WW                            WW         WW                  
//  WW    W WW WW  WWW WWW   WWW  WW   WW    WW    W WW  WW   WW 
//  WW    WW   WW W WW WW W W WW  WW  WWWW   WW    WW   WWWW WWWW
//  WW    WW   WW W WW WW W W WW  WW  WW     WW    WW   WW   WW  
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW   WW    WW    WWW  WWW
//                            WW                                 
//''''''''''''''''''''''''''WWW''''''''''''''''''''''''''''''''''


#include<stack>
#include "TriangleTree.h"
#include "intersect_tri_cube.h"
#include "debug.h"
#include "Ray.h"


const int TriangleTree::delta[8][3] = {
  0, 0, 0,
  0, 0, 1,
  0, 1, 0,
  0, 1, 1,
  
  1, 0, 0,
  1, 0, 1,
  1, 1, 0,
  1, 1, 1
};
  

TriangleTree::TriangleTree(){
  min[X] = min[Y] = min[Z] = 1e99;
  max[X] = max[Y] = max[Z] = -1e99;
  shiny_min.setValue(1e99f, 1e99f, 1e99f);
  shiny_max.setValue(-1e99f, -1e99f, -1e99f);
  root = new node;
  for (int i=0; i<8; i++)
    root->child[i] = NULL;
  root->tri_list = &tri_list;
  root->mid[0] = root->mid[1] = root->mid[2] = 0.0;
  num_tris = 0;
  max_dist_of_vert = 0.0;
  octreeDebug = 0;
  maxDepth = 8;
  maxItems = 10;
  rayId = 0;
  pthread_mutex_init(&lock, NULL);
  rays = 0;
}

// recursive deletes nodes/children of nodes (but not triangles)
void TriangleTree::deleteNode(node *n){
  if (n == NULL)
    return;

  // delete children
  if (n->tri_list == NULL)
    for (int i=0; i<8; i++)
      deleteNode(n->child[i]);

  // delete self
  delete n;
}
  
TriangleTree::~TriangleTree(){
  struct timeval start, end;
  cerr << "Deleting triangle tree..." << flush;
  gettimeofday(&start, NULL); // start delay measure

  // delete triangles
  cerr << "(triangles...)" << flush;
  tri_list.deleteContents();

  cerr << "(octree nodes...)" << flush;
  // delete nodes
  deleteNode(root);

  // destroy mutex
  pthread_mutex_destroy( &lock );

  gettimeofday(&end, NULL);   // end delay measure
  cerr << " done. (" << delay(start, end) << "ms)" << endl;
}


// Resets tree, deleting all nodes
// This is called by rescanScene() to regenerate an octree
// because of changed geometry.
void TriangleTree::resetTree(){

  // clear tree visualization
  showCube(SbVec3f(0.0, 0.0, 0.0), SbVec3f(0.0, 0.0, 0.0),
	   SbVec3f(0.0, 0.0, 0.0), 1);

  // clear tree
  tri_list.deleteContents();
  deleteNode(root);

  // reset variables
  root = new node;
  for (int i=0; i<8; i++){
    root->child[i] = NULL;
  }
  root->tri_list = &tri_list;
  root->mid[0] = root->mid[1] = root->mid[2] = 0.0;
  min[X] = min[Y] = min[Z] = 1e99;
  max[X] = max[Y] = max[Z] = -1e99;
  shiny_min.setValue(1e99f, 1e99f, 1e99f);
  shiny_max.setValue(-1e99f, -1e99f, -1e99f);
  max_dist_of_vert = 0.0;
}
  
// set triangle tree parameters
void TriangleTree::setParams(int maxD, int maxI, int debug){
  maxDepth = maxD;
  maxItems = maxI;
  octreeDebug = debug;
}

// creates the octree structure
void TriangleTree::refine(int simpleDiv){
    
  // adjust first node boundaries
  
  // foreach triangle
  vector<Triangle *>::iterator I = tri_list.begin();
  for (; I != tri_list.end(); I++){
    
    // foreach vertex
    for (int i=0; i<3; i++)
      
      // find max/min value
      for (int j=0; j<3; j++){
	
	if ((*I)->vert[i][j] > max[j]){
	  //showPoint((*I)->vert[i], 3.0); // debugging
	  max[j] = (*I)->vert[i][j];
	}
	if ((*I)->vert[i][j] < min[j]){
	  //showPoint((*I)->vert[i], 3.0); // debugging
	  min[j] = (*I)->vert[i][j];
	}

	// specular bounding box
	if ((*I)->shiny){
	  if ((*I)->vert[i][j] > shiny_max[j])
	    shiny_max[j] = (*I)->vert[i][j];
	  if ((*I)->vert[i][j] < shiny_min[j])
	    shiny_min[j] = (*I)->vert[i][j];
	}

      }
  }

  // setup boundaries of first node
  for (int i=0; i<3; i++){
    min[i] -= EPSILON_OS;
    max[i] += EPSILON_OS;
    shiny_min[i] -= EPSILON_OS;
    shiny_max[i] += EPSILON_OS;
    root->mid[i] = (min[i]+max[i])/2.0;
    if (max[i]-min[i] > max_dist_of_vert)
      max_dist_of_vert = max[i]-min[i];
  }

  // find bounding sphere
  center = SbVec3f(min[X]+max[X], min[Y]+max[Y], min[Z]+max[Z])/2.0;
  SbVec3f maxCorner(max[X], max[Y], max[Z]);
  radius = (maxCorner-center).length();
  //showPoint(center, radius, SbColor(0.0, 0.0, 1.0));

  // shiny bounding sphere
  shiny_center = (shiny_min+shiny_max)/2.0;
  shiny_radius = (shiny_max-shiny_center).length();
  //showPoint(shiny_center, shiny_radius, SbColor(1.0, 0.0, 0.0));
  
#ifdef SHOWTRAV  
  // debugging
  root->min[0] = min[0];
  root->min[1] = min[1];
  root->min[2] = min[2];
  root->max[0] = max[0];
  root->max[1] = max[1];
  root->max[2] = max[2];
  root->depth = 0;
#endif
  
  if (min[X] == max[X] ||
      min[Y] == max[Y] ||
      min[Z] == max[Z]){
    cerr << "You forgot to set the size of the tree before refine().\n";
    exit(1);
  }
    
  refineNode(root, 0, simpleDiv, min, max);

  if (octreeDebug){
    showCube(SbVec3f(min[X], min[Y], min[Z]),
	     SbVec3f(max[X], max[Y], max[Z]),
	     SbColor(0.0, 1.0, 1.0));
//    showCube(SbVec3f(-0.5, -0.5, -0.5),
//	     SbVec3f(0.5, 0.5, 0.5),
//	     SbColor(0.0, 0.0, 1.0));
  }
}


const static int mod3[5] = { 0, 1, 2, 0, 1 };

double TriangleTree::findDividingPlane(node *n, int a, 
				       double vmin[3], double vmax[3]){
  
  int maxErr = MAX2((int)(0.05*n->tri_list->size()), 1);
  int splits = 0;
  double x0 = vmin[a], x1 = vmax[a];
  int c0, c1, cb;
  int diff = maxErr+10; // guarantees >=1 iteration, which inits this correctly
  double xm = (vmin[a]+vmax[a])*0.5; // middle point
  double xn;                         // median point
  
  // binary search for median
  while (diff>maxErr && splits++<10){
//     cerr << "axis=" << a << " "
// 	 << "maxErr=" << maxErr << " "
// 	 << "x0,x1=" << x0 << "," << x1 << " ";
    
    xn = (x0+x1)*0.5;

    // count triangles less than and greater than xn
    c0 = c1 = 0;
    vector<Triangle *>::iterator I = n->tri_list->begin();
    for (; I != n->tri_list->end(); I++){
      int l=0, r=0;
      for (int i=0; i<3; i++){
	if ((*I)->vert[i][a] > xn)
	  r=1;
	else
	  l=1;
      }
      if (l&&r)
	continue;
      if (l)
	c0++;
      if (r)
	c1++;
    }

//    cerr << "c0,c1=" << c0 << "," << c1 << "\n";

    if (c0>=c1)
      x1 = xn;
    else
      x0 = xn;
    diff = abs(c1-c0);
  }


  // minimize E from equation 8 in Octree-R paper
  int divs = 10;

  double va = vmax[a]-vmin[a];
  double vb = vmax[mod3[a+1]]-vmin[mod3[a+1]];
  double vc = vmax[mod3[a+2]]-vmin[mod3[a+2]];

  //cerr << "va,vb,vc = " << va << ", " << vb << ", " << vc << "\n";

  double min_e = 1e99;
  double min_x = xm;

  // loop over points between xm and xn
  for (int k=0; k<=divs; k++){
    double x = (xn-xm)/divs * k + xm;
    double t = x-vmin[a];

    // count triangles less than and greater than x
    c0 = c1 = cb = 0;
    vector<Triangle *>::iterator I = n->tri_list->begin();
    for (; I != n->tri_list->end(); I++){
      int l=0, r=0;
      for (int i=0; i<3; i++){
	if ((*I)->vert[i][a] > x)
	  r=1;
	else
	  l=1;
      }
      if (l&&r)
	cb++;
      else if (l)
	  c0++;
      else // if (r)
	c1++;
    }

    double d = va*(vb+vc)+vb*vc;
    double e =
      (     t*(vb+vc)+vb*vc)/d      * c0 +
      ((va-t)*(vb+vc)+vb*vc)/d      * c1 +
      (    va*(vb+vc)+2.0*vb*vc)/d  * cb;

    if (e<min_e){
      min_e = e;
      min_x = x;
//       cerr << "k=" << k << " "
// 	   << "e=" << e << "\n";
    }
    
  }
//  cerr << "\n";
  return(min_x);
}


void TriangleTree::refineNode(node *n, int depth, int simpleDiv,
			      double vmin[3], double vmax[3]){

  SbMatrix scale;
  double nmin[3], nmax[3];

  // subdivide node
  if (depth<maxDepth && n->tri_list && n->tri_list->size()>maxItems){
    int i;

    // debugging info
    if (0 && octreeDebug){
      for (i=0; i<depth; i++)
	cerr << "  ";
      cerr << "refineNode(" << n << ") "
	   << "(" << vmin[X] << ", " << vmax[X] << ") "
	   << "(" << vmin[Y] << ", " << vmax[Y] << ") "
	   << "(" << vmin[Z] << ", " << vmax[Z] << ")"
	   << " depth=" << depth
	   << " items=" << n->tri_list->size() << endl;;
    }

    // set new size
    if (simpleDiv){        // midpoint 
      n->mid[X] = 0.5*(vmax[X]-vmin[X])+vmin[X]; 
      n->mid[Y] = 0.5*(vmax[Y]-vmin[Y])+vmin[Y];
      n->mid[Z] = 0.5*(vmax[Z]-vmin[Z])+vmin[Z];
    }
    else {                 // optimized point for subdivision
      n->mid[X] = findDividingPlane(n, X, vmin, vmax);
      n->mid[Y] = findDividingPlane(n, Y, vmin, vmax);
      n->mid[Z] = findDividingPlane(n, Z, vmin, vmax);
    }

    // for each child
    int childCreated = 0;
    for (i=0; i<8; i++){

      childCreated = 0;

      nmin[X] = delta[i][X] ? n->mid[X] : vmin[X]; 
      nmin[Y] = delta[i][Y] ? n->mid[Y] : vmin[Y]; 
      nmin[Z] = delta[i][Z] ? n->mid[Z] : vmin[Z];
      nmax[X] = delta[i][X] ? vmax[X] : n->mid[X];
      nmax[Y] = delta[i][Y] ? vmax[Y] : n->mid[Y];
      nmax[Z] = delta[i][Z] ? vmax[Z] : n->mid[Z];

      // debugging info: show octree/child cell
      SbColor color;
      color.setHSVValue(clamp((double)n->tri_list->size(),
			      (double)0.0,(double)36.0)/36.0,
			1.0, 0.5);

      // add appropriate children to new child node
      //   run triangle cube intersection for each triangle
      double boxcenter[3], boxhalfsize[3];
      boxcenter[0] = (nmax[X]+nmin[X])/2.0;
      boxcenter[1] = (nmax[Y]+nmin[Y])/2.0;
      boxcenter[2] = (nmax[Z]+nmin[Z])/2.0;
      boxhalfsize[0] = nmax[X]-boxcenter[0];
      boxhalfsize[1] = nmax[Y]-boxcenter[1];
      boxhalfsize[2] = nmax[Z]-boxcenter[2];

      // test each triangle
      int children = 0;
      int tri_num = 0;
      vector<Triangle *>::iterator I = n->tri_list->begin();
	
      for (; I != n->tri_list->end(); I++){
	tri_num++;

	if (intersect_tri_cube(boxcenter, boxhalfsize, **I)){
	
	  if (!childCreated){
	    // create child
	    n->child[i] = new node;
	    for (int j=0; j<8; j++)
	      n->child[i]->child[j] = NULL;
	    n->child[i]->tri_list = new TriangleList();
	    n->child[i]->mid[0] = boxcenter[0];
	    n->child[i]->mid[1] = boxcenter[1];
	    n->child[i]->mid[2] = boxcenter[2];
	    childCreated++;

#ifdef SHOWTRAV	    
	    // debugging
	    n->child[i]->min = SbVec3f(nmin[X], nmin[Y], nmin[Z]);
	    n->child[i]->max = SbVec3f(nmax[X], nmax[Y], nmax[Z]);
	    n->depth=depth+1;
#endif	    

	    if (octreeDebug)
	      showCube(SbVec3f(nmin[X], nmin[Y], nmin[Z]),
		       SbVec3f(nmax[X], nmax[Y], nmax[Z]),
		       color);
	  }
	    
	  n->child[i]->tri_list->add(*I);
	  children++;
	}
      }

      // refine new child node node
      if (childCreated)
	refineNode(n->child[i], depth+1, simpleDiv, nmin, nmax);
	
    } // end for(1..8)

      // remove Triangles from tri_list, set terminal to 0
    if (n != root){
      n->tri_list->clear();
      delete n->tri_list;
    }
    n->tri_list = NULL;
      
  } // end if (depth<max_depth...)
    
  else{ // if (n->tri_list.size()){

    // debugging info: print cell info
    if (0 && octreeDebug){
      if (1 || n->tri_list->size()){ // all or just empty cells
	for (int x=0; x<depth; x++)
	  cerr << "  ";
	cerr << n << " ";
	cerr << n->tri_list->size() << " triangles depth=" << depth << endl;
      }
    }
      
  }

} // end of refineNode();


  /*

  This Octree Traversal code is taken from
  "An Efficient Parametric Algorithm for Octree Traversal"
  by J. Revelles, C. Urena, M. Lastra

  */

  // returns i-th int, where i'th float is minimum
inline int TriangleTree::new_node(double &t1, int i1, double &t2, int i2,
				  double &t3, int i3) const{
  return( (t1<t2&&t1<t3) ? i1 : (t2<t3 ? i2 : i3) );
}

// returns first child node crossed
int TriangleTree::first_node(double &tx0, double &ty0, double &tz0,
			     double &txm, double &tym, double &tzm) const{
    
  unsigned char b = 0;
  if (tx0>ty0 && tx0>tz0){   // tx0 max
    if (tym<tx0)
      b |= 2;
    if (tzm<tx0)
      b |= 1;
  }
  else if (ty0>tz0){         // ty0 max
    if (txm<ty0)
      b |= 4;
    if (tzm<ty0)
      b |= 1;
  }
  else {                     // tz0 max
    if (txm<tz0)
      b |= 4;
    if (tym<tz0)
      b |= 2;
  }

  return(b);
}

int TriangleTree::intersect_ray(Ray &r, Triangle *ignoreTri) const{

  // count rays
  r.ctx->rays++;
  
  // return value pointers (private class members)
  treeParam tp;
  tp.r_p = &r;
  tp.ignoreTri_p = ignoreTri;

  // increment rayId (must be non-zero) <-- relaxed because risk is remote
  //if (++rayId==0)
  /*
  pthread_mutex_lock( &lock );
  tp.rayId = ++rayId;
  pthread_mutex_unlock( &lock );
  */

  // function variables
  SbVec3f pos = r.pos;
  SbVec3f dir = r.dir;
  double mid[3];
  double tx0, tx1, ty0, ty1, tz0, tz1;

  // calculate midplanes
  mid[0] = (max[0]+min[0])/2.0;
  mid[1] = (max[1]+min[1])/2.0;
  mid[2] = (max[2]+min[2])/2.0;
    
  tp.a = 0;
  if (dir[X]<0.0){
    pos[X] = 2.0*mid[X]-pos[X];
    dir[X] = -dir[X];
    tp.a |= 4;
  }
  if (dir[Y]<0.0){
    pos[Y] = 2.0*mid[Y]-pos[Y];
    dir[Y] = -dir[Y];
    tp.a |= 2;
  }
  if (dir[Z]<0.0){
    pos[Z] = 2.0*mid[Z]-pos[Z];
    dir[Z] = -dir[Z];
    tp.a |= 1;
  }

  tx0 = (min[X]-pos[X])/dir[X];
  tx1 = (max[X]-pos[X])/dir[X];
  ty0 = (min[Y]-pos[Y])/dir[Y];
  ty1 = (max[Y]-pos[Y])/dir[Y];
  tz0 = (min[Z]-pos[Z])/dir[Z];
  tz1 = (max[Z]-pos[Z])/dir[Z];

  if (MAX3(tx0,ty0,tz0) >= MIN3(tx1,ty1,tz1))
    return(0);
  
  return( proc_subtree(tx0,ty0,tz0, tx1,ty1,tz1, root, tp) );
}

int TriangleTree::proc_subtree(double &tx0, double &ty0, double &tz0,
			       double &tx1, double &ty1, double &tz1,
			       node *n, treeParam &tp) const{

  double txm, tym, tzm;
  unsigned int currNode;

  // don't examine volumes behind the position
  if (n == NULL || tx1<0.0 || ty1<0.0 || tz1<0.0)
    return(0);

#ifdef SHOWTRAV  
  // debugging
  if (DEBUG_STATE){
    cerr << "Size: " << n->max-n->min << "\n";
    SbColor color;
    color.setHSVValue(n->depth/8.0, 1.0, 0.5);
    showCube(n->min, n->max, color);
  }
#endif
  
  // proc_terminal(n);
  if (n->tri_list != NULL){

    if (n->tri_list->intersect_ray(*tp.r_p, tp.ignoreTri_p) &&
	// special case: must be in cell
	tp.r_p->t < MIN3(tx1, ty1, tz1)+EPSILON_OI
	/*&& tp.r_p->t > MAX3(tx0, ty0, tz0)-EPSILON_OI*/) // unnecessary
	return(1);
    else
      return(0);
  }


  // this code is for axis-aligned direction handling (special case)
  // see paper Sec.3.3 in Efficient Parametric Algorithm for Octree Traversal
  if (isinf(tx1) == 1){
    if (tp.r_p->pos[X] < n->mid[X])
      txm = tx1;
    else
      txm = -tx1;
  }
  else
    txm = (n->mid[X]-tp.r_p->pos[X])/tp.r_p->dir[X]; // 0.25*(tx0+tx1);

  if (isinf(ty1) == 1){
    if (tp.r_p->pos[Y] < n->mid[Y])
      tym = ty1;
    else
      tym = -ty1;
  }
  else
    tym = (n->mid[Y]-tp.r_p->pos[Y])/tp.r_p->dir[Y]; // 0.25*(ty0+ty1);
  
  if (isinf(tz1) == 1){
    if (tp.r_p->pos[Z] < n->mid[Z])
      tzm = tz1;
    else
      tzm = -tz1;
  }
  else
    tzm = (n->mid[Z]-tp.r_p->pos[Z])/tp.r_p->dir[Z]; // 0.25*(tz0+tz1);
  
  currNode = 0;
  if (tx0>ty0 && tx0>tz0){   // tx0 max
    if (tym<tx0)
      currNode |= 2;
    if (tzm<tx0)
      currNode |= 1;
  }
  else if (ty0>tz0){         // ty0 max
    if (txm<ty0)
      currNode |= 4;
    if (tzm<ty0)
      currNode |= 1;
  }
  else {                     // tz0 max
    if (txm<tz0)
      currNode |= 4;
    if (tym<tz0)
      currNode |= 2;
  }
  
  do {
    switch (currNode){
    case 0:
      if ( proc_subtree(tx0, ty0, tz0, txm, tym, tzm, n->child[tp.a], tp) )
	return(1);
      currNode = new_node(txm, 4, tym, 2, tzm, 1);
      break;
	
    case 1:
      if ( proc_subtree(tx0, ty0, tzm, txm, tym, tz1, n->child[1^tp.a], tp) )
	return(1);
      currNode = new_node(txm, 5, tym, 3, tz1, 8);
      break;

    case 2:
      if ( proc_subtree(tx0, tym, tz0, txm, ty1, tzm, n->child[2^tp.a], tp) )
	return(1);
      currNode = new_node(txm, 6, ty1, 8, tzm, 3);
      break;

    case 3:
      if ( proc_subtree(tx0, tym, tzm, txm, ty1, tz1, n->child[3^tp.a], tp) )
	return(1);
      currNode = new_node(txm, 7, ty1, 8, tz1, 8);
      break;

    case 4:
      if ( proc_subtree(txm, ty0, tz0, tx1, tym, tzm, n->child[4^tp.a], tp) )
	return(1);
      currNode = new_node(tx1, 8, tym, 6, tzm, 5);
      break;

    case 5:
      if ( proc_subtree(txm, ty0, tzm, tx1, tym, tz1, n->child[5^tp.a], tp) )
	return(1);
      currNode = new_node(tx1, 8, tym, 7, tz1, 8);
      break;

    case 6:
      if ( proc_subtree(txm, tym, tz0, tx1, ty1, tzm, n->child[6^tp.a], tp) )
	return(1);
      currNode = new_node(tx1, 8, ty1, 8, tzm, 7);
      break;

    case 7:
      if ( proc_subtree(txm, tym, tzm, tx1, ty1, tz1, n->child[7^tp.a], tp) )
	return(1);
      currNode = 8;
      break;
	
    default:
      //cerr << "Invalid case " << currNode << " in switch(currNode)!" << endl;
      break;
    } // end switch
      
  } while (currNode<8); // end do {} while

  return(0);
} // end proc_subtree()

  // single sided intersection routine
  //   calls regular intersection routine
int TriangleTree::ss_intersect_ray(Ray &r){
  int retVal = intersect_ray(r);
    
  if (retVal && r.tri->getNormal().dot(r.dir)>0.0){
    Ray new_r = Ray(r.pos+(r.t+0.0001)*r.dir, r.dir, r.ctx);
    retVal = ss_intersect_ray(new_r);
    r.t += new_r.t;
  }

  return(retVal);
}

void TriangleTree::printStats(node *ptr, int *numCubes, int *numTris){

  // if no parameters, call self on root
  if (ptr == NULL){
    int ncubes = 1, ntris = 0;
    printStats(root, &ncubes, &ntris);
    cerr << "TriangleTree statistics: "
	 << ncubes << " cubes, "
	 << ntris << " triangles, "
	 << num_tris << " initially ("
	 << ntris/float(num_tris)*100.0 << "%).\n";
    return;
  }

  // if leaf node, count tets
  if (ptr->tri_list != NULL){
    *numTris += ptr->tri_list->size();
  }

  // otherwise, recurse over children
  else {
    for (int i=0; i<8; i++){
      if (ptr->child[i] != NULL){
	(*numCubes)++;
	printStats(ptr->child[i], numCubes, numTris);
      }
    }
  }

  return;
}


void TriangleTree::reorderNodes(){

  
  // count nodes
  node *p;
  int nodeCnt = 0;
  stack<node *> nodes;
  nodes.push(root);

  while (nodes.empty() == 0){
    
    p = nodes.top();
    nodes.pop();
    nodeCnt++;
    
    for (int i=0; i<8; i++)
      if (p->child[i])
	nodes.push(p->child[i]);
  }

  cerr << "Counted " << nodeCnt << " nodes.\n";

  // allocate space for nodes
  compactNodes = new node[nodeCnt];
  int nodeIdx = 0;
  node *q;
  
  compactNodes[nodeIdx] = *root;
  delete root;
  root = &compactNodes[nodeIdx++];
  nodes.push(root);

  while (nodes.empty() == 0){
    
    p = nodes.top(); // p already in compactNodes
    nodes.pop();

    // move p's children
    for (int i=0; i<8; i++)
      if (p->child[i]){

	// copy node
	compactNodes[nodeIdx] = *(p->child[i]);
	delete p->child[i];
	p->child[i] = &compactNodes[nodeIdx++];

	// add node to processing list
	nodes.push(p->child[i]);
      }
  }

  cerr << "Moved " << nodeIdx << " nodes.\n";

  return;
}
