////////////////////////////////////////////////////////////////////////////
//
//WWWW'''''''''''''''''''''''''WWW''''''''''''''''''
//WW  W                         WW                  
//WW  W  WWW WWW   WW      WWW  WW   WWW  WWWW  WWWW
//WWWW  W WW WW W WWWW    WW W  WW  W WW WWW   WWW  
//WW    W WW WW W WW      WW    WW  W WW   WWW   WWW
//WW     WWW WW W  WWW     WWW WWWW  WWW WWWW  WWWW 
//                                                  
//''''''''''''''''''''''''''''''''''''''''''''''''''
//
// File: Pane.cxx
// Description: Pane class (Core Ray Tracing Engine) implementation
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Rendering engine for the Pane ray tracer
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// 
// $Id: Pane.cxx,v 1.127 2005/09/21 04:48:03 beason Exp beason $
//
////////////////////////////////////////////////////////////////////////////

#include <limits.h>
#include <stack>
#include "Pane.h"
#include "sobol.h"
#include "Halton.h"
#include "debug.h"
#include "Texture.h"
#include "Material.h"

static char rcsid[] = "@(#) $Id: Pane.cxx,v 1.127 2005/09/21 04:48:03 beason Exp beason $";
const double Pane::N_AIR = 1.0;   // refr. index for air, 1.00029;
const double Pane::N_GLASS = 1.7; // refr. index for glass, 1.5-1.7

//'WWW''''''''''''''''''''WW''''''''''''''''''WW''''''''''''''''''
//WW  W                   WW                  WW                  
//WW     WW   WWW   WWWW WWWW W WW W WW  WWW WWWW  WW   W WW  WWWW
//WW    WW W  WW W WWW    WW  WW   W WW WW W  WW  WW W  WW   WWW  
//WW  W WW W  WW W   WWW  WW  WW   W WW WW    WW  WW W  WW     WWW
// WWW   WW   WW W WWWW   WW  WW    WWW  WWW  WW   WW   WW   WWWW 
//                                                                
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Constructors
//

Pane::Pane(){
  gv = NULL;
  pix = NULL;
  pixDepth = NULL;
  gPhotonMap = NULL;
  cPhotonMap = NULL;
  totalPower = 0.0;
  max_photons = 0;
  firedPhotons = 0;
  minApproxDistance = 0.0;
  rejectedICPoints = 0;
  toneMapped = 0;
  sampMap = NULL;
  numLights = 0;
  maxEmisAvg = 0.0;
  totalEmisAvg = 0.0;
  renderRaysFired = 0;
  photonRaysFired = 0;
  lastImageYi = 0.0;
  envMap = NULL;
}

// Constructor with initializer
Pane::Pane(GlobalVars &globVar){
  gv = &globVar;
  pix = NULL;
  pixDepth = NULL;
  perspectiveCamera = NULL;
  headLight = NULL;
  totalPower = 0.0;
  max_photons = 0;
  firedPhotons = 0;
  minApproxDistance = 0.0;
  gPhotonMap = NULL; 
  cPhotonMap = NULL;
  rejectedICPoints = 0;
  toneMapped = 0;
  sampMap = NULL;
  numLights = 0;
  maxEmisAvg = 0.0;
  totalEmisAvg = 0.0;
  renderRaysFired = 0;
  photonRaysFired = 0;
  lastImageYi = 0.0;
  envMap = NULL;
}


//WWWW''''''''''''''''''''''''''''''WW''''''''''''''''''WW''''''''''''
//WW  W                             WW                  WW            
//WW  W  WW   WWW  WW   WWW   WWWW WWWW W WW W WW  WWW WWWW  WW   W WW
//WW  W WWWW WW W WW W  WW W WWW    WW  WW   W WW WW W  WW  WW W  WW  
//WW  W WW   WW   WW W  WW W   WWW  WW  WW   W WW WW    WW  WW W  WW  
//WWWW   WWW  WWW  WW   WW W WWWW   WW  WW    WWW  WWW  WW   WW   WW  
//                                                                    
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Deconstructor
//
Pane::~Pane(){

  // remove headlight at front of litPnts
  if (gv->cam)
    litPnts[0]->unref();

  delete [] pix;
  delete [] pixDepth;
  delete envMap;
}
  

//WW''''WW'''''''WW''WW'''''''''''WW''''''
//WWW  WWW       WW  WW           WW      
//W WWW WW  WW  WWWW WWW   WW    WWW  WWWW
//W  W  WW WWWW  WW  WW W WW W  W WW WWW  
//W     WW WW    WW  WW W WW W  W WW   WWW
//W     WW  WWW  WW  WW W  WW    WWW WWWW 
//                                        
//''''''''''''''''''''''''''''''''''''''''


// Return string containing version number
const char* Pane::getVersion(){
  static char info[80] =
    "$Revision: 1.127 $";
  char *p = strchr(info, ' ')+1;
  char *q = strsep(&p, " ");
  return(q);
}

// Rebuild octree (for animation and demo purposes)
void Pane::rescanScene(){
  tritree.resetTree();
  litPnts.resize(1);    // remove all but headlight
  litTris.clear();
  readTris();
}   
  
// PointLight CB Handler
// Stores PointLight information when traversing scene graph
SoCallbackAction::Response
Pane::pntLtCBHandler(void *data, SoCallbackAction *action,
		     const SoNode *node){
  Pane *p = (Pane *)data;
  p->litPnts.push_back((SoPointLight *)node);
  p->gv->sceneHasLights = 1;
  p->numLights++;
  return(SoCallbackAction::CONTINUE);
}

// SoMaterial CB Handler
// reduces transparency for viewing purposes
// Note: If the node is referenced more than once
// the reduction will accumulate.
SoCallbackAction::Response
Pane::transMatCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node){
  
  SoMaterial *mat = (SoMaterial *)node;
  Pane *pane = (Pane *)data;

  // reduce material transparency by 25%
  int num = mat->transparency.getNum();
  float *trans = mat->transparency.startEditing();
  for (int i=0; i<num; i++)
    trans[i] *= pane->gv->transpScale;
  mat->transparency.finishEditing();
  
  return(SoCallbackAction::CONTINUE);
}

// Material CB Handler
// turns surfaces black, for photon viewing
SoCallbackAction::Response
Pane::matBlackCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node){
  SoMaterial *mat = (SoMaterial *)node;

  for (int i=0; i<mat->ambientColor.getNum(); i++)
    mat->ambientColor.set1Value(i, 0.0, 0.0, 0.0);
  for (int i=0; i<mat->diffuseColor.getNum(); i++)
    mat->diffuseColor.set1Value(i, 0.0, 0.0, 0.0);
  for (int i=0; i<mat->specularColor.getNum(); i++)
    mat->specularColor.set1Value(i, 0.0, 0.0, 0.0);
  for (int i=0; i<mat->emissiveColor.getNum(); i++)
    mat->emissiveColor.set1Value(i, 0.0, 0.0, 0.0);
  
  return(SoCallbackAction::CONTINUE);
}

// Triangle CB Handler
// Puts triangle into octtree
void Pane::triCBHandler(void *data, SoCallbackAction *action,
			const SoPrimitiveVertex *v0,
			const SoPrimitiveVertex *v1,
			const SoPrimitiveVertex *v2){

  Triangle *t = new Triangle(data, action, v0, v1, v2);
  if (t->area<=0.0)
    return;

  // add triangle to tree
  Pane *pane = (Pane *)data;
  pane->tritree.add(t);

  // add triangle to lights list
  if (t->emissive>0.0){

    // record keeping
    pane->numLights++;
    pane->totalEmisAvg += t->emisAvg;
    if (t->emisAvg > pane->maxEmisAvg)
      pane->maxEmisAvg = t->emisAvg;
    
    pane->litTris.add(t);
    
  }

}


// keep a pointer to the actual current material
// complex because it's accurate and
// keeps track of position in scene graph,
// changing when a group containing a material
// is left

typedef struct matListItem {
  SoMaterial *mat;     // is this necessary?
  int        matIndex;
  SoGroup    *matParent;
} matListItem_t;

SoMaterial   *currMaterial;
SoGroup      *currGroup;
SoShapeHints *currShapeHints;
stack <matListItem_t *> materialStack;
stack <SoGroup *> groupStack;

// Material CB Handler
// Stores pointer to most recent SoMaterial node in
// scene graph traversal.
SoCallbackAction::Response
Pane::materialTrackerCB(void *data, SoCallbackAction *action,
			const SoNode *node){
  
  currMaterial = (SoMaterial *)node;
  Pane *pane = (Pane *)data;

  // create material
  int i = pane->materialList.size();
  pane->materialList[i] = Material(currMaterial);

  // put on stack
  matListItem_t *item = new matListItem_t;
  item->mat = currMaterial;
  item->matIndex = i;
  item->matParent = currGroup;
  
  return(SoCallbackAction::CONTINUE);
}

// Group CB Handler
// Stores pointer to most recent SoGroup node in
// sene graph traversal.
SoCallbackAction::Response
Pane::groupPreCBHandler(void *data, SoCallbackAction *action,
			const SoNode *node){

  // place group on stack
  currGroup = (SoGroup *)node;
  groupStack.push(currGroup);
  
  return(SoCallbackAction::CONTINUE);
}

SoCallbackAction::Response
Pane::groupPostCBHandler(void *data, SoCallbackAction *action,
			 const SoNode *node){

  // remove group from stack
  currGroup = groupStack.top();
  groupStack.pop();

  // if parent of material, pop material from stacc
  matListItem_t *currItem = materialStack.top();
  if (currItem->matParent == currGroup)
    materialStack.pop();
  delete currItem;
  
  return(SoCallbackAction::CONTINUE);
}




// Material CB Handler
// Stores pointer to most recent SoMaterial node in
// scene graph traversal.
SoCallbackAction::Response
materialCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node){
  currMaterial = (SoMaterial *)node;
  
  return(SoCallbackAction::CONTINUE);
}

// Group CB Handler
// Stores pointer to most recent SoGroup node in
// sene graph traversal.
SoCallbackAction::Response
groupCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node){
  currGroup = (SoGroup *)node;
  
  return(SoCallbackAction::CONTINUE);
}

// ShapeHints CB Handler
// Stores pointer to most recent SoShapeHints node in
// sene graph traversal.
SoCallbackAction::Response
shapeHintsCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node){
  currShapeHints = (SoShapeHints *)node;
  
  return(SoCallbackAction::CONTINUE);
}




//WW''''''WW''WW'
//            WW 
//WW WWW  WW WWWW
//WW WW W WW  WW 
//WW WW W WW  WW 
//WW WW W WW  WW 
//               
//'''''''''''''''
//
// class initializer
// builds octree, creates headlight, pixel array
//
int Pane::init(){

  // create photon maps
  max_photons = MAX2(gv->numPhotons, gv->numCPhotons)*2;
  gPhotonMap = new Photon_map(max_photons, gv);
  cPhotonMap = new Photon_map(max_photons, gv);
  importanceMap = new Photon_map(gv->numIPhotons*2, gv);

  // create headlight
  headLight = new SoPointLight();
  headLight->intensity.setValue(1.0);
  litPnts.insert(litPnts.begin(), headLight);

  // initialize pix array
  if (!gv->inImgFile){
    pix = new SbColor[gv->imgDim[X]*gv->imgDim[Y]]; // should i initialize this?
    pixDepth = new float[gv->imgDim[X]*gv->imgDim[Y]];
    for (int j=0; j<gv->imgDim[Y]; j++)
      for (int i=0; i<gv->imgDim[X]; i++)
	pixDepth[gv->imgDim[X]*j+i] = 0.0;
  }
  // load image into pix array (aka framebuffer)
  else 
    readHDR();

  // reset completion meters
  for (int i=0; i<16; i++){
    completion[i] = 0.0;
    completion_stage[i] = ' ';
  }

  // adjust worldBG
  gv->worldBG[R] = pow(gv->worldBG[R], gv->gamma);
  gv->worldBG[G] = pow(gv->worldBG[G], gv->gamma);
  gv->worldBG[B] = pow(gv->worldBG[B], gv->gamma);

  // set intelligent default of lightsamples:
  // roughly 100/pixelsamples. however, since
  // both the pixel and the light sampling
  // use stratified sampling, the number of
  // pixel samples and light samples used are the
  // next smallest perfect square. this code chooses a
  // perfect square for the light samples in order
  // to gaurantee at least 100 light samples per pixel.
  // 
  // lightSamples = next largest perfect square (
  //                      100/smallest perfect square(pixel samples)
  //                )
  gv->totalLightSamples = gv->lightSamples;
  if (gv->lightSamples == -1){
    if ( (gv->algorithm == 1 || gv->algorithm == 0)
	 && gv->areaLights == 1 ){
      
      int samples = (int)sqrt((float)gv->numSamples);
      samples *= samples;
      gv->totalLightSamples = 100;
      gv->lightSamples = MAX2(100/samples, 1); // at least 1

      // find next largest perfect square
      while ( gv->lightSamples -
	      ((int)sqrt((float)gv->lightSamples))*
	      ((int)sqrt((float)gv->lightSamples)) > 0.01 )
	gv->lightSamples++;
      cerr << "Using " << gv->lightSamples << " light samples.\n";
    }
    else {
      gv->lightSamples = 1;
      gv->totalLightSamples = 1;
    }
  }

  // determine number of processors
  if (gv->numThreads == -1){

    int procs = 0;
    char line[1000];
    ifstream cpuinfo("/proc/cpuinfo");
    regex_t regex;
    int reti;
    reti = regcomp(&regex, "^processor", 0);
    if (reti)
      cerr << "Warning: Regular expression parsing failure, reading # cpus!"
	   << "Try using -jN to manually set number of threads.\n";
    else {
      while (cpuinfo){
	cpuinfo.getline(line, 1000, '\n');
	reti = regexec(&regex, line, 0, NULL, 0);
	if (!reti)
	  procs++;
      }
    }
    regfree(&regex);
    cpuinfo.close();
    gv->numThreads = MAX2(1, procs);
    cerr << "Using " << gv->numThreads << " threads.\n";
  }

  // read environment map
  if (gv->environMapFile){
    envMap = new EnvironMap();
    envMap->init(gv);
  }


  //'''''''''''''''W'''''''''''''''''''''''WW'''''''''''''''''''''''
  //               W                        W                       
  // WW  WW   WW  WW     WWW  WW WW W  WW   W   WW     WW W   WW WW 
  //W   Wmmm W W W W     WW  W W W W W W W  W  Wmmm    W W W W W W W
  //W    WW   WW  WW    WWW   WW W W W WW  WWW  WW     W W W  WW WW 
  //                                   W                         W  
  //'''''''''''''''''''''''''''''''''''W'''''''''''''''''''''''''W''
  // read sample map
  if (gv->sampleMap){
    ifstream pgm;
    cerr << "Reading sample map " << gv->sampleMap << "...";
    pgm.open(gv->sampleMap);
    if (!pgm){
      cerr << "Error: Could not open sample map " << gv->sampleMap << ".\n";
      exit(1);
    }
    char buf[100];
    pgm >> buf;
    if (strcmp(buf, "P2")){
      cerr << "Error: Sample map " << gv->sampleMap << " does not appear "
	" to be a PGM file.\n";
      cerr << "Could not recognize: \"" << buf << "\"\n";
      exit(1);
    }

    // read width
    while (pgm >> buf){
      if (buf[0] == '#')
	pgm.getline(buf, 100);
      else
	break;
    }
    sampMapWidth = atoi(buf);

    // read height
    while (pgm >> buf){
      if (buf[0] == '#')
	pgm.getline(buf, 100);
      else
	break;
    }
    sampMapHeight = atoi(buf);

    // read max value
    while (pgm >> buf){
      if (buf[0] == '#')
	pgm.getline(buf, 100);
      else
	break;
    }
    int maxVal = atoi(buf);

    // create space for map
    sampMap = new float[sampMapWidth * sampMapHeight];
    
    // read values
    int f;
    for (int j=sampMapHeight; j>0; j--)
      for (int i=0; i<sampMapWidth; i++){
	pgm >> f;
	sampMap[j*sampMapWidth+i] = f/(float)maxVal;
      }

    // finish up
    pgm.close();
    cerr << "done.\n";
  }
  
  // read scene
  cerr << "Inserting triangles into octree acceleration structure..." << flush;
  return(readTris());
}


//''''''''''''''''''''''''WW''''WW'''''''''''WW''W''''''WWW'''''''''''
//                        WWW  WWW           WW  W       WW           
// WWWW  WWW WWWW W  WWW  W WWW WW  WWW WWW  WW  W  WWW  WW  W WW  WW 
//WWW   W WW WW WW W WW W W  W  WW W WW WW W WW  W W WW  WW  W WW WWWW
//  WWW W WW WW WW W WW W W     WW W WW WW W WW W  W WW  WW  W WW WW  
//WWWW   WWW WW WW W WWW  W     WW  WWW WWW   WW    WWW WWWW  WWW  WWW
//                   WW                 WW                            
//'''''''''''''''''''WW'''''''''''''''''WW''''''''''''''''''''''''''''
//
// returns value in sample map (gray scale, 0<=x<1.0)
// uses hermite smoothing to return smoothly varying result.
//
float Pane::sampMapValue(float x, float y){

  float g;
  float s, t;
  float i_f, j_f;
  float sp, tp;
  int i, j, i1, j1;

  int size[2] = { sampMapWidth, sampMapHeight };

  // calculate image coordinates (interpolate)
  s = x; 
  t = y;

  //  clamp
  if (s<0.0 || s>1.0 || t<0.0 || t>1.0)
    return(0.0);
  
  // calculate texture coordinates
  i_f = size[0]*s-0.5;
  j_f = size[1]*t-0.5;

  // calculate indices
  i  = (int)floor(i_f);
  j  = (int)floor(j_f);
  i1 = i+1;
  j1 = j+1;
	 
  i  = clamp( i, 0, size[0]-1);
  i1 = clamp(i1, 0, size[0]-1);
  j  = clamp( j, 0, size[1]-1);
  j1 = clamp(j1, 0, size[1]-1);
	 
  // calculate texel ratios
  sp = i_f-floor(i_f);
  tp = j_f-floor(j_f);
	 
  // perform hermite smoothing
  // (see Peter Shirley's Realistic Ray Tracing 1st ed.)
  sp = 3.0*sp*sp-2.0*sp*sp*sp;
  tp = 3.0*tp*tp-2.0*tp*tp*tp;

  g = (1.0-sp)*(1.0-tp)*sampMap[size[0]*j +i ] +
      sp*(1.0-tp)      *sampMap[size[0]*j +i1] +
      (1.0-sp)*tp      *sampMap[size[0]*j1+i ] +
      sp*tp            *sampMap[size[0]*j1+i1];
      
  return(g);
}


//'''''''''''''''''WW'WWWWWW'''''''WW''''''
//                 WW   WW                 
//W WW  WW   WWW  WWW   WW    W WW WW  WWWW
//WW   WWWW W WW W WW   WW    WW   WW WWW  
//WW   WW   W WW W WW   WW    WW   WW   WWW
//WW    WWW  WWW  WWW   WW    WW   WW WWWW 
//                                         
//'''''''''''''''''''''''''''''''''''''''''
//
// Read the triangles into internal data structure
//
int Pane::readTris(){
  SoCallbackAction cb;
//    cb.addPreCallback(SoGroup::getClassTypeId(),
//  		    Pane::groupPreCBHandler, this);
//    cb.addPostCallback(SoGroup::getClassTypeId(),
//  		    Pane::groupPostCBHandler, this);
//    cb.addPostCallback(SoMaterial::getClassTypeId(),
//  		     Pane::materialTrackerCB, this);
  cb.addTriangleCallback(SoShape::getClassTypeId(),
			 Pane::triCBHandler, this);
  cb.addPreCallback(SoPointLight::getClassTypeId(),
		    Pane::pntLtCBHandler, this);
  cb.apply(gv->scene);

  // adjust transparency, so we can see 100% transparent objects in viewer
  SoCallbackAction trancb;
  trancb.addPostCallback(SoMaterial::getClassTypeId(),
			 Pane::transMatCBHandler, this);
  trancb.apply(gv->scene);
   
  // refine tree
  tritree.setParams(gv->octreeMaxDepth, gv->octreeMaxItems,
		    gv->octreeDebug);
  tritree.refine(gv->simpleOctree);
  tritree.printStats();

  // Set default meters per unit for scene size of 10m
  if (gv->metersPerUnit<0.0){
    gv->metersPerUnit = 10.0/getMaxDist();
    gv->metersPerUnit2 = gv->metersPerUnit*gv->metersPerUnit;
    gv->unitsPerMeter = 1.0/gv->metersPerUnit;
  }
  
  // set up irradiance cache
  irradCache.setDims(SbVec3f(tritree.min[X], tritree.min[Y], tritree.min[Z]),
		     tritree.max_dist_of_vert, 0);
  irradCache.setA(gv->irradCacheError);
  irradCache.setResolution(gv->ambientResolution);
  irradCache.setGV(gv);

  // set photon map parameter
  if (!gv->sceneHasShininess && gv->useGlassModel)
    gv->numCPhotons = 0;

  Triangle().printGlobalStats();
  
  return(tritree.size());
}

// refract ray
// Computes refracted ray direction or signals
// Total Inteneral Reflection (tir) if that occurs
void Pane::refract(SbVec3f &ret, int &tir,
		   SbVec3f &d, SbVec3f n, double nnt){ // n/nt
  
  double ddn = d.dot(n);
  double costsq = 1.0-nnt*nnt*(1.0-ddn*ddn);
  if (costsq<0.0){  // total internal reflection
    tir = 1;        // signal tir
    //t = 2.0*n.dot(-d)*n+d; // return mirror
  }
  else {
    //ret = nnt*(d-n*ddn) - n*sqrt(costsq);
    //ret = nnt*d - ddn*nnt*n - sqrt(costsq)*n;
    ret = nnt*d - (ddn*nnt + sqrt(costsq))*n;

    // happens with
    // ./pane -a3 --environ=seminar.hdr open.iv -o test.png -r -d200x200 -s100
//     if (ret[0]>1.0 || ret[1]>1.0 || ret[2]>1.0)
//       cerr << "oh no!\n";
    
    tir = 0;
  }
}


//'''''''''''WW'''''''WW''WW''''WW''''''WW''''WW'''WWW''''''''''''''WW'''''''
//                    WW  WW            WW    WW  WW  W             WW       
//WWW   WW   WW WWW  WWWW WW    WW  WWW WWW  WWWW WW     WW   WWW  WWWW W WW 
//WW W WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW    WW W  WW W  WW  WW   
//WW W WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW  W WW W  WW W  WW  WW   
//WWW   WW   WW WW W  WW  WWWWW WW  WWW WW W  WW   WWW   WW   WW W  WW  WW   
//WW                                 WW                                      
//WW'''''''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''
//
// Calculates point lights contribution
//
SbColor Pane::pointLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng){
  
  SbColor plcolor = SbColor(0.0, 0.0, 0.0);
  float value;
  SbVec3f l;
  vector<SoPointLight *>::iterator pI = litPnts.begin();
  if (gv->headLightPos == HL_OFF)
    pI++;

  // the following is unnecessary because the caller has already
  // given us a nice upward facing normal (it had better!)
  // fix inward pointing normals
  // if (n.dot(ng)<0.0)
  //   n = -n;


  // debugging
  // if (drand48()<0.02){
  //   Ray(x,n).show(50.0, SbColor(1.0, 1.0, 0.0));
  // }
  
  for (; pI != litPnts.end(); pI++){

    // compute light direction, l
    l = (*pI)->location.getValue() - r.x;
    l.normalize();

    // dont use lights below horizon
    if (ng.dot(l) <= 0.0)
      continue;

    // don't illuminate areas that think they are facing away
    float dot1 = n.dot(l);
    if (dot1<=0.0)
      continue;
    
    // trace shadow ray
    Ray sr((*pI)->location.getValue(), -l, r.ctx);

    // does light ray reach triangle?
    if (! (tritree.intersect_ray(sr) && sr.tri == r.tri) )
      continue;

    // from jensen p.36 eq. 3.2 (frd = diffuse brdf, f_r,d)
    // L = frd*(w.n)/dist^2*Phi_l/(4pi)
    value = dot1 / gv->meters2(sr.t*sr.t) /
      (4.0*M_PI) * (*pI)->intensity.getValue();
    // (l.n)/dist^2 * 1/(4pi)

    // flip normal if surface normal is facing away from light
    //if (value < 0.0)
    //  value = -value;  // continue; // triangle must face light

    // compute color
    // from jensen p.36 eq. 3.2

    SbColor lcolor = (*pI)->color.getValue(); // in W
    plcolor += SbColor(            
      r.diffCol[0]*M_1_PI * value * lcolor[0],          // r.diffCol = reflectance
      r.diffCol[1]*M_1_PI * value * lcolor[1],
      r.diffCol[2]*M_1_PI * value * lcolor[2]);
    // frd * [ (l.n)/dist^2 * 1/(4pi) ] * Phi_l        note: frd = p/pi
  }
  return(plcolor);
}




//'''''''''''WW'''''''WW''WW''''WW''''''WW''''WW'''WWW''''''''''''''WW'''''''
//                    WW  WW            WW    WW  WW  W             WW       
//WWW   WW   WW WWW  WWWW WW    WW  WWW WWW  WWWW WW     WW   WWW  WWWW W WW 
//WW W WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW    WW W  WW W  WW  WW   
//WW W WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW  W WW W  WW W  WW  WW   
//WWW   WW   WW WW W  WW  WWWWW WW  WWW WW W  WW   WWW   WW   WW W  WW  WW   
//WW                                 WW                                      
//WW'''''''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''
//
// Calculates point lights contribution (does not require surface)
//
SbColor Pane::pointLightContrib2(Ray &r, SbVec3f &n, SbVec3f &ng){
  
  SbColor plcolor = SbColor(0.0, 0.0, 0.0);
  float value;
  SbVec3f l;
  vector<SoPointLight *>::iterator pI = litPnts.begin();
  if (gv->headLightPos == HL_OFF)
    pI++;

  for (; pI != litPnts.end(); pI++){

    // compute light direction, l
    l = (*pI)->location.getValue() - r.x;
    float dist = l.normalize();

    // dont use lights below horizon
    if (ng.dot(l) <= 0.0)
      continue;

    // don't illuminate areas that think they are facing away
    float dot1 = n.dot(l);
    if (dot1<=0.0)
      continue;
    
    // trace shadow ray
    //Ray sr((*pI)->location.getValue(), -l);
    // Ray sr(r.x+0.1*l, l, r.ctx); // shadow offset improves things
    Ray sr(r.x, l, r.ctx);
    //sr.show2(dist);

    // does light ray reach triangle?
    if (tritree.intersect_ray(sr, r.tri) && sr.t<dist)
      continue;

    sr.t = dist;

    // from jensen p.36 eq. 3.2 (frd = diffuse brdf, f_r,d)
    // L = frd*(w.n)/dist^2*Phi_l/(4pi)
    value = dot1 / gv->meters2(sr.t*sr.t) /
      (4.0*M_PI) * (*pI)->intensity.getValue();
    // (l.n)/dist^2 * 1/(4pi)

    // compute color
    // from jensen p.36 eq. 3.2

    SbColor lcolor = (*pI)->color.getValue(); // in W
    plcolor += SbColor(            
      r.diffCol[0]*M_1_PI * value * lcolor[0],          // r.diffCol = reflectance
      r.diffCol[1]*M_1_PI * value * lcolor[1],
      r.diffCol[2]*M_1_PI * value * lcolor[2]);
    // frd * [ (l.n)/dist^2 * 1/(4pi) ] * Phi_l        note: frd = p/pi
  }
  return(plcolor);
}




//'''''''''''WWW''WWWW''''''''WW'''''''WW''WW''''WW''''''WW''''WW'''WWW'''''''
//            WW  WW  W                WW  WW            WW    WW  WW  W      
//WW W  WW    WW  WW  W  WW   WW WWW  WWWW WW    WW  WWW WWW  WWWW WW     WW  
//WW W WW W   WW  WWWW  WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW    WW W 
//WW W WW W   WW  WW    WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW  W WW W 
// WW   WW   WWWW WW     WW   WW WW W  WW  WWWWW WW  WWW WW W  WW   WWW   WW  
//                                                    WW                      
//''''''''''''''''''''''''''''''''''''''''''''''''''WWW'''''''''''''''''''''''
// Experimental volume scattering code
//
SbColor Pane::volPointLightContrib(Ray &r, SbColor &c){

  // ray march
  float sigma_s = gv->sigma_s; //0.5;   // in meters!
  float sigma_a = 0.0; // -sigma_s
  float sigma_t = sigma_s+sigma_a;
  float p = 1.0/(4.0*M_PI);
  float dt = gv->stepsize;
  //dt = gv->units(dt); // convert meters to units
  float fdt = r.ctx->mydrand48()*dt;

  // iterate over length of ray
  // (from back to front)
  SbColor color = c;
  float t=r.t;
  if (t<0)
    t=1.0*getMaxDist();

  int first = 0;
  float edt;
  
  while (t>0){

    // random first length
    if (first++ == 0)
      edt=fdt;
    else 
      edt=dt;

    t -= edt;
    SbVec3f rx = r.pos+t*r.dir;

    // clip last (nearest) length
    if (t<0){
      edt = -t;
      rx = r.pos;
    }

    // convert to meters
    edt = gv->meters(edt);

    // find contribution of point lights
    SbColor plcolor = SbColor(0.0, 0.0, 0.0);
    float value;
    SbVec3f l;
    vector<SoPointLight *>::iterator pI = litPnts.begin();
    if (gv->headLightPos == HL_OFF)
      pI++;

    for (; pI != litPnts.end(); pI++){
      
      // compute light direction, l
      l = (*pI)->location.getValue() - rx;
      float dist = l.normalize();
      
      // dont use lights below horizon
//      if (ng.dot(l) <= 0.0)
//	continue;

      // don't illuminate areas that think they are facing away
//       float dot1 = n.dot(l);
//       if (dot1<=0.0)
// 	continue;
    
      // trace shadow ray
      //Ray sr((*pI)->location.getValue(), -l);
      // Ray sr(r.x+0.1*l, l, r.ctx); // shadow offset improves things
      Ray sr(rx, l, r.ctx);
      //sr.show2(dist);
      
      // does light ray reach triangle?
     if (tritree.intersect_ray(sr, r.tri) && sr.t<dist)
	continue;

      sr.t = dist;
      
      // from jensen p.36 eq. 3.2 (frd = diffuse brdf, f_r,d)
      // L = frd*(w.n)/dist^2*Phi_l/(4pi)
      value = 1.0 / gv->meters2(sr.t*sr.t) /
	(4.0*M_PI) * (*pI)->intensity.getValue();
      // (l.n)/dist^2 * 1/(4pi)
//      value *= exp(-sigma_t*gv->meters(sr.t)); // no absorbtion

      // compute color
      // from jensen p.36 eq. 3.2

      SbColor lcolor = (*pI)->color.getValue(); // in W
      plcolor += SbColor(            
      value * lcolor[0],          // r.diffCol = reflectance
      value * lcolor[1],
      value * lcolor[2]);
      // frd * [ (l.n)/dist^2 * 1/(4pi) ] * Phi_l        note: frd = p/pi
    }

    c = plcolor*p*sigma_s*edt + /*exp(-sigma_t*edt)* */c; // no absorbtion
  }

  color = c;
  return(c);
}


//'''''''''''''''''W''''m'''''W''''''''WW''''''''''''''''''m'W''''WW''WWW'
//                 W          W    W  W  W          W        W   W  W W  W
// WW  WW  WW   WW W    W  WW WW  WWW W     W  WW  WWW  WW W WW    W  W  W
//W W W   Wmmm W W W    W W W W W  W  W  W W W W W  W  W   W W W  W   W  W
// WW W    WW   WW WWWW W  WW W W  W   WW   W  W W  W  W   W WW  WWWW WWW 
//                          W                                             
//                        WW                                              
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Calculates area lights contribution in 2 Dimensions
//
SbColor Pane::areaLightContrib2D(Ray &r, SbVec3f n, SbVec3f ng){

  float value;
  double cA, cB, cC;
  SbVec3f l;
  SbVec3f lx;
    
  SbColor alcolor = SbColor(0.0, 0.0, 0.0);

  vector<Triangle *>::iterator aI = litTris.begin();

  // fix inward pointing normals
  if (n.dot(ng)<0.0)
    n = -n;

  // flatten normals
  n[1] = 0.0;
  n.normalize();
  ng[1] = 0.0;
  ng.normalize();

  // do nothing if no light samples
  if (gv->lightSamples<=0)
    return(alcolor); // black
  
  // loop over area lights
  for (; aI != litTris.end(); aI++){

    // intersect light with plane
    Triangle *tri = *aI;
    SbVec3f c[2];
    int crossing=0;

    // loop over vertices
    for (int i=0; i<3; i++){
      int j = (i+1)%3;
      if (crossing>2){
	cerr << "crossing error!\n";
	continue;
      }

      if ( (tri->vert[i][1]-r.x[1])*(tri->vert[j][1]-r.x[1]) < 0 ){
	float ratio = (r.x[1] - tri->vert[i][1]) /
	  (tri->vert[j][1] - tri->vert[i][1]);
	c[crossing][0] = ratio * (tri->vert[j][0] - tri->vert[i][0]) +
	                 tri->vert[i][0];
	c[crossing][1] = r.x[1];
	c[crossing][2] = ratio * (tri->vert[j][2] - tri->vert[i][2]) +
	                 tri->vert[i][2];
	crossing++;
      }
    }

    // debugging
    // if (crossing){
    //   showPoint(c[0], 3.0);
    //   showPoint(c[1], 3.0);
    //   showPoint(tri->vert[0], 1.0);
    //   showPoint(tri->vert[1], 1.0);
    //   showPoint(tri->vert[2], 1.0);
    // }

    // Loop over samples
    SbColor colora = SbColor(0.0, 0.0, 0.0);
    for (int lsamp=0; lsamp<gv->lightSamples; lsamp++){

      if (gv->areaLights){
	// choose random position inside light
	cA = r.ctx->mydrand48();
      }
      else
	cA = 0.5;

      // compute sample location
      lx = cA*c[0] + (1.0-cA)*c[1];

      // compute light direction, l
      l = lx - r.x;
      l[1] = 0.0;
      l.normalize();

      // dont illuminate lights below horizon
      if (ng.dot(l) <= 0.0)
        continue;
	
      // trace from point to light point
      Ray lr = Ray(r.x, l, r.ctx);
      if (! (tritree.intersect_ray(lr, r.tri) && lr.tri == *aI) ){
	//cerr << "lr.t=" << lr.t << "\n";
	continue; // does not hit light
      }

      // don't illuminate area that think they are facing away
      double dot1 = n.dot(l);
      if (dot1<0.0)
	continue;

      // calculate light normal scalar
      SbVec3f bc;
      tri->calcBaryCoords(lx, bc);
      SbVec3f np = (*aI)->getIntNormal(bc);
      np[1] = 0.0;
      np.normalize();
      double dot2 = -np.dot(l);
      if (dot2<0.0)
	continue;

      // compute value (fraction of light reaching surface)
      // from shirley p.120 eq. 15.5
      // Le = Phi-l/area/pi
      // L(x) = (emissive Le+) p/pi*Le(x')*cos(theta1)*area*cos(theta2)/dist^2
      value = dot1 * dot2 * gv->meters( (c[1]-c[0]).length() ) /
	gv->meters(lr.t);
      // (cos(theta)*area*cos(theta')/dist^2

      // from shirley p.124 middle of page
      SbColor Le = (*aI)->emission / // in W
	gv->meters2( (*aI)->area ) / 2.0;
      // Le = E(x') / pi = Phi_l/area / pi       note: Phi_l = power of light
		   
      // compute color
      // shirley p.120 eq. 15.5
      colora += SbColor(r.diffCol[R]/2.0 * Le[R] * value,
			r.diffCol[G]/2.0 * Le[G] * value,
			r.diffCol[B]/2.0 * Le[B] * value);
    } // end for each sample

    alcolor += colora/gv->lightSamples;
      
  } // end for each light

  return(alcolor);
}


//'''''''''''''''''W''''m'''''W''''''''WW''''''''''''''''''m'W''
//                 W          W    W  W  W          W        W  
// WW  WW  WW   WW W    W  WW WW  WWW W     W  WW  WWW  WW W WW 
//W W W   Wmmm W W W    W W W W W  W  W  W W W W W  W  W   W W W
// WW W    WW   WW WWWW W  WW W W  W   WW   W  W W  W  W   W WW 
//                          W                                   
//                        WW                                    
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Calculates area lights contribution
// Samples every light. Less efficient than
// than the importance sampling method (below).
//
SbColor Pane::areaLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng){

  float value;
  double cA, cB, cC;
  SbVec3f l;
  SbVec3f lx;
    
  SbColor alcolor = SbColor(0.0, 0.0, 0.0);

  vector<Triangle *>::iterator aI = litTris.begin();

  // no longer necessary - caller is responsible
  // for passing upward facing normal
  //
  // fix inward pointing normals
  // if (n.dot(ng)<0.0)
  //   n = -n;

  // do nothing if no light samples
  if (gv->lightSamples<=0)
    return(alcolor); // black

  // loop over area lights
  for (; aI != litTris.end(); aI++){

  SbColor colora = SbColor(0.0, 0.0, 0.0);
    for (int lsamp=0; lsamp<gv->lightSamples; lsamp++){

      if (gv->areaLights){
	// choose random position inside light
	cB = 1.0-sqrt(r.ctx->mydrand48());
	cC = (1.0-cB)*r.ctx->mydrand48();
	cA = 1.0-cB-cC;
      }
      else
	cB = cC = cA = 1.0/3.0;
	
      lx = cA*(*aI)->vert[0] + cB*(*aI)->vert[1] + cC*(*aI)->vert[2];

      // compute light direction, l
      l = lx - r.x;
      
      // 2d adjustment
      if (gv->useDimensions<3)
	l[1] = 0.0;
      
      l.normalize();
      //float lrt = l.normalize();

      // dont illuminate lights below horizon
      if (ng.dot(l) <= 0.0)
        continue;
	
      // trace from point to light point
      Ray lr = Ray(r.x, l, r.ctx);
      if ( (tritree.intersect_ray(lr, r.tri) && lr.tri == *aI)
	){ //|| lr.t > gv->tempFloat ){
	// hits light
	//lr.t = lrt;
      }
      else
	continue; // does not hit light

      // don't illuminate area that think they are facing away
      double dot1 = n.dot(l);
      if (dot1<0.0)
	continue;

      // calculate light normal scalar
      SbVec3f bc(cA, cB, cC);
      SbVec3f np = (*aI)->getIntNormal(bc);
      double dot2 = -np.dot(l);
      if (dot2<0.0)
	continue;

      // compute value (fraction of light reaching surface)
      // from shirley p.120 eq. 15.5
      // Le = Phi-l/area/pi
      // L(x) = (emissive Le+) p/pi*Le(x')*cos(theta1)*area*cos(theta2)/dist^2
      value = dot1 * dot2 * gv->meters2( (*aI)->area ) /
	gv->meters2(lr.t*lr.t);
      // (cos(theta)*area*cos(theta')/dist^2

      // from shirley p.124 middle of page
      SbColor Le = (*aI)->emission / // in W
	gv->meters2( (*aI)->area ) / M_PI;
      // Le = E(x') / pi = Phi_l/area / pi       note: Phi_l = power of light
		   
      // compute color
      // shirley p.120 eq. 15.5
      colora += SbColor(r.diffCol[R]*M_1_PI * Le[R] * value,
			r.diffCol[G]*M_1_PI * Le[G] * value,
			r.diffCol[B]*M_1_PI * Le[B] * value);
    } // end for each sample

    alcolor += colora/gv->lightSamples;
      
  } // end for each light

  return(alcolor);
}


//'''''''''''''''''WWW'''''''''''W''''m'''''W''''''''WW''''''''''''''''''m'W''
//                  W            W          W    W  W  W          W        W  
// WW  WW  WW   WW  W  WW W  WW  W    W  WW WW  WWW W     W  WW  WWW  WW W WW 
//W W W   Wmmm W W  W  W W W W W W    W W W W W  W  W  W W W W W  W  W   W W W
// WW W    WW   WW WWW W W W WW  WWWW W  WW W W  W   WW   W  W W  W  W   W WW 
//                           W            W                                   
//                           W          WW                                    
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Calculates area lights contribution
// Uses multiple luminaire importance sampling
// Discussed in Peter Shirleys Realistic Ray Tracing 1st ed. p. 124
// Basically, we simulate a discrete random variable
// to sample a probability mass function, where the masses
// are proportional to the importance (proximity and direction)
// of the lights.
//
SbColor Pane::areaImpLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng, int simple){

  if (gv->useDimensions==2)
    return(areaLightContrib2D(r, n, ng));
  
  SbColor color = SbColor(0.0, 0.0, 0.0);
  SbVec3f l, xP, xPP, nP;
  Triangle *lTri;

  // the following is unnecessary because the caller has already
  // given us a nice upward facing normal (it had better!)
  // fix inward pointing normals
  // if (n.dot(ng)<0.0)
  //   n = -n;

  int numLights = litTris.size();
  if (numLights == 0 || gv->lightSamples <= 0)
    return(color); // black

  double cA, cB, cC;
  double e_i, dot1, dot2, dot3;
  double lP, lP2, lPP;

  // initialize precomputation tables
  typedef struct cacheEntry {
    float sarea, alpha, cos_alpha, sin_alpha;
    float product; // sin-alpha*cos_c
    SbVec3f CgA;
  } cacheEntry_t;

//   cacheEntry_t cv[numLights];
//   unsigned int cvComputed[numLights];
  float *weights = (float *)alloca(sizeof(float)*numLights);
  int *sampleMethod = (int *)alloca(sizeof(int)*numLights);   // 0=approx, 1=acc
  cacheEntry_t *cv = (cacheEntry_t *)alloca(sizeof(cacheEntry_t)*numLights);
  unsigned int *cvComputed = (unsigned int *)alloca(sizeof(unsigned int)*numLights);
  
  /////////////////////////
  // compute weights[]
  /////////////////////////
  double eSum = 0;
  int lNum = 0;
  vector<Triangle *>::iterator aI = litTris.begin();
  for (; aI != litTris.end(); aI++,lNum++){
    
    // L*cosO*cosO'*A/dist^2
    
    cvComputed[lNum] = 0;

    // compute Lavg
    lTri = *aI;
    e_i = lTri->emisAvg/lTri->area;  // ~ L(i)

    // Figure purposes - only allow certain lights
    if (gv->figureMode && gv->useGlassModel==0 &&
	ABSF(e_i-0.15)>0.01)
      continue;
    
    // find normal
    nP = lTri->getNormal();

    //                      /\
    //                     /  \
    //                    /    \
    // 	     nP<------xP \
    //                  '---/----'
    // 		    /
    //           nPP      /
    //            ^\  p3 /l
    //              \ /\/        
    //               \ /\
    //              / xPP\
    //           p2'-/----p1
    //              x

    // find center
    xP = (lTri->vert[X]+lTri->vert[Y]+lTri->vert[Z]) * 1.0/3.0;

    // make rough estimate
    l = xP-r.x;
    lP = l.normalize();

    // calculate dot products and cos's
    float dp[3], dp1[3];
    
    xP = lTri->vert[X]-r.x;
    dp[X] = ng.dot(xP);
    dp1[X] = dp[X]/xP.length();

    xP = lTri->vert[Y]-r.x;
    dp[Y] = ng.dot(xP);
    dp1[Y] = dp[Y]/xP.length();

    xP = lTri->vert[Z]-r.x;
    dp[Z] = ng.dot(xP);
    dp1[Z] = dp[Z]/xP.length();

    // calculate average angle between normal and light
    int dpcnt = 0;
    dot1 = 0.0;
    if (dp1[X]>0.0){
      dot1 += dp1[X];
      dpcnt++;
    }
    if (dp1[Y]>0.0){
      dot1 += dp1[Y];
      dpcnt++;
    }
    if (dp1[Z]>0.0){
      dot1 += dp1[Z];
      dpcnt++;
    }
    dot1 /= dpcnt;  // can be delayed, for speed

    // check luminaire is above horizon
    //if (dot1 <= 0.0){
    if (dpcnt == 0){
      weights[lNum] = 0.0;
      continue;
    }
      
    // calculate angle of light normal and light vector
    dot2 = -nP.dot(l);
    
    // check to see if luminaire pointing away
    if (dot2 <= 0.0){
      weights[lNum] = 0.0;
      continue;
    }

    // calculate projected solid angle (max pi, area of unit circle)
    float w1 = lTri->area / (lP*lP); // max projected area
    float w2 = dot1 * dot2;          // approximate projected area scalar components

    // if relatively far away (max solid angle is small)
    // or miniscule solid angle -> approximate sampling
    // if partially below horizon and not very small
    // otherwise -> accurate

    // weight
    //
    //
    // sampling
    // approximate if w_ < 0.001 || w1 < pi/20
    // acccurate if w_ > 0.001

    // find length of longest edge
//     float maxlen = 0.0, len2[3];
//     SbVec3f edge[3];
//     edge[0] = lTri->vert[1]-lTri->vert[0];
//     edge[1] = lTri->vert[2]-lTri->vert[0];
//     edge[2] = lTri->vert[2]-lTri->vert[1];
//     len2[0] = edge[0].dot(edge[0]);
//     len2[1] = edge[1].dot(edge[1]);
//     len2[2] = edge[2].dot(edge[2]);
//     maxlen = sqrt(MAX3(len2[0], len2[1], len2[2]));

    
    // if maximum solid angle is very small,
    // or if at extreme angle, use approximate weighting & sampling 
    if (w1 < 0.001 || //(dpcnt == 3 && w1 < M_PI/40.0) ||
	w2<0.001){
    //if (r.x[X]>50.0){
      
      // use approximate sampling
      sampleMethod[lNum] = 0;

      e_i *= w1 * w2;

      //return(SbColor(1,0,0));
    }
    else {    
      //return(SbColor(0,1,0));

      // use accurate sampling
      sampleMethod[lNum] = 1;
      
      // compute projected solid area using Stoke's thm
      // from Improving Radiosity Solutions
      // through the Use of Analytically Determined
      // Form Factors by Baum, Rushmeier, and Winget

      SbVec3f pts[4];
      int ptscnt = 0;

      // temporary bug fix
      //weights[lNum] = 1.0;
      //eSum += 1.0;
      //continue;
    
      // clip triangle against horizon
      for (int i=0; i<3; i++){
	int j = (i+1)%3;
	if (dp[i]>=0.0)      // check vertex (don't clip pts on horizon)
	  pts[ptscnt++] = lTri->vert[i];
	if (dp[i]*dp[j]<0.0) // clip crossing edge against horizon
	  pts[ptscnt++] = (lTri->vert[i]*(-dp[j]) + lTri->vert[j]*dp[i])/(dp[i]-dp[j]);
      }

      // guaranteed to be 3-4 points
      // integrate over edges
      float projarea = 0.0;
      for (int i=0; i<ptscnt; i++){
	int j = (i+1)%ptscnt;
	SbVec3f r1 = pts[i]-r.x;
	SbVec3f r2 = pts[j]-r.x;
	SbVec3f tau = r1.cross(r2);
	r1.normalize();
	r2.normalize();
	float dotp = r1.dot(r2);
	float gamma = acos(dotp);
	if (!finite(gamma)){
	  if (dotp>=1.0)
	    gamma = 0.0;
	  else
	    gamma = M_PI;
	}
      
	tau.normalize();
	tau *= gamma;
	projarea -= ng.dot(tau);
	
	if (!finite(projarea))
	  cerr << "problem\n";
      }

      //projarea /= 2.0*M_PI;
      e_i *= projarea*0.5;       // 0.5 why?? i dunno

      //cerr << "\nApprox proj area= " << w1 * w2 << "\n";
      //cerr << "Accura proj area= " << projarea << "\n";

    
      // debugging
      //if (gv->tempInt == 39)
      //  e_i = 1;

    }
    
    weights[lNum] = e_i;
    eSum += e_i;
  }

  // debugging
  //if (gv->tempInt == 40)
  //  return(SbColor(eSum, eSum, eSum));
  
  if (eSum == 0.0)
    return(color); // black
  
  for (lNum=0; lNum<numLights; lNum++){
    weights[lNum] /= eSum;
    //if (DEBUG_STATE)
    //  cerr << "weight for light " << lNum << " = " << weights[lNum] << "\n";
  }

  // debugging
  if (gv->tempInt == 41)
    return(SbColor(weights[0], weights[1], 0.0));
  
  /////////////////////////////////
  // importance sample the lights
  /////////////////////////////////

  // stratified sampling
  int sqrt_samples = (int)sqrt((float)gv->lightSamples);
  
  // use one sample in final gather
  if (simple)
    sqrt_samples = 1;

  for (int sampx=0; sampx<sqrt_samples; sampx++){
    for (int sampy=0; sampy<sqrt_samples; sampy++){

      // jittered
      double eps1 = (sampx + r.ctx->mydrand48()) / sqrt_samples;
      double eps2 = (sampy + r.ctx->mydrand48()) / sqrt_samples;
      //double eps1 = drand48(), eps2 = drand48();

      // pick light
      eSum = 0.0;
      for (lNum=0; lNum<numLights-1; lNum++){
	if (eps1<(eSum+weights[lNum]))
	  break;
	else
	  eSum += weights[lNum];
      }

      //if (DEBUG_STATE)
      //cerr << "chose light " << lNum << endl;
      eps1 = (eps1 - eSum) / weights[lNum];

      lTri = litTris[lNum];

      // Generate sample on solid angle of triangle
      // from "Stratified Sampling of Spherical Triangles"
      // by James Arvo
      // Code mostly glibbed from Java version

      // this would be better if it picked
      // a point on the CLIPPED luminaire
      // oh well, future work
           
      // 1. find p1, p2, p3, points on unit sphere surrounding x
      SbVec3f p1, p2, p3;
      p1 = lTri->vert[0]-r.x;
      p1.normalize();
      p2 = lTri->vert[1]-r.x;
      p2.normalize();
      p3 = lTri->vert[2]-r.x;
      p3.normalize();

      if (sampleMethod[lNum] == 1){

	//return(SbColor(0,1,0));

	// Only compute certain values once and store in
	// a cache. Then use cache values in future computations.
	if (cvComputed[lNum] == 0){
	
	  cvComputed[lNum] = 1;

	  // Compute dihedral angles
	  SbVec3f u12, u13, u23;
	  u12 = p1.cross(p2); u12.normalize();
	  u13 = p1.cross(p3); u13.normalize();
	  u23 = p2.cross(p3); u23.normalize();

// 	cv[lNum].cos_alpha = (-u13).dot(-u12); // p3xp1 . p2xp1
// 	float cos_beta  =    u12.dot(-u23);    // p1xp2 . p3xp2
// 	float cos_gamma =    u13.dot( u23);    // p1xp3 . p2xp3
	  cv[lNum].cos_alpha =   u13.dot(u12);  // p3xp1 . p2xp1
	  float cos_beta  =    -(u12.dot(u23)); // p1xp2 . p3xp2
	  float cos_gamma =      u13.dot(u23);  // p1xp3 . p2xp3

	  cv[lNum].alpha = acos(cv[lNum].cos_alpha);
	  float beta  = acos(cos_beta);
	  float gamma = acos(cos_gamma);

	  // Compute solid angle (area of tri on unit sphere)
// 	if (DEBUG_STATE)
// 	  cerr << "cos_alpha=" << cv[lNum].cos_alpha
// 	       << " alpha=" << cv[lNum].alpha
// 	       << " sarea=" << cv[lNum].sarea << "\n";
	  
	  cv[lNum].sarea = cv[lNum].alpha+beta+gamma - M_PI;

	  // debugging
	  if (!finite(cv[lNum].sarea))
	    cerr << "problem\n";
	
	  // compute other cache values
	  cv[lNum].sin_alpha = sin(cv[lNum].alpha);
	  float cos_c = p1.dot(p2);
	  cv[lNum].product = cv[lNum].sin_alpha*cos_c;
	  cv[lNum].CgA = p3-p1*(p3.dot(p1)/(p1.dot(p1)));
	  cv[lNum].CgA.normalize();
	}

	// temporary bug fix 
	if (cv[lNum].sarea <= 1.0e-6)
	  continue;

	// find sample 
      
	float phi = eps1*cv[lNum].sarea - cv[lNum].alpha;
	float s = sin(phi);
	float t = cos(phi);
	float u = t-cv[lNum].cos_alpha;

	float v = s+cv[lNum].product;
	float q =
	  (cv[lNum].cos_alpha * (v*t-u*s) - v) /
	  (cv[lNum].sin_alpha * (u*t+v*s)    );
	double temp = MAX2(1.0-q*q, 0.0);

	SbVec3f T = q*p1 + sqrt(temp)*cv[lNum].CgA;
	T.normalize();
	float z = 1.0-eps2 * (1.0-T.dot(p2));
	T -= p2 * (T.dot(p2)/p2.dot(p2));
	T.normalize();
	l = z*p2 + sqrt(1.0-z*z)*T;
	//l.normalize();

      }
      else {

	//return(SbColor(1,0,0));

	// choose sample point using flat triangle

	// 1. find p1, p2, p3, points on unit sphere surrounding x
	p1 += r.x;
	p2 += r.x;
	p3 += r.x;

	cB = 1.0-sqrt(1.0-eps2);
	cC = (1.0-cB)*eps1;
	cA = 1.0-cB-cC;

	xPP = cA*p1 + cB*p2 + cC*p3;
    
	// 4. Find light direction, l
	l = xPP - r.x;

      }
      
      // 2d adjustment
      if (gv->useDimensions<3)
	l[1] = 0.0;
    
      lPP = l.normalize();

      // dont illuminate lights below horizon
      if (ng.dot(l) <= 0.0){
	continue;
      }

      // don't illuminate area that think they are facing away
      dot1 = n.dot(l);
      if (dot1<=0.0)
	continue;      

      // trace from point to light point
      // i add a small amount to make it not intersect overlapping surfaces
      // i really hate doing this
      // optimize: fix this vv^^
      // fix: could solve with min_t
      const float t_eps = 0.002;
      Ray lr = Ray(r.x+l*t_eps, l, r.ctx);
      // Ray lr = Ray(r.x+l*0.0002, l, r.ctx);
      if (! tritree.intersect_ray(lr, r.tri))
 	continue; // does not anything (?)
      if (lr.tri != lTri)
	continue; // does not hit light

      double value;
      
      if (sampleMethod[lNum] == 1){

	value = dot1 * cv[lNum].sarea;
      }
      else {

	// calculate edges and normal
	SbVec3f edge1 = p2-p1;
	SbVec3f edge2 = p3-p1;
	SbVec3f nPP = edge1.cross(edge2);

	double area = 0.5*nPP.normalize();

	nP = lTri->getNormal();
	dot2 = -nP.dot(l);
	
	// FIXME: test seems needless
	double dot3 = -nPP.dot(l);
	if (dot3<0.0)
	  dot3 = -dot3;
    
	// 6. Find p(x) =                  ||x-x"||^2 * cosO'
	//                 -----------------------------------
	//                  area(p1p2p3) * ||x-x'||^2 * cosO"

	lP = lr.t+t_eps;	
	double prob = (gv->meters2(lPP*lPP) * dot2) /
	  (gv->meters2(area) * gv->meters2(lP*lP) * dot3);

	value = dot1 * dot2 / gv->meters2(lP*lP) / prob;
      }

      SbColor Le = lTri->emission / // in W
	gv->meters2( lTri->area ) * M_1_PI;
      // Le = E(x') / pi = Phi_l/area / pi       note: Phi_l = power of light

      // weight light
      value /= weights[lNum];

      // debugging
      if (!finite(value))
	cerr << "Problem in areaLightImpContrib: non-finite value\n";
      
      // compute color
      // shirley p.120 eq. 15.5
      color += SbColor(r.diffCol[R] * M_1_PI * Le[R] * value,
		       r.diffCol[G] * M_1_PI * Le[G] * value,
		       r.diffCol[B] * M_1_PI * Le[B] * value) ;
      
      if (DEBUG_STATE){
	cerr << flush;
	print(" value: %f ", value);
	printCol(color);
	cerr << endl;
      }
      
    } // end stratified sampling x
  } // end stratified sampling x

  color /= sqrt_samples*sqrt_samples;
  
  return(color);
}



//'''''''''''''''''''''''''''''''''''''''''''''WW'''''''WW''''WW'WWWW'''WWW''
//                                             WW       WWW  WWW WW  W WW  W 
//WWW  W WW  WW   WWW  WW   WWWW W  WWW  W WW WWWW  WW  W WWW WW WW  W WW    
//WW W WW   WWWW WW W WW W  WW WW W WW W W WW  WW  WWWW W  W  WW WW  W WW    
//WW W WW   WW   WW   WW W  WW WW W WW W W WW  WW  WW   W     WW WW  W WW  W 
//WWW  WW    WWW  WWW  WW   WW WW W WWW   WWW  WW   WWW W     WW WWWW   WWW  
//WW                                WW                                       
//WW''''''''''''''''''''''''''''''''WW'''''''''''''''''''''''''''''''''''''''
//
// precompute look up tables
//
void Pane::precomputeMDCTables(){

  // Calculate M and N to match a user specified total # of gather rays
  // Mmdc = MIN2((int)sqrt(gv->gatherRays/2.0), Mmax);
  // Nmdc = 2*Mmdc;
  Mmdc = MIN2((int)sqrt(gv->gatherRays/3.0), Mmax);
  Nmdc = 3*Mmdc;
  
  // theta, M, j
  for (int j=0; j<Mmdc; j++){
    float sinTm = sqrt( j/(double)Mmdc );
    float sinTp = sqrt( (j+1)/(double)Mmdc );
    float thetam = asin(sinTm);
    float thetap = asin(sinTp);
    float cosTm = cos(thetam);

    // sin(Oj_)*cos^2(Oj_)
    // sin(O_) = sinTm
    sinTxcos2T[j] = sinTm*cosTm*cosTm;

    // sin(Oj+)-sin(Oj_)
    sinTmsinT[j] = sinTp - sinTm;
  }

  // phi, N, k
  float phi;
  for (int k=0; k<Nmdc; k++){
    // phi_k_ + pi/2
    phi = 2.0 * M_PI * k/(float)Nmdc + M_PI/2.0;
    vk_X[k] = cos(phi);
    vk_Z[k] = sin(phi);
  }
	
}


//'''''''''''''WWW'''WW''WW''''''WWW'''''''WWWW''WW'''WW'''WW'''''''''''''''''
//              WW   WW           WW       WW  W     WW   WW                  
//WWWW W  W WW  WW  WWWW WW WWW   WW   WW  WW  W WW WWWW WWWW W WW  WWWW  WW  
//WW WW W W WW  WW   WW  WW WW W  WW  WWWW WW  W WW  WW   WW  W WW WWW   WWWW 
//WW WW W W WW  WW   WW  WW WW W  WW  WW   WW  W WW  WW   WW  W WW   WWW WW   
//WW WW W  WWW WWWW  WW  WW WWW  WWWW  WWW WWWW  WW  WW   WW   WWW WWWW   WWW 
//                          WW                                                
//''''''''''''''''''''''''''WW''''''''''''''''''''''''''''''''''''''''''''''''
// (mdc)
//
// calculates the contribution from multiple diffuse reflections
// (indirect illumination). Uses Monte Carlo integration
// and the Photon Mapping radiance estimate to estimate
// indirect illumination (no lights). The irradiance is then
// multiplied by the BRDF to return the reflected radiance.
//
// Uses Irradiance Caching with Irradiance Gradients to cache and
// interpolate the results, since this is an expensive computation.
//
SbColor Pane::multipleDiffuseContrib(Ray &r, SbVec3f &n, SbVec3f &ngl,
				     int simpleCalc,
				     int prepass,
				     int inOrOut){

  // already done
  // fix inward pointing normals
  //if (n.dot(ngl)<0.0)
  //  n = -n;

//   int useIrradCache =
//     gv->useIndirectImpSampling == 0 &&
//     gv->irradCacheError > 0.0 &&
//     gv->bumpMapping == 0;
  
  // Cached value exist?
  SbColor ptcolor = SbColor(0.0, 0.0, 0.0);
  
  int cache_result = 0;
  if (gv->useIrradCache)
    cache_result = irradCache.getIrradiance(&ptcolor, r, &n, &ngl, prepass);

  if (cache_result == 0){

    const int n_theta = 4;
    const int n_phi = 16;
    float *flux_map;

    // flux map section disabled
    if (gv->useIndirectImpSampling){
      
      flux_map = (float *)alloca(sizeof(float)*n_theta*n_phi);
      // get flux map for importance sampling
      const int n_photons = gv->gPhotonEstimate;


      gPhotonMap->get_flux_map((float *)flux_map, n_theta, n_phi, r.x.getValue(),
			       n.getValue(), n_photons);

      // find zero_regions and total_energy
      int zero_regions = 0;
      float total_energy = 0.0;
      for (int u=0; u<n_theta; u++)
	for (int v=0; v<n_phi; v++){
	  if (flux_map[u*n_phi+v] > 0.0)
	    total_energy += flux_map[u*n_phi+v];
	  else
	    zero_regions++;
//	cerr << "flux for (" << u << "," << v << ") = " << flux_map[u][v]
//	     << "\n";
	}

      // add small fraction of energy to zero regions
//  cerr << "te = " << total_energy;
      float de;
      if (total_energy>0.0)
	de = total_energy / (n_phi*n_theta); // * 0.5;
      else
	de = 1.0;

//    cerr << " de = " << de;
//    cerr << " zr = " << zero_regions;
      total_energy += de*(n_phi*n_theta);
//    cerr << " te = " << total_energy << "\n";

      // scale flux map to create c.d.f.
      for (int u=0; u<n_theta; u++)
	for (int v=0; v<n_phi; v++){
//	if (flux_map[u][v] > 0.0)

	  // directional
	  flux_map[u*n_phi+v] += de;
	  flux_map[u*n_phi+v] /= total_energy;

	  // stratified
//	flux_map[u][v] = 1.0/(n_phi*n_theta);
	
//        else
//	  flux_map[u][v] = de/total_energy;
	  //if (flux_map[u][v])
//         cerr << "weight for (" << u << "," << v << ") = " << flux_map[u][v]
//	  << "\n";
	}

      // draw debugging histogram
      if (DEBUG_STATE)
	drawHist((float *) flux_map, n_phi, n_theta,
		 n_phi*n_theta*getMaxDist()/10.0, r.x, n);

//      return(ptcolor);
    }

    
    // integrate hemisphere
    SbRotation sbr;
    float phi, theta;
    SbVec3f neww;
    sbr.setValue(SbVec3f(0.0, 1.0, 0.0),
		 ngl);

    // stratified sampling
    ptcolor = SbColor(0.0, 0.0, 0.0);

    // number of stratifications in theta, and phi
    int M, N;
    if (simpleCalc){
      if (r.depth<=2){
	M=3;
	N=4;
      }
      else {
	M = N = 1;
      }
    }
    else {
      M=Mmdc;
      N=Nmdc;
    }

    // 2d adjustment
    if (gv->useDimensions<3){
      M=MIN2(gv->gatherRays, Mmax);
      N=1;
      if (simpleCalc)
	M=4;
    }
      
    // Ray bounce = Ray(r.x, ngl, r.ctx, 1);
    Ray bounce = Ray(r.x, ngl, r.ctx, r.depth+1, r.atten);
    float harm_mean = 0.0;
    SbVec3f Er = SbVec3f(0.0, 0.0, 0.0);
    SbVec3f Et = SbVec3f(0.0, 0.0, 0.0);
    SbVec3f v_k;
    float *Ljk   = (float *)alloca(sizeof(float)*N*M);
    float *rjk   = (float *)alloca(sizeof(float)*N*M);
    float *phijk = (float *)alloca(sizeof(float)*N*M);
    int rejectPoint = 0; // for trashing (not storing) statistically bad irrpnt

    // debugging
    float maxrad = -1.0;
    int maxk=-1, maxj=-1;
    
    // loop over phi
    for (int k=0; k<N; k++){

      float rot_sum = 0.0;

      // loop over theta
      for (int j=0; j<M; j++){

	// choose new direction, neww (cosine weighted)
	//double eps1 = 0.5, eps2 = 0.5;
	//double eps1 = ctx->mydrand48(), eps2 = ctx->mydrand48();
	//phi = 2.0 * M_PI * (k+eps1)/N;
	//theta = asin(sqrt( (j+eps2)/M ));

	// search for new direction using importance sampling
	double eps1 = (j + r.ctx->mydrand48()) / M;  // stratified
	double eps2 = (k + r.ctx->mydrand48()) / N;  // stratified
//	double eps1 = (j+0.5)/M;
//	double eps2 = (k+0.5)/N;

	float imp_scale;
	
	if (gv->useIndirectImpSampling){

	  // use eps1 to pick region
	  double eSum = 0.0;
	  int u, v;
	  int found = 0;
	  for (u=0; u<n_theta; u++){
	    for (v=0; v<n_phi; v++)
	      if (eps2<(eSum+flux_map[u*n_phi+v])){
		found = 1;
		break;
	      }
	      else
		eSum += flux_map[u*n_phi+v];

	    if (found)
	      break;
	  }

	  // check if reached end (shouldn't, but might...)
	  if (!found){
	    u = n_theta-1;
	    v = n_phi-1;
	  }

	  //cerr << "chose region " << u << "," << v << endl;

	  // compute scaling factor
	  imp_scale = 1.0/(flux_map[u*n_phi+v]*n_theta*n_phi);

	  // adjust eps
	  eps2 = (eps2 - eSum) / flux_map[u*n_phi+v];

	  // calculate phi and theta from eps1 and eps2
	  // this uses a non-linear transformation to give
	  // a cosine distrubition in theta, that is
	  // theta takes [0,1]->[0,pi/2], and if
	  // [0,1] is uniform, theta is ~ cos(theta).
	  // the actual mapping is [0,1]->[pi/2,0].
	  // i reversed it to save a 1-eps subtraction operation.

	  // importance
	  theta = acos(sqrt( (u+eps1)/n_theta ));
	  phi = 2.0 * M_PI * (v+eps2)/n_phi;

	}
	else {
	
	  // stratified
	  theta = acos(sqrt( eps1 ));
	  phi = 2.0 * M_PI * eps2;
	  imp_scale = 1.0;
	}
	
	neww.setValue(sin(theta)*cos(phi),
		      cos(theta),
		      sin(theta)*sin(phi));
    
	sbr.multVec(neww, bounce.dir);
	
	// 2d adjustment
	if (gv->useDimensions<3){
	  SbVec3f fngl = SbVec3f(ngl[0], 0.0, ngl[2]);
	  fngl.normalize();
	  SbVec3f fngl_perp = SbVec3f(fngl[2], 0.0, -fngl[0]);
	  theta = asin(2.0*eps1-1.0);
	  bounce.dir = fngl*cos(theta) + fngl_perp*sin(theta);
	  // if (DEBUG_STATE && !simpleCalc)
	  // bounce.show(200, SbColor(1.0, 1.0, 1.0));
	}
	
	bounce.dir.normalize();

	SbVec3f Lrad;
	int useApprox = 0; //(simpleCalc)?1:0;

	// debugging
	//int state = DEBUG_STATE;
	//DEBUG_STATE = 0;
	Lrad = getRadiance1(bounce,
			 C_SIMPLE | C_SPECULAR | (useApprox?C_APPROX:0) |
			 C_DIRECT | C_CAUSTIC | C_MDIFFUSE,
			 r.tri);
	//DEBUG_STATE = state;

	// can uncoment to create sort of a dome light with worldBG.. buggy though
	//if (bounce.t<=0.0)
	//  bounce.t = getMaxDist();

	// if hit something...
	if (bounce.t>0.0){

	  // bump map
	  float bumpVal = 1.0;
	  if (gv->bumpMapping && simpleCalc == 0){
	    bumpVal = bounce.dir.dot(n)/bounce.dir.dot(ngl);
// 	    if (bumpVal>gv->bumpScale)
// 	      bumpVal = gv->bumpScale;
	    if (bumpVal<0.0)
	      continue;
	  }
	  
	  // don't do irradiance_estimate on non-diffuse surfaces - pointless&slow!!
	  ptcolor += Lrad*imp_scale*bumpVal;

	  // continue with phi iterations if simple calculation
	  if (simpleCalc || gv->useIrradCache == 0)
	    continue;
	  
	  // update harmonic mean
	  harm_mean += 1.0/bounce.t;
	  
	  // for rotational gradiant calculation
	  //float l_avg = (Lrad[X]+Lrad[Y]+Lrad[Z])/3.0;
	  float l_avg = 0.27*Lrad[X]+0.67*Lrad[Y]+0.06*Lrad[Z];
	  rot_sum += - tan(theta)*l_avg;	  

	  // Reject ridiculously close to pi/2 values of theta.
	  // The following line of code rejects points if a sample
	  // is used, satisfying the condition:
	  // Prob(x>sample)*num_samples < 1/500.
	  // In other words, if the probability of having a
	  // sample closer to Pi/2 in this set is less than 0.2%,
	  // junk the whole set.
	  // This effectively skips over the craziest 0.2%.
	  // The reasoning behind this is, if
	  // Prob(x) = cos(x), then
	  // Prob(x>sample) = 1-sin(sample)  (by integrating pdf)
	  // If #samples*above < 1/500, then this sample is too close.
	  // Close samples make tan() (see previous line of code)explode.
	  if ((1.-1./(500*M*N))<sin(theta)){
	    rejectPoint = 1;

	    // record keeping
	    pthread_mutex_lock(&mutexes[LOCK_ICREJECTS]);
	    rejectedICPoints++;
	    pthread_mutex_unlock(&mutexes[LOCK_ICREJECTS]);
	  }

	  // for translational gradient
	  // [j][k] = N*k+j
	  Ljk[k*M+j] = l_avg;
	  rjk[k*M+j] = bounce.t;
	  phijk[k*M+j] = phi;

	  // debugging
	  if (0 && DEBUG_STATE){
	    //bounce.show(hit_t, Lrad*0.3);
	    //bounce.show(getMaxDist()/10.0, Lrad);
	    cerr << bounce.t << endl;
	  }
        }

	// ray did not hit geometry
	else { 

	  // for translational gradient
	  Ljk[k*M+j] = 0;
	  rjk[k*M+j] = 1e8;
	  phijk[k*M+j] = phi;

	  // show missed ray
	  if (0 &&DEBUG_STATE){ 
	    SbColor Lrad = SbColor(0,0,0);
	    bounce.show(getMaxDist()/10.0, Lrad);
	    cerr << "missed: " << "??" << endl;
	  }
	}
      
      } // end theta loop

      // skip IC stuff
      if (simpleCalc || gv->useIrradCache == 0)
	continue;
      
      // calculate v_k (theta=pi/2 phi=phi+pi/2)
      double phi_kp2 = phi+M_PI/2.0;
      v_k = SbVec3f(cos(phi_kp2), 0.0, sin(phi_kp2));
	    
      // update rotational gradient
      Er += v_k*rot_sum;

      if (0&&DEBUG_STATE){
	cerr << "Er += " << rot_sum << " * "; printVec(v_k);
	cerr << " = "; printVec(Er); cerr << "\n";
      }
	
    } // end phi loop

    // Calculate irradiance (and gradients)
    double factor = M_PI/(M*N);

    // 2d adjustment
    if (gv->useDimensions<3)
      factor = 2.0/(M*N);
    
    ptcolor *= factor;

    // skip the gradient calulcation stuff if simple calculation
    if (gv->useIrradCache && simpleCalc == 0){

      sbr.multVec(Er, Er);
      Er *= factor;

      // Calculate translational gradient
      SbVec3f u_k, v_k_;
      for (int k=0; k<N; k++){

	SbVec3f tran_sum1 = SbVec3f(0.0, 0.0, 0.0);

	// compute right sum for j = 0
	int km1 = (k==0)?N-1:k-1;
	double tran_sum2 =
	  sinTmsinT[0] /
	  MIN2(rjk[k*M], rjk[km1*M]) *
	  (Ljk[k*M] - Ljk[km1*M]);

	// perform inner sum
	for (int j=1; j<M; j++){

	  // left sum
	  u_k.setValue(cos(phijk[k*M+j]), 0.0, sin(phijk[k*M+j]));
	  sbr.multVec(u_k, u_k);
	  //u_k.normalize();        // accurate computation not needed
	  
	  int jm1 = j-1;
	  double s = sinTxcos2T[j] *
	    (Ljk[k*M+j] - Ljk[k*M+jm1]) /
	    MIN2(rjk[k*M+j], rjk[k*M+jm1]);
	  tran_sum1 += u_k*s;

	  // right sum
	  tran_sum2 +=
	    sinTmsinT[j] /
	    MIN2(rjk[k*M+j], rjk[km1*M+j]) *
	    (Ljk[k*M+j] - Ljk[km1*M+j]);
	}

	// finish inner sum
	
	// compute v_k_
	v_k_.setValue(vk_X[k], 0.0, vk_Z[k]);
	sbr.multVec(v_k_, v_k_);
	//v_k_.normalize();      // accurate computation not needed

	// add to final
	Et += (2.0*M_PI/N)*tran_sum1 + v_k_*tran_sum2;

      }
      
      // Calculate harmonic mean
      // 1/harmonic mean = 1/n * Sum[1/x] => H = n/Sum[1/x]
      if (harm_mean >0.0)
	harm_mean = (M*N) / harm_mean;
      else // if no objects hit, set mean to bounding box size
	harm_mean = irradCache.size();

      if (!rejectPoint){

	// store irradiance
	if (DEBUG_STATE){
	  cerr << "STORING: r=" << harm_mean << " E="; printCol(ptcolor);
	  cerr << " Et="; printVec(Et);
	  cerr << " Er="; printVec(Er);
	  cerr << " n="; printVec(n);
	  cerr << " x="; printVec(r.x);
	  cerr << "\n";
	}

	// 2d adjustment
	if (gv->useDimensions<3)
	  Er = Et = SbVec3f(0.0, 0.0, 0.0);

	// store point
	pthread_mutex_lock(&mutexes[LOCK_ICACHE]);
	irradCache.storeIrradiance(ptcolor, r.x, n, ngl, harm_mean, Er, Et);
	pthread_mutex_unlock(&mutexes[LOCK_ICACHE]);    

	// draw point
	if (gv->renderQuit == 0)
	  showPhoton(r.x, SbColor(0.0, 1.0, 0.0));
      }
    }

  }
      
  // multiply irradiance by brdf to get radiance
  if (gv->useDimensions<3){
    ptcolor[R] *= r.diffCol[R]*1.0/2.0;
    ptcolor[G] *= r.diffCol[G]*1.0/2.0;
    ptcolor[B] *= r.diffCol[B]*1.0/2.0;
  }
  else {
    ptcolor[R] *= r.diffCol[R]*M_1_PI;
    ptcolor[G] *= r.diffCol[G]*M_1_PI;
    ptcolor[B] *= r.diffCol[B]*M_1_PI;
  }

  // return radiance
  return(ptcolor);
}


//''''''''''''WW'''''''''''''''''WW''WW'''''''''''''WWW''''''''''''''WW'''''''
//           WW                  WW                WW  W             WW       
//W WW  WW  WWWW W WW  WWW  WWW WWWW WW  WW   WWW  WW     WW   WWW  WWWW W WW 
//WW   WWWW  WW  WW   W WW WW W  WW  WW WW W  WW W WW    WW W  WW W  WW  WW   
//WW   WW    WW  WW   W WW WW    WW  WW WW W  WW W WW  W WW W  WW W  WW  WW   
//WW    WWW  WW  WW    WWW  WWW  WW  WW  WW   WW W  WWW   WW   WW W  WW  WW   
//                                                                            
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Calculates refracted component contribution
// Returns radiance due to transparent dielectric,
// specifically the specularly reflected and refracted components.
//
SbColor Pane::refractionContrib(Ray &r, SbVec3f &n, SbVec3f &ngl, int flags){
  
  SbColor ncolor;
  SbVec3f tdir;
  double c;
  SbVec3f k;
  double R0;
  double Re;
  
  SbVec3f rdir = r.dir-2.0*n.dot(r.dir)*n;
  
  // grazing angle problem
  // don't let reflected ray go beneath surface
  if (rdir.dot(ngl)<0.0){
    rdir += (-rdir.dot(ngl)+0.0015) * ngl;
    rdir.normalize();
  }
  Ray reflect(r.x, rdir, r.ctx, r.depth+1, r.atten);

  // From RRT by PS
  double nc, nt;
  float transparency = r.tri->transparency;
//  transparency = 0.5;
  //double ar, ag, ab;
  //ar = ag = ab = log(transparency)*gv->meters(1.0); // convert a/m -> a/u
  int tir;

  if (n.dot(ngl)>0.0){  // into object
    nc = N_AIR;         // index of refraction for vacuum/air
    nt = gv->refractionIndex;       // index of refraction for glass
    refract(tdir, tir, r.dir, n, nc/nt);
    k = SbVec3f(1.0, 1.0, 1.0);
    c = -r.dir.dot(n);
  }
  else {                // out of object
    nc = gv->refractionIndex;
    nt = N_AIR;
    refract(tdir, tir, r.dir, -n, nc/nt);
    float t = gv->meters(r.t);
    k = SbVec3f(1.0, 1.0, 1.0);
//     k[R] = exp(gv->attenuation[R]*t);
//     k[G] = exp(gv->attenuation[G]*t);
//     k[B] = exp(gv->attenuation[B]*t);
    //k = SbVec3f(exp(ar*r.t), exp(ag*r.t), exp(ab*r.t));
    c = tdir.dot(n);
  }

  // total internal reflection
  if (tir){
    SbColor rcolor = (*this.*getRadiance)(reflect, flags, r.tri);
    ncolor =
      SbColor(k[R] * rcolor[R],
	      k[G] * rcolor[G],
	      k[B] * rcolor[B]);
  }

  // reflect+refract (Fresnel)
  else {

    // grazing angle problem (normal interpolation error)
    // don't let transmitted ray be on same side
    if (tdir.dot(ngl)>0.0){
      tdir -= (tdir.dot(ngl)+0.0015) * ngl;
      tdir.normalize();
    }
    
    // due to errors for grazing angles
    // with normal interpolation, c may
    // incorrectly be < 0.0. This clamps it.
    if (c<0.0)
      c = 0.0;
    
    R0 = (nt-nc)/(nt+nc);
    R0 *= R0;     // same no matter which way (for n=1.7, R0~0.67
    
    double temp = 1.0-c; 
    Re = R0+(1.0-R0)*temp*temp*temp*temp*temp;

    // use russian roulette
    // (after a few bounces, prescribed in getRadiance1)
    if (flags & C_RUSSIANR){
      double heur = 0.5;
      double P = heur/2.0 + (1.0-heur)*Re;
      
      // reflect
      // OPTIMIZE: remove transparency check below (added for debugging)
      if (r.ctx->mydrand48()<P || transparency==0.0){
	reflect.importance = r.importance*Re;
	SbColor rcolor = (*this.*getRadiance)(reflect, flags, r.tri);
	ncolor =
	  SbColor(k[R] * Re*rcolor[R],
		  k[G] * Re*rcolor[G],
		  k[B] * Re*rcolor[B]);
	ncolor /= P;
      }

      // transmit
      else {
	Ray transmit = Ray(r.x, tdir, r.ctx, r.depth+1, 1-r.atten);
	transmit.importance = r.importance*(1.0-Re);
	SbColor tcolor = (*this.*getRadiance)(transmit, flags, r.tri);
	ncolor =
	  SbColor( k[R]*( (1.0-Re)*tcolor[R] ),
		   k[G]*( (1.0-Re)*tcolor[G] ),
		   k[B]*( (1.0-Re)*tcolor[B] ) );
	ncolor /= (1.0-P);
      }
    }

    // no russian roulette
    else {

      double Fr = Re;
      //double Ft = (1.0-Re)*(nt/nc)*(nt/nc);
      double Ft = 1.0-Re;

// Several experimental Frenel reflection component calculations

//       // WinOSI
//       float ce1, e1, c1, e2;
//       ce1 = r.dir.dot(n);
//       if(fabs(nt-nc)<0.001){
// 	Fr = 0.0f;
// 	goto winosidone;
//       }
//       e1 = acos(fabs(ce1));
//       if(e1<0.01f){
// 	Fr = (nc-nt)/(nc+nt);
// 	Fr *= Fr;
// 	goto winosidone;
//       }
//       c1 = sin(e1) * nc/nt;
//       if(c1 < 1.0){
// 	e2 = asin(c1);
// 	Fr = 0.5f*(float)( sin(e1-e2)*sin(e1-e2)/(sin(e1+e2)*sin(e1+e2)) +
// 			   (tan(e1-e2)*tan(e1-e2))/(tan(e1+e2)*tan(e1+e2)) );
//       }
//       else
// 	Fr = 1.0f;
//
//       winosidone:

		    

//       // WinOSI (optimized)
//       double c12,z,A2B2,AB2,C2,D2,ZC12,face1;
//       float ce1 = r.dir.dot(n);
//       if(fabs(nt-nc)<0.001){
// 	Fr = 0.0;
// 	goto winosidone;
//       }
//       face1=fabs(ce1);
//       if(acos(face1)<0.01f){
// 	Fr = (nc-nt)/(nc+nt);
// 	Fr *= Fr;
// 	goto winosidone;
//       }
//       z=1-face1*face1;
//       c12=z*(nc/nt)*(nc/nt);
//       if(c12<1.0){
// 	ZC12=z*c12;
// 	A2B2=z+c12-ZC12-ZC12;
// 	AB2=sqrt((z-ZC12)*(c12-ZC12));AB2=AB2+AB2;
// 	C2=A2B2-AB2;D2=A2B2+AB2;
// 	Fr = (C2*(1+(1-D2)/(1-C2))/(D2+D2));
// 	goto winosidone;
//       }
//       else Fr = 1.0f;
//
// 	winosidone:
//       Ft = 1.0-Fr;

      
//       // based on glassner p. 737
//       double F0 = R0;
//       double eta = (1.0+sqrt(F0))/(1.0-sqrt(F0));
//       c = ABSF(r.dir.dot(n));
//       double g = eta*eta+c*c-1.0;
//       g = sqrt(g);
//       double aux = c*(g+c);
//       double F = 0.5* ((g-c)*(g-c)) /
//                  ((g+c)*(g+c)) *( 1.0 + ((aux-1.0)*(aux-1.0)) /
// 		    ((aux+1.0)*(aux+1.0)));
//       Fr = clamp(F, 0.0, 1.0);
//       Ft = 1.0-Fr;

//       // from wikipedia
//       float theta1, theta2;
//       theta1 = acos(ABSF(r.dir.dot(n)));
//       theta2 = acos(ABSF(tdir.dot(n)));
//       float Rs = sin(theta1-theta2)/sin(theta1+theta2);
//       Rs *= Rs;
//       float Rp = tan(theta1-theta2)/tan(theta1+theta2);
//       Rp *= Rp;
//       Fr = (Rs+Rp)/2.0;
//       Ft = 1.0 - Fr;


//       // from stanford project
//       // http://graphics.stanford.edu/courses/cs348b-competition/cs348b-02/ong/
//       float rhoParallel;
//       float rhoPerp;
//       float index1 = nc;
//       float index2 = nt;
//       double cosTheta1 = ABSF(r.dir.dot(n));
//       double cosTheta2 = ABSF(tdir.dot(n));
//
//       rhoParallel = ( index2 * cosTheta1 - index1 * cosTheta2 ) /
//             ( index2 * cosTheta1 + index1 * cosTheta2 );
//       rhoPerp = ( index1 * cosTheta1 - index2 * cosTheta2 ) /
//             ( index1 * cosTheta1 + index2 * cosTheta2 );
//       Fr = 0.5 * ( rhoParallel*rhoParallel + rhoPerp*rhoPerp );
//       Ft = 1. - Fr;

      // FIXME: setting deptha and ctx repeatedly is repetitive
      //        (as is calculating depth+1)
      Ray transmit = Ray(r.x, tdir, r.ctx, r.depth+1, 1-r.atten);

      // DEBUG_STATE = 0;
      SbColor tcolor = (*this.*getRadiance)(transmit, flags, r.tri);
      // transmit.show(transmit.t>0.0?transmit.t:100.0, SbColor(Ft, Ft, Ft));
      // DEBUG_STATE = 0;
      SbColor rcolor = (*this.*getRadiance)(reflect, flags, r.tri);
      // reflect.show(reflect.t>0.0?reflect.t:100.0, SbColor(Fr, Fr, Fr));
      // DEBUG_STATE = 1;


//       double cos1 = ABSF(r.dir.dot(n));
//       double cos2 = ABSF(tdir.dot(n));
//       float theta1 = acos(cos1);
//       float theta2 = acos(cos2);
//       theta1 *= 360/(2.0*M_PI);
//       theta2 *= 360/(2.0*M_PI);
// //      cerr << "Angles ( r) " << theta1 << " ( t) " << theta2 << endl;
// //      cerr << "       (Fr) " << Fr     << " (Ft) " << Ft << endl;
// //      Fr = MAX2(Fr, 0.2);

      ncolor =
	SbColor( k[R]*( Fr*rcolor[R]+Ft*tcolor[R] ),
		 k[G]*( Fr*rcolor[G]+Ft*tcolor[G] ),
		 k[B]*( Fr*rcolor[B]+Ft*tcolor[B] ) );

//       ncolor =
// 	SbColor( k[R]*( Re*rcolor[R]+(1.0-Re)*tcolor[R] ),
// 		 k[G]*( Re*rcolor[G]+(1.0-Re)*tcolor[G] ),
// 		 k[B]*( Re*rcolor[B]+(1.0-Re)*tcolor[B] ) );

    }

    
  }
  
  return(ncolor);
}


//WWWW''WW''''''''''WW''''''''''''''''WW''''WW''''''''''''''''WW''''''''''
//WW  W WW          WW                WWW  WWW                            
//WW  W WWW   WW   WWWW  WW   WWW     W WWW WW  WWW WWW  WWW  WW WWW   WWW
//WWWW  WW W WW W   WW  WW W  WW W    W  W  WW W WW WW W WW W WW WW W W WW
//WW    WW W WW W   WW  WW W  WW W    W     WW W WW WW W WW W WW WW W W WW
//WW    WW W  WW    WW   WW   WW W    W     WW  WWW WWW  WWW  WW WW W  WWW
//                                                  WW   WW             WW
//''''''''''''''''''''''''''''''''''''''''''''''''''WW'''WW'''''''''''WWW'
// (getRadiance1)
//
// Returns radiance for a ray, using Photon-mapping path tracing.
// This is vastly more efficient than pure Monte Carlo path tracing,
// which is available in getRadiance2.
//
// Accurate and Approxmate solutions are available, specified
// using the flags variable, which is the binary OR of the following
// available flags:
//
// C_APPROX, C_PREPASS, C_SIMPLE
// C_EMITTED, C_MDIFFUSE,  C_CAUSTIC, C_SPECULAR, C_DIRECT,  
//
// Note the last five flags above refer to the
// the emitted radiance component, which is trivial,
// plus the four rendering components 
// described in Henrik Jensen's Two-Pass Photon Mapping
// Algorithm, in his book "Realistic Image Synthesis Using
// Photon Mapping" Ch.9.
//
// In a nutshell, incoming radiance is split into
// direct, indirect caustic, and indirect diffuse components,
// while the BRDF is split into specular and diffuse reflection
// components.
//
// Use accurate (no C_APPROX) if:
// seen directly by the eye or maybe a few sepcular
// reflections, or if intersection ray is short
// (to prevent inaccurate color bleeding in corners)
//
// Use approximate (C_APPROX) if:
// ray has been reflected diffusely or ray
// contributes little to pixel radiance
//
// Used for pixel color algorithms (-aN):
// 0: Ray tracing (no indirect illumination, but w/ caustics)
// 1: Photon mapping (all four components, using Photon Mapping)
//
SbColor Pane::getRadiance1(Ray &r, int flags, Triangle *ignoreTri){

  SbColor color = SbColor(0.0, 0.0, 0.0);
  r.col = &color;

  // intersection?
  if (tritree.intersect_ray(r, ignoreTri) == 0){

    NOHIT_PM:
    
    // debugging
    if (DEBUG_STATE)
      //r.show(getMaxDist()/4.0, color);
      r.show2(getMaxDist()*4.0, color);

    if (flags & C_EMITTED){
      if (envMap)
	color = envMap->getRadiance(r.dir);
      else
	color = gv->worldBG;
    }
    
    r.t = -1.0;

    // experimental
    if (r.depth==1 && gv->stepsize>0)
      volPointLightContrib(r, color);
    
    return(color);
  }
    
  
    // set flags
    int c_approx   = flags & C_APPROX;
    int c_emitted  = flags & C_EMITTED;
    int c_mdiffuse = flags & C_MDIFFUSE;
    int c_caustic  = flags & C_CAUSTIC;
    int c_specular = flags & C_SPECULAR;
    int c_direct   = flags & C_DIRECT;
    int c_prepass  = flags & C_PREPASS;
    int c_simple   = flags & C_SIMPLE;

    // more flags
    int c_russianr  = c_prepass || c_simple || r.depth>=gv->rrStartDepth;
    if (c_russianr)
      flags |= C_RUSSIANR;
    int c_recurse  = (r.depth < gv->maxDepth)?1:0;
    
    // use approximate values if several specular bounces
    // (if doing global illumination)
    //if (!c_approx && r.depth > 5 && gv->algorithm!=0){
    if (r.importance<0.01 && gv->algorithm==1){
      //cerr << "boo";
      //return(SbColor(1.0, 0.0, 1.0));
      c_approx = 1;
      flags |= C_APPROX;
    }
    
    //return color.setHSVValue(((int)r.tri & 255)/255.0, 1., 1.);
    
    // use approximate calculation for final gather
    // unless t is small.
    if (c_simple){
      if (r.t>minApproxDistance){
	c_approx = 1;
	flags |= C_APPROX;
      }
    }

    // single sided triangles
    if (r.depth == 1 && gv->seeSingleSided==0 /*&& r.tri->emission[0]>0.0 */){
      while (r.tri->getNormal().dot(r.dir)>0.0){
	Ray new_r = Ray(r.pos+r.t*r.dir, r.dir, r.ctx, r.depth, r.atten);
	if (tritree.intersect_ray(new_r, r.tri)){
	  r.t += new_r.t;
	  r.tri = new_r.tri;
	  r.bc = new_r.bc;
	}
	else
	  goto NOHIT_PM;
      }
    }

    // compute point of intersection
    r.x = r.pos + r.t*r.dir;

    // compute normal
    SbVec3f n = r.tri->getIntNormal(r.bc);  // interpolated normal
    SbVec3f ng = r.tri->getNormal();        // geometry normal    
    SbVec3f nl = n;                         // points toward ray source
    SbVec3f ngl = ng;                       // points toward ray source

    // normal correction (two sided lighting)
    if (ng.dot(r.dir)>0.0){
      nl = -nl;
      ngl = -ngl;
    }
    
    if (DEBUG_STATE && gv->showNormals){
      print("t=%f\n",r.t);
      Ray(r.x,n,r.ctx).show(r.t/7.0, SbColor(1.0, 1.0, 0.0));     // yellow
      //printVec(n);
      Ray(r.x,ng,r.ctx).show(r.t/6.0, SbColor(0.0, 1.0, 1.0));    // green
      //printVec(ng);
      Ray(r.x,ngl,r.ctx).show(r.t/5.0, SbColor(1.0, 0.0, 1.0));    // purple
      r.tri->show();
      //cerr << *(r.tri);
    }

    // frequently used constants
    double trans_comp = r.tri->transparency;
    double reg_comp = 1.0-trans_comp;    
    double shiny_comp = r.tri->shininess;
    double diff_comp = 1.0-shiny_comp;

    // do plain OpenGL transparency
    if (gv->useGlassModel == 0){
      if (trans_comp>0.01){
	Ray transmit = Ray(r.x, r.dir, r.ctx, r.depth+1, r.atten);
	color += getRadiance1(transmit, flags, r.tri) * trans_comp;
      }
      
      shiny_comp *= reg_comp;
      diff_comp *= reg_comp;
      reg_comp = 1.0;
      trans_comp = 0.0;
    }

    // fake alpha
    if (gv->figureMode && gv->tempInt==31){
      SbColor surf = r.tri->getIntDiffColor(r.bc);
      if ((surf-SbColor(0.9, 0.9, 0.99)).length()<0.01){
	const float alpha = gv->tempFloat;
	Ray transmit = Ray(r.x, r.dir, r.ctx, r.depth+1, r.atten);
	color += getRadiance1(transmit, flags, r.tri) * alpha;
      
	shiny_comp *= 1.0-alpha;
	diff_comp *= 1.0-alpha;
      }
    }

    // some constants
    const double invMetersConversion2 = 1.0/gv->meters2(1.0);
    double invPi= M_1_PI; // 1.0/M_PI;
    
    // different brdf
    if (gv->useDimensions<3)
      invPi = 1.0/2.0;
    
    // emitted (emissive)
    if (c_emitted && r.tri->emissive && ng.dot(r.dir)<0.0)
      color += r.tri->getIntEmission(r.bc)/gv->meters2(r.tri->area)*invPi;


    //'WWW''''''''''''''''''WW'''''''''''''''''''''''W'''''WWW'WW'''''''''''''
    //W                      W                       W    W     W             
    // WW  WW   WW   WW W W  W   WW  WW     WW WW   WW    W WW  W   W   WWW  W
    //   W W W Wmmm W   W W  W  W W W      W W W W W W    W  W  W  W W  WW   W
    //WWW  WW   WW   WW  WW WWW  WW W       WW W W  WW     WWW WWW  W  WWW  WW
    //'''''W''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    // (specular)

    if (c_specular && c_recurse && r.tri->shiny){

      SbColor specCol = r.tri->specular;

      // fresnel effect
      float fresnel = 1.0 + (r.dir.dot(nl)); // schlick approx + H=n for specular dir
      fresnel = fresnel*fresnel*fresnel*fresnel*fresnel;
      specCol += (SbColor(1.0,1.0,1.0)-specCol)*fresnel;

      // russian roulette
      float prob_recurse;
      if (c_russianr){

	// find maximum component
	float maxComp = 0.0;
	for (int i=0; i<3; i++)
	  if (specCol[i]>maxComp)
	    maxComp = specCol[i];
	
	prob_recurse = shiny_comp*maxComp;

	// increase chance of ray termination
	// this ensures termination, eventually
	if (r.depth>15 || (r.depth>10 && (c_prepass || c_simple))  )
	  prob_recurse *= 0.9;
      }
      else
	prob_recurse = 1.0;

      if (prob_recurse == 1.0 || r.ctx->mydrand48()<prob_recurse){

	float prob_refract;
	float prob_reflect;
	if (c_russianr){

	  // reflect OR refract
	  if (trans_comp != 0.0 && r.ctx->mydrand48() < trans_comp){
	    prob_refract = trans_comp;
	    prob_reflect = 0.0;
	  }
	  else {
	    prob_refract = 0.0;
	    prob_reflect = reg_comp;
	  }

	  // do NOT do any diffuse computation
	  diff_comp = 0.0;
	}
	else {
	  prob_refract = trans_comp?1.0:0.0;
	  prob_reflect = reg_comp?1.0:0.0;
	}
	
	// Refraction component contribution
	if (prob_refract > 0.0){

	  SbColor tcolor = refractionContrib(r, n, ngl, flags);

	  tcolor[R] *= specCol[R];
	  tcolor[G] *= specCol[G];
	  tcolor[B] *= specCol[B];
	
	  color += tcolor * trans_comp * shiny_comp / prob_refract;

	  // dont waste time computing other components
//	  if (trans_comp>0.99 && shiny_comp>0.99 || c_prepass){
//	    goto RETURN_PM;
// 	    if (DEBUG_STATE)
// 	      r.show(r.t, color);
// 	    return(color);
//	  }
	}

	// Reflection component
	if (prob_reflect > 0.0){
	  SbVec3f rdir = r.dir-2.0*n.dot(r.dir)*n;

	  // grazing angle problem
	  // don't let reflected ray go beneath surface
	  if (rdir.dot(ngl)<0.0){
	    rdir += (-rdir.dot(ngl)+0.0015) * ngl;
	    rdir.normalize();
	  }

	  Ray reflect(r.x, rdir, r.ctx, r.depth+1, r.atten);
	
	  SbColor rcolor = getRadiance1(reflect, flags, r.tri);

	  rcolor[R] *= specCol[R];
	  rcolor[G] *= specCol[G];
	  rcolor[B] *= specCol[B];
	
	  color += rcolor * reg_comp * shiny_comp / prob_reflect;

	  // dont waste time computing other components
//	  if (shiny_comp>0.99 || c_prepass){
//	    goto RETURN_PM;
// 	    if (DEBUG_STATE)
// 	      r.show(r.t, color);
// 	    return(color);
//	  }
	
	}// end prob_reflect test

      }// end prob_recurse test

    } // end specular component

    // fetch diffuse color
    SbColor surf = r.tri->getIntDiffColor(r.bc);
    r.diffCol = surf;

    // for transparency
    SbVec3f rn = -n;
    SbVec3f rnl = -nl;
    SbVec3f rng = -ng;
    SbVec3f rngl = -ngl;

    // Only do following computations if reasonable amount of diff_comp
    if (diff_comp>0.01){      
    
      //'WW'''''''''''''''''''''''m'''''''''''''''''''
      //W  W                                   W      
      //WWWW WW  WW   WW  W  W W  W WW W   WW WWW  WW 
      //W  W W W W W W   W W  W   W W W W W W  W  Wmmm
      //W  W WW  WW  W    W  W W  W W W W  WW  W   WW 
      //'''''W'''W''''''''''''''''''''''''''''''''''''
      // (approx)
      
      // approximate solution to Direct, Caustic, and MDiffuse
      if (c_approx && (c_direct || c_caustic || c_mdiffuse)){
	
	float irrad[3];

	// precomputed estimate
	if (gv->precomputeIrr){
	  gPhotonMap->irradiance_estimate_pre(
	    irrad, r.x.getValue(), ngl.getValue());
	}
	
	// regular estimate
	else {
	  gPhotonMap->irradiance_estimate(
	    irrad, r.x.getValue(), nl.getValue(),
	    gv->gPhotonEstimate);
	}

	// from jensen p.21 eq 2.22
	color += SbColor(surf[R] * irrad[R],
			 surf[G] * irrad[G],
			 surf[B] * irrad[B]) *
	  ( invPi * invMetersConversion2 * reg_comp * diff_comp );
	// frd * E(x) = p/pi * E(x)

	if (trans_comp>0.0){
    
	  // precomputed estimate
	  if (gv->precomputeIrr){
	    gPhotonMap->irradiance_estimate_pre(
	      irrad, r.x.getValue(), rngl.getValue());
	  }
	
	  // regular estimate
	  else {
	    gPhotonMap->irradiance_estimate(
	      irrad, r.x.getValue(), rnl.getValue(),
	      gv->gPhotonEstimate);
	    }

	  // from jensen p.21 eq 2.22
	  color += SbColor(surf[R] * irrad[R],
			   surf[G] * irrad[G],
			   surf[B] * irrad[B]) *
	    ( invPi * invMetersConversion2 * trans_comp * diff_comp );
	  // frd * E(x) = p/pi * E(x)
	}

      }

      //'WW''''''''''''''''''''''''''''''
      //W  W                      W      
      //WWWW  WW  WW W W  WW  WW WWW  WW 
      //W  W W   W   W W W   W W  W  Wmmm
      //W  W  WW  WW  WW W    WW  W   WW 
      //'''''''''''''''''''''''''''''''''
      else {

    
	//WWW''m'''''''''''''''''''''WWW'WW''WW''''''''''''m'''''''''''''m''''''''
	//W  W                 W      W   W   W                       W           
	//W  W W  WW  WW   WW WWW     W   W   W  W W WW W  W WW   WW WWW W  W  WW 
	//W  W W W   Wmmm W    W      W   W   W  W W W W W W W W W W  W  W W W W W
	//WWW  W W    WW   WW  W     WWW WWW WWW  WW W W W W W W  WW  W  W  W  W W
	//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	// (direct illumination)

	// accurate solution uses Monte Carlo integration of lights
	// approximate solution uses photon map (see above)
	if (c_direct && diff_comp>0.3){ // don't bother unless important

	  // REGULAR
	  // Point lights contribution
	  color += pointLightContrib2(r, nl, ngl)
	    * reg_comp * diff_comp;

	  // Area lights contribution (incident)
	  // Le from shirley p.124

	  // reflected
	  if (gv->areaLights){
	    color += areaImpLightContrib(r, nl, ngl, c_simple)
	      * reg_comp * diff_comp;
	    
	    if (envMap)
	      color += envMap->impSample(r, nl, ngl, c_simple)
		* reg_comp * diff_comp;
	  }
	  else
	    color += areaLightContrib(r, nl, ngl)
	      * reg_comp * diff_comp;

	  // TRANSMITTED
	  if (trans_comp>0.0){

	    // accurate solution uses Monte Carlo integration of lights
	    // Point lights contribution
	    color += pointLightContrib2(r, rnl, rngl)
	      * trans_comp * diff_comp;

	    // Area lights contribution (incident)
	    // Le from shirley p.124

	    // reflected
	    if (gv->areaLights){
	      color += areaImpLightContrib(r, rnl, rngl, c_simple)
		* trans_comp * diff_comp;
	      
	      if (envMap)
		color += envMap->impSample(r, nl, ngl, c_simple)
		  * reg_comp * diff_comp;
	    }
	    else
	      color += areaLightContrib(r, rn, rngl)
		* trans_comp * diff_comp;
	  }
      
	}

	//'WW'''''''''''''''''''m'''''''''
	//W  W               W            
	//W     WW W W  WWW WWW W  WW  WWW
	//W  W W W W W  WW   W  W W    WW 
	// WW   WW  WW WWW   W  W  WW WWW 
	//''''''''''''''''''''''''''''''''
	// (caustics)

	// REGULAR
	if (c_caustic){
	  float irrad[3];

	  cPhotonMap->irradiance_estimate(
	    irrad, r.x.getValue(), ngl.getValue(),
	    gv->cPhotonEstimate);

	  //double c = 1.0/gv->meters2(1.0);

	  // from jensen p.21 eq 2.22
	  color += SbColor(surf[R] * irrad[R],
			   surf[G] * irrad[G],
			   surf[B] * irrad[B]) *
	    ( invPi * invMetersConversion2 * reg_comp * diff_comp );
	  // frd * E(x) = p/pi * E(x)
	}

	// TRANSMITTED
	if (c_caustic && trans_comp>0.0){
	  float irrad[3];

	  cPhotonMap->irradiance_estimate(
	    irrad, r.x.getValue(), rngl.getValue(),
	    gv->cPhotonEstimate);

	  double c = 1.0/gv->meters2(1.0);

	  // from jensen p.21 eq 2.22
	  color += SbColor(surf[R] * irrad[R],
			   surf[G] * irrad[G],
			   surf[B] * irrad[B]) *
	    ( invPi * invMetersConversion2 * trans_comp * diff_comp );
	  // frd * E(x) = p/pi * E(x)
	  
	}

	// approximate evaluation included in
	// approximate direct lighting estimate

	//W''''W'''''WW''''''m'''''WW''''''''''WWW''m'''WW'''WW''''''''''''''
	//WW  WW      W   W         W          W  W    W    W                
	//W WW W W W  W  WWW W WW   W   WW     W  W W WWW  WWW  W W  WWW  WW 
	//W    W W W  W   W  W W W  W  Wmmm    W  W W  W    W   W W  WW  Wmmm
	//W    W  WW WWW  W  W WW  WWW  WW     WWW  W  W    W    WW WWW   WW 
	//'''''''''''''''''''''W'''''''''''''''''''''''''''''''''''''''''''''
	// multiple diffuse (md)
	if (c_mdiffuse && c_recurse && diff_comp >0.01){

	  // choose random cosine weighted direction (on hemisphere) to fire
	  
	  // REGULAR
	  if (trans_comp<1.0){

	    // debugging
	    SbColor ret;
	    color += ret = multipleDiffuseContrib(r, nl, ngl, c_simple, c_prepass,
					    FRONTSIDE)
	      * reg_comp * diff_comp;

	  }

	  // TRANSMITTED
	  if (trans_comp>0.0){
	    color += multipleDiffuseContrib(r, rnl, rngl, c_simple, c_prepass,
					    BACKSIDE)
	      * trans_comp * diff_comp;
	
	  }

	} // end Multiple Diffuse component calculation

      }// end Accurate computations
    }// end diffuse component computations



    // do Beer's law attenuation
    if (r.atten){
      color[R] *= exp(gv->attenuation[R]*r.t);
      color[G] *= exp(gv->attenuation[G]*r.t);
      color[B] *= exp(gv->attenuation[B]*r.t);
    }
    
    // debugging
    if (DEBUG_STATE)
      r.show(r.t, color);

    // experimental
    if (r.depth==1 && gv->stepsize>0)
      volPointLightContrib(r, color);

    return(color);

}


//WWWW''''''''WW''WW''''''WWWWWW'''''''''''''''''WW''''''''''
//WW  W       WW  WW        WW                               
//WW  W  WWW WWWW WWW       WW    W WW  WWW  WWW WW WWW   WWW
//WWWW  W WW  WW  WW W      WW    WW   W WW WW W WW WW W W WW
//WW    W WW  WW  WW W      WW    WW   W WW WW   WW WW W W WW
//WW     WWW  WW  WW W      WW    WW    WWW  WWW WW WW W  WWW
//                                                         WW
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''WWW'
//  
// Used for pixel color algorithms (-aN):
// 2: MCPT w/ Explicit Lighting test, and   (faster)
// 3: MCPT w/o
//
SbColor Pane::getRadiance2(Ray &r, int flags, Triangle *ignoreTri){
  
  SbColor color = SbColor(0.0, 0.0, 0.0);

  // intersection?
  if (tritree.intersect_ray(r, ignoreTri)){

    // set flags
    int c_emitted  = flags & C_EMITTED;
    int c_explicitl  = flags & C_EXPLICITL;
    
    // single sided triangles
    if (r.depth == 1 && gv->seeSingleSided==0){
      while (r.tri->getNormal().dot(r.dir)>0.0){
	Ray new_r = Ray(r.pos+r.t*r.dir, r.dir, r.ctx, r.depth, r.atten);
	if (tritree.intersect_ray(new_r, r.tri)){
	  r.t += new_r.t;
	  r.tri = new_r.tri;
	  r.bc = new_r.bc;
	}
	else
	  goto NOHIT_MCPT;
      }
    }

    // compute point of intersection
    r.x = r.pos + r.t*r.dir;

    // compute normal
    SbVec3f n = r.tri->getIntNormal(r.bc);  // interpolated normal
    SbVec3f ng = r.tri->getNormal();        // geometry normal    
    SbVec3f nl = n;                         // points toward ray source
    SbVec3f ngl = ng;                       // points toward ray source

    // normal correction (two sided lighting)
    if (ng.dot(r.dir)>0.0){
      nl = -nl;
      ngl = -ngl;
    }
    
    if (DEBUG_STATE && gv->showNormals){
      Ray(r.x,n,r.ctx).show(r.t/7.0, SbColor(1.0, 1.0, 0.0));     // yellow
      //printVec(n);
      Ray(r.x,ng,r.ctx).show(r.t/6.0, SbColor(0.0, 1.0, 1.0));    // green
      //printVec(ng);
      Ray(r.x,ngl,r.ctx).show(r.t/5.0, SbColor(1.0, 0.0, 1.0));    // purple
      r.tri->show();
    }

    // frequently used constants
    double trans_comp = r.tri->transparency;
    double reg_comp = 1.0-trans_comp;    
    double shiny_comp = r.tri->shininess;
    double diff_comp = 1.0-shiny_comp;

    // do plain OpenGL transparency
    if (gv->useGlassModel == 0){
      if (trans_comp>0.01){
	Ray transmit = Ray(r.x, r.dir, r.ctx, r.depth+1, r.atten);
	color += getRadiance2(transmit, flags, r.tri) * trans_comp;
      }
      
      shiny_comp *= reg_comp;
      diff_comp *= reg_comp;
      reg_comp = 1.0;
      trans_comp = 0.0;
    }

//     // for transparency
//     if (r.ctx->mydrand48() < trans_comp){
//       n = -n;
//       nl = -nl;
//       ng = -ng;
//       ngl = -ngl;
//     }

    // get colors
    SbColor specCol = r.tri->specular;
    
    // fresnel effect
    float fresnel = 1.0 + (r.dir.dot(nl)); // schlick approx + H=n for specular directions
    fresnel = fresnel*fresnel*fresnel*fresnel*fresnel;
    specCol += (SbColor(1.0,1.0,1.0)-specCol)*fresnel;
    
    SbColor surf = r.tri->getIntDiffColor(r.bc);
    r.diffCol = surf;

    if (c_explicitl){
      // Point lights contribution
      color += pointLightContrib(r, nl, ngl)
	* reg_comp * diff_comp;
    
      // Area lights contribution (incident)
      color += areaImpLightContrib(r, nl, ngl, C_SIMPLE)
	* reg_comp * diff_comp;

      // Area lights contribution (incident)
      color += envMap->impSample(r, nl, ngl, C_SIMPLE)
	* reg_comp * diff_comp;
    }

    // Emissive contribution
    if (c_emitted){
      if (r.dir.dot(ng)<0.0)
	color += r.tri->emission /           // in W
	  gv->meters2( r.tri->area ) / M_PI;
    }
    else
      flags |= C_EMITTED; // reset flag for future rays

    int c_recurse  = (r.depth < gv->maxDepth)?1:0;
    double prob_recurse;
    if (c_recurse){

      // do refraction OR reflection OR diffuse
      int choice; // 0 = refraction, 1=reflection, 2=diffuse
      double u = r.ctx->mydrand48();

      // diffuse
      if (u < diff_comp)
	choice = 2;
      // reflection
      else if (u < diff_comp + reg_comp*shiny_comp)
	choice = 1;
      // refraction
      else {
	choice = 0;
      }

      // Refraction component contribution
      if (choice == 0){
	prob_recurse = (specCol[R]+specCol[G]+specCol[B])/3.0;
	if (r.ctx->mydrand48()<prob_recurse){

	  SbColor rcolor = refractionContrib(r, n, ngl, flags);

	  rcolor[R] *= specCol[R];
	  rcolor[G] *= specCol[G];
	  rcolor[B] *= specCol[B];
	  
	  color += rcolor * 1.0/prob_recurse;
	}
      }

      // Shiny component contribution
      else if (choice == 1){

	prob_recurse = (specCol[R]+specCol[G]+specCol[B])/3.0;
	if (r.ctx->mydrand48()<prob_recurse){
	
	  SbVec3f rdir = r.dir-2.0*n.dot(r.dir)*n;

	  // grazing angle problem
	  // don't let reflected ray go beneath surface
	  if (rdir.dot(ngl)<0.0){
	    rdir += (-rdir.dot(ngl)+0.0015) * ngl;
	    rdir.normalize();
	  }

	  Ray reflect(r.x, rdir, r.ctx, r.depth+1, r.atten);
	  
	  SbColor rcolor = getRadiance2(reflect, flags, r.tri);

	  rcolor[R] *= specCol[R];
	  rcolor[G] *= specCol[G];
	  rcolor[B] *= specCol[B];
	  
	  color += rcolor * 1.0/prob_recurse;
	}
      }

      // Path Tracing component (interreflection)
      else {

	prob_recurse = (surf[R]+surf[G]+surf[B])/3.0;
	prob_recurse = 0.5 + prob_recurse/2.0;
	if (r.ctx->mydrand48()<prob_recurse){

	  // transmit
	  if (r.ctx->mydrand48()<trans_comp)
	    ngl = -ngl;

	  // choose new direction according to cos(theta)
	  SbRotation sbr;
	  SbVec3f neww;
	  Ray bounce = Ray(r.x, ngl, r.ctx, r.depth+1, r.atten);
	  sbr.setValue(SbVec3f(0.0, 1.0, 0.0),
		       nl);
	
	  double eps1 = r.ctx->mydrand48();
	  double eps2 = r.ctx->mydrand48();
	  double theta = acos(sqrt( eps1 ));
	  double phi = 2.0 * M_PI * eps2;
	
	  neww.setValue(sin(theta)*cos(phi),
			cos(theta),
			sin(theta)*sin(phi));
    
	  sbr.multVec(neww, bounce.dir);

	  // make sure vector goes up
	  if (bounce.dir.dot(ngl)>0.0){

	    if (c_explicitl)
	      flags ^= C_EMITTED;
	    SbColor dcolor = getRadiance2(bounce, flags, r.tri);

	    dcolor[R] *= surf[R];
	    dcolor[G] *= surf[G];
	    dcolor[B] *= surf[B];

	    color += dcolor * 1.0/prob_recurse; 
	    // /M_PI; (cancels with cos0 normalization)
	  }
	}

      }
    }
    
    // debugging
    if (DEBUG_STATE)
      r.show(r.t, color);

    // do Beer's law attenuation
    if (r.atten){
      color[R] *= exp(gv->attenuation[R]*r.t);
      color[G] *= exp(gv->attenuation[G]*r.t);
      color[B] *= exp(gv->attenuation[B]*r.t);
    }
    
  }
  // light did not intersect anything
  else {
    NOHIT_MCPT:
    
    r.t = -1;
    if (gv->environMapFile && (flags & C_EMITTED)){
      color = envMap->getRadiance(r.dir);
      const SbColor topColor = SbColor(0.8196, 0.9412, 1.0)*1.5;
      const SbColor bottomColor = SbColor(0.1313, 0.3628, 0.0765)*1.5;
//     if (1) // crude horizon map
//       if (r.dir[1]>0.0)
// 	color = topColor;
//                 //SbColor(0.7216, 0.9098, 1.0);
//                 //SbColor(0.4863, 0.8118, 1.0);
//       else
// 	color = bottomColor;
//                 //SbColor(0.2627, 0.7255, 0.1529);
//                 //SbColor(0.6, 0.3686, 0.0);
    }
    else
      color = gv->worldBG;
    
    // debugging
    if (DEBUG_STATE)
      r.show(getMaxDist()/4.0, color);
  }

  return(color);
}

// Sets the gv->cam variable.
void Pane::findCamera(){

  // Search for camera in scene graph
  SoSearchAction sa;
  sa.setType(SoCamera::getClassTypeId());
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.apply(gv->root);
  if (sa.getPath())
    gv->cam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();

  // else Look in root at second child (put there in Main.cxx)
  else if (gv->root->getChild(1)->isOfType(SoCamera::getClassTypeId()))
    gv->cam = (SoCamera *)gv->root->getChild(1);

  // or give up
  else {
    cerr << "Could not find camera!" << endl;
    return;
  }
}

// Positions the headlight in the litPts list.
// Call findCamera() first.
void Pane::configureHeadlight(){

  // position headlight (to above and left of camera)
  if (gv->headLightPos==HL_OFFCENTER){
    SbVec3f e, g, vup, w, u, v, x, hl;
    SbRotation sbr;
    
    gv->cam->orientation.getValue().multVec(SbVec3f(0.0, 1.0, 0.0), vup);
    e = SbVec3f(gv->cam->position.getValue());
    sbr = gv->cam->orientation.getValue();
    sbr.multVec(SbVec3f(0.0, 0.0, -1.0), g);
    g.normalize();
    w = SbVec3f(-g);
    x = SbVec3f(vup.cross(w));
    u = SbVec3f(x/x.length());
    v = SbVec3f(w.cross(u));
    hl = -u*getMaxDist()/3.0 + v*getMaxDist()/3.0 + e;
    (*(litPnts.begin()))->location.setValue(hl);
  }
  else if (gv->headLightPos != HL_OFF){
    (*(litPnts.begin()))->location.setValue(gv->cam->position.getValue());
  }

  // still need to account for orthographic camera
}




//''''''''''''''''''''''''''''''''WWW''''''''WW'''''''''''
//                               WW  W       WW           
// WWW  WWW WWW  WW W  WWW  WWWW  WW    WW  WWWW W WW WWW 
//WW W W WW WW W WW W W WW WWW     WW  WWWW  WW  W WW WW W
//WW   W WW WW W WW W W WW   WWW W  WW WW    WW  W WW WW W
// WWW  WWW WW W  WW   WWW WWWW   WWW   WWW  WW   WWW WWW 
//                                                    WW  
//''''''''''''''''''''''''''''''''''''''''''''''''''''WW''
// sets basic canvas parameters, such as viewing direction,
// frustrum
void Pane::canvasSetup(){

  // ripped from Realistic Ray Tracing p.40 by Peter Shirley

  // Up vector
  if (gv->cam)
    gv->cam->orientation.getValue().multVec(SbVec3f(0.0, 1.0, 0.0),
					    canvas.vup);
  else
    canvas.vup = SbVec3f(0.0, 1.0, 0.0);

  // Camera center
  // FIXME: more code is needed if you want to use an orthographic camera.
  if (gv->cam)
    canvas.e = SbVec3f(gv->cam->position.getValue());
  else
    canvas.e = SbVec3f(0.0, 0.0, tritree.max_dist_of_vert*2.0);

  // Camera direction
  if (gv->cam){
    canvas.sbr = gv->cam->orientation.getValue();
    canvas.sbr.multVec(SbVec3f(0.0, 0.0, -1.0), canvas.g);
    canvas.g.normalize();
  }
  else
    canvas.g = SbVec3f(0.0, 0.0, -1.0);

  // Distance to viewplane
  canvas.s = gv->cam->focalDistance.getValue();

  // height, width
  if (gv->cam->isOfType(SoPerspectiveCamera::getClassTypeId())){
    float angle = ((SoPerspectiveCamera *)gv->cam)->heightAngle.getValue();
    float dist = ((SoPerspectiveCamera *)gv->cam)->focalDistance.getValue();
    canvas.height = tanf(angle/2.0)*dist*2.0;;
    canvas.orthographic = 0;
  }
  else if (gv->cam->isOfType(SoOrthographicCamera::getClassTypeId())){
    canvas.height = ((SoOrthographicCamera *)gv->cam)->height.getValue();
    canvas.orthographic = 1;
  }
  else {
    cerr << "WARNING: Unknown camera type!\n";
    canvas.height = 45.0;
  }
  canvas.width = ABSF(gv->imgDim[X]/(double)gv->imgDim[Y])*canvas.height;

  // Orthonormal basis for viewplane [0]
  canvas.w = SbVec3f(-canvas.g);

  // Orthonormal basis for viewplane [1]
  canvas.x = SbVec3f(canvas.vup.cross(canvas.w));
  canvas.u = SbVec3f(canvas.x/canvas.x.length());

  // Orthonormal basis for viewplane [2]
  canvas.v = SbVec3f(canvas.w.cross(canvas.u));

}



//'''''''''''''''''WW''''''''''''WW'WW'
//                 WW           WW   WW
//W WW  WW  WWW   WWW  WW  W WW WW   WW
//WW   WWWW WW W W WW WWWW WW   WW   WW
//WW   WW   WW W W WW WW   WW   WW   WW
//WW    WWW WW W  WWW  WWW WW    WW WW 
//                                     
//'''''''''''''''''''''''''''''''''''''
// (render)
//
// Called via a keypress or the main program.
//
// Prepares to compute the image pixel values, then
// creates threads which call renderTile() to compute the values
// for each sub-region.
//
// Assigns each thread one row, except in the case of
// using a sample map, where each thread uses random samples.
void Pane::render(int trace){

  cerr << "Preparing to render...\n";
  
  // check for lights
  if (gv->environMapFile==NULL && gv->sceneHasLights==0){
    cerr << "Scene has no lights!! Aborting render!\n";
    return;
  }
  
  // Set default photon estimate
  // better to have too many that not enough for given radius size
  if (gv->gPhotonEstimate<0){
    gv->gPhotonEstimate = clamp(gPhotonMap->get_stored_photons()/100, 8, 100);
    cerr << "Using global photon estimate of " << gv->gPhotonEstimate << ".\n";
  }
  if (gv->cPhotonEstimate<0){
    gv->cPhotonEstimate = clamp(cPhotonMap->get_stored_photons()/100, 8, 50);
    cerr << "Using caustic photon estimate of " << gv->cPhotonEstimate << ".\n";
  }

  // balance Photon map
  static int last_stored_photons = 0;
  if (last_stored_photons !=
      (gPhotonMap->get_stored_photons() +
       cPhotonMap->get_stored_photons()) ){
    
    last_stored_photons = gPhotonMap->get_stored_photons()
      + cPhotonMap->get_stored_photons();

    // start balancing timer
    struct timeval start, end;
    gettimeofday(&start, NULL); 
    
    cerr << "Balancing global photon map...";
    TIME(gPhotonMap->balance(gv->gPhotonEstimate));

    cerr << "Balancing caustic photon map...";
    TIME(cPhotonMap->balance(gv->cPhotonEstimate));


    // stop balancing timer
    gettimeofday(&end, NULL);   // end delay measure
    gv->times.photonBalancing = delay(start, end);

    // set mininum distance for ray to travel to use approximation
    minApproxDistance = gPhotonMap->max_radius(gv->gPhotonEstimate);

    // precompute irradiance
    if ((gv->algorithm==1 && gv->precomputeIrr) || gv->showPhotons){

      cerr << "Precomputing irradiance for global photon map...\n";
      gettimeofday(&start, NULL); // start delay measure
      float estdist = gPhotonMap->precompute_irradiance(gv->gPhotonEstimate,
							gv->numThreads,
							gv->showPhotons);
      minApproxDistance = estdist;
      
      cerr << "Average maximum photon distance in estimate is " << estdist << "\n";
      gettimeofday(&end, NULL);   // end delay measure
      gv->times.photonIrradiance = delay(start, end);
      cerr << "Time: " << gv->times.photonIrradiance << "ms" << endl;
    }

    // report minimum approx distance
    cerr << "Using minimum approximation distance " << minApproxDistance << ".\n";
  }

  // feedback
  struct timeval start, end;
  cerr << "Rendering...\n";
  gettimeofday(&start, NULL); // start delay measure
  struct tms renderTimes;
  times(&renderTimes);
  clock_t startTicks = renderTimes.tms_utime + renderTimes.tms_cutime;
    
  // find camera
  findCamera();

  // positon headlight
  configureHeadlight();
  
  // precompute multipleDiffuseContrib tables
  precomputeMDCTables();
  
  // reset image
  for (int j=0; j<gv->imgDim[Y]; j++)
    for (int i=0; i<gv->imgDim[X]; i++)
      setPix(i, j, SbColor(0.0, 0.0, 0.0), 0.0);

  // basic canvas setup, mostly camera settings
  canvasSetup();

  // further canvas setup
  // set up flags
  if (gv->renderFlags){
    canvas.flags = gv->renderFlags |
      gv->approximate * Pane::C_APPROX;
    cerr << "Rendering with: ";
    if (canvas.flags&C_APPROX)
      cerr << "approx ";
    if (canvas.flags&C_SPECULAR)
      cerr << "specular ";
    if (canvas.flags&C_DIRECT)
      cerr << "direct ";
    if (canvas.flags&C_EMITTED)
      cerr << "emitted ";
    if (canvas.flags&C_CAUSTIC)
      cerr << "caustic ";
    if (canvas.flags&C_MDIFFUSE)
      cerr << "mdiffuse ";
    cerr << "\n";
  }
  else {
    canvas.flags =
      Pane::C_NORMAL |
      (gv->algorithm==1?1:0) * Pane::C_MDIFFUSE |
      gv->approximate * Pane::C_APPROX;
  }
  
  // set function pointer for getRadiance()
  if (gv->algorithm==0 ||           // Ray tracing
      gv->algorithm==1)             // Photon Mapping
    getRadiance = &Pane::getRadiance1;
  else if (gv->algorithm==2 ||      // MCPT w/EL
	   gv->algorithm==3){       // MCPT
    getRadiance = &Pane::getRadiance2;
    canvas.flags |= C_RUSSIANR;     // No branching
    
    if (gv->algorithm==2)
      canvas.flags |= C_EXPLICITL;
  }
  else 
    getRadiance = &Pane::getRadiance1;    // default
  
  // further setup
  gv->useIrradCache = gv->approximate?0:1 &&
    canvas.flags&C_MDIFFUSE &&
    gv->algorithm==1 &&
    gv->irradCacheError>0.0 &&
    gv->useIndirectImpSampling == 0 &&
    gv->bumpMapping == 0;
  canvas.prepass = gv->useIrradCache && canvas.flags;
  canvas.sqrt_samples = (int)sqrt((float)gv->numSamples);
  
  // trace pixel 
  canvas.trace = trace;
  if (trace){      // clear previous rays
    Ray r;
    r.show(0.0, SbColor(0.0, 0.0, 0.0), 1);
  }

  // initialize mutexes
  for (int i=0; i<mutexNum; i++)
    pthread_mutex_init(&mutexes[i], NULL);

  // launch threads
  MemberFunction *memberFunc = (MemberFunction *)alloca(
    sizeof(MemberFunction)*gv->maxThreads); // used to pass object pointer
  thread_context *ctx = (thread_context *)alloca(
    sizeof(thread_context)*gv->maxThreads);   // random number generators

  // scramble random number generators
  for (int i=0; i<gv->numThreads; i++){
    ctx[i].mydrand48 = RNG(lrand48());
    ctx[i].id = i;
    ctx[i].rays = 0;
    ctx[i].ticks = 0;
  }
  
  // render iv
  if (gv->renderIv)
    renderIv(&ctx[0]);
  
  // render with one thread
  else if (gv->numThreads == 1){
    renderTile(&ctx[0]);
    renderRaysFired += ctx[0].rays;
  }
  
  // render with multiple threads
  else {

    for (int i=0; i<gv->numThreads; i++){

      // pack object pointer, function pointer, and argument into struct
      memberFunc[i].object = this;
      memberFunc[i].pmf = &Pane::renderTile;
      memberFunc[i].ctx = &ctx[i];

      // call wrapper function with struct
      pthread_create(&gv->threads[i], NULL, Pane::renderTileWrapper,
		     (void*)&memberFunc[i]);
    }
  
    // wait for threads to finish
    for (int i=0; i<gv->numThreads; i++){
      pthread_join(gv->threads[i], NULL);
      renderRaysFired += ctx[i].rays;
      gv->times.renderTicks += ctx[i].ticks;
      gv->times.totalTicks += ctx[i].ticks;
    }

  }

  // destroy mutexes
  for (int i=0; i<mutexNum; i++)           
    pthread_mutex_destroy(&mutexes[i]);


  // normalize image if sample map
  if (sampMap){
    // normalize sample values
    for (int j=0; j<gv->imgDim[Y]; j++)
      for (int i=0; i<gv->imgDim[X]; i++){
	int idx = gv->imgDim[X]*j+i;
	if (pixDepth[idx]>0.0)
	  pix[idx] /= pixDepth[idx];
      }
  }
  
  print("\n%d stored points in irradiance cache (%d rejected).\n",
	irradCache.numPoints(),
	  rejectedICPoints);

  // timing
  times(&renderTimes);
  clock_t endTicks = renderTimes.tms_utime + renderTimes.tms_cutime;
  gv->times.renderTicks += endTicks - startTicks;
  long ticksPerSec = sysconf(_SC_CLK_TCK);
  float ticksMs = gv->times.renderTicks/(double)ticksPerSec*1000.0;
  print ("\nRendering CPU ticks: %d, Ticks/s: %d\n", gv->times.renderTicks, ticksPerSec);
  print("Rendering CPU time: (%f ms, %f s, %f m, %f h)\n",
	ticksMs, ticksMs/1000.0, ticksMs/1000.0/60.0, ticksMs/1000.0/60.0/60.0);
  
  gettimeofday(&end, NULL);   // end delay measure
  gv->times.render = delay(start, end); // renderTime - photon balancing + irradiance
  int ms = gv->times.render;

  print("Actual Time:        (%d ms, %f s, %f m, %f h)\n", ms, ms/1000.0, ms/1000.0/60.0,
	ms/1000.0/60.0/60.0);

  // rays statistics
  print("%d total rays fired, for %.2f rays/sec (cpu time), %.2f rays/sec (actual time)\n",
	renderRaysFired,
	renderRaysFired/ticksMs*1000.0,
	renderRaysFired/(ms/1000.0));
  
  if (!gv->silent)
    print("\a");
  print("Done.\n");
  
  toneMapped = 0;

} // end render()
  

//'''''''''''''''''WW'''''''''''WWWWWW''WW'WWW''''''''WW'WW'
//                 WW             WW        WW       WW   WW
//W WW  WW  WWW   WWW  WW  W WW   WW    WW  WW   WW  WW   WW
//WW   WWWW WW W W WW WWWW WW     WW    WW  WW  WWWW WW   WW
//WW   WW   WW W W WW WW   WW     WW    WW  WW  WW   WW   WW
//WW    WWW WW W  WWW  WWW WW     WW    WW WWWW  WWW  WW WW 
//                                                          
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
// (renderTile)
//
// Renders a tile (row) of the image. Computes pixel
// color for each pixel in tile (except when using a
// sample map, then computes values for random pixels).
//
// Computing pixel colors is done in two passes.
// This is to reduce artifacts from irradiance caching.
// With one pass, subsequent pixels will see and interpolate
// indirect lighting samples that previous pixels didn't have yet,
// creating visible lines in the indirect illumination.
// 
// The first pass computes at least one indrect illumination
// sample for each pixel, roughly populating the cache.
//
// The second, final pass, computes the color for each pixel.
//
void Pane::renderTile(thread_context *ctx){

  // for each pixel
  int i, j;
  Ray r;
  r.ctx = ctx;
  r.depth = 1;
  r.pos = canvas.e;
  
  // image size calc (in image space)
  SbColor color = SbColor(0.0, 0.0, 0.0);
  int lowerLeft[2];
  int upperRight[2];
  double offset[2];
  double samples[2];
  if (ctx->id==0)
    print("Using %d samples.\n", canvas.sqrt_samples*canvas.sqrt_samples);

  // render range
  lowerLeft[X] = 0;
  lowerLeft[Y] = 0;
  upperRight[X] = gv->imgDim[X];
  upperRight[Y] = gv->imgDim[Y];

  // Restrict pixels to debugging range
  if (gv->debugRange >= 0){

    // get pixel center
    int center[2];
    if (gv->debugPixel[X]<0 && gv->debugPixel[Y]<0){
      // trace center
      center[X] = gv->imgDim[X]/2;
      center[Y] = gv->imgDim[Y]/2;
    }
    else {
      center[X] = gv->debugPixel[X];
      center[Y] = gv->imgDim[Y]-gv->debugPixel[Y]-1;
    }
    
    int size = gv->debugRange;
    lowerLeft[X]  = center[X] -size;
    lowerLeft[Y]  = center[Y] -size;
    upperRight[X] = center[X] +size+1;
    upperRight[Y] = center[Y] +size+1;

    // adjust for reverse Y
//     lowerLeft[Y] = gv->imgDim[Y] - lowerLeft[Y] - 1;
//     upperRight[Y] = gv->imgDim[Y] - upperRight[Y] - 1;
//     upperRight[X]++; upperRight[Y]++; // inclusive

    //canvas.prepass = 0;
    //DEBUG_STATE=1;
  }

  // Trace a single pixel
  if (canvas.trace){

    if (gv->debugPixel[X]<0 && gv->debugPixel[Y]<0){
      // trace center of image
      i = gv->imgDim[X]/2;
      j = gv->imgDim[Y]/2;
    }
    else {
      i = gv->debugPixel[X];
      j = gv->imgDim[Y]-gv->debugPixel[Y]-1;
    }

    lowerLeft[X] = i;
    upperRight[X] = i+1;      
    lowerLeft[Y] = j;
    upperRight[Y] = j+1;
  }
      
  // bounds checking
  lowerLeft[X] = MAX2(lowerLeft[X], 0);
  lowerLeft[Y] = MAX2(lowerLeft[Y], 0);
  upperRight[X] = MIN2(upperRight[X], gv->imgDim[X]);
  upperRight[Y] = MIN2(upperRight[Y], gv->imgDim[Y]);    


  //'WWW'''''''''''''''WW''''''''''W''''W''''''''
  //W                   W          WW  WW        
  // WW   WW WW W  WW   W   WW     W WW W  WW WW 
  //   W W W W W W W W  W  Wmmm    W    W W W W W
  //WWW   WW W W W WW  WWW  WW     W    W  WW WW 
  //               W                          W  
  //'''''''''''''''W''''''''''''''''''''''''''W''
  //
  // If a sample map is provided, use rejection sampling
  // of the map (scaled) to choose pixels to sample.
  if (sampMap){

    // divvy up samples
    for (int i=0; i<gv->numSamples/gv->numThreads; i++){

      // completion meter
      completion[ctx->id] = 100.0*i/(gv->numSamples/gv->numThreads);
      if (ctx->id==0 && i%100==0){
	print("\r");
	for (int i=0; i<gv->numThreads; i++)
	  print("%6.2f%%%c ", completion[i], completion_stage[i]);
      }
    
      // rejection sample the map
      float du, dv;
      do {
	du = ctx->mydrand48();
	dv = ctx->mydrand48();
      } while (ctx->mydrand48()>sampMapValue(du, dv));

      // set direction
      r.dir = SbVec3f(

	// sideways component
	( -canvas.width/2.0 + du * canvas.width ) *
	canvas.u +

	// vertical component
	( -canvas.height/2.0 + dv * canvas.height ) *
	canvas.v +
	
	// distance ahead component
	-canvas.s*canvas.w  
	);
      
      // adjust for orthographic camera
      if (canvas.orthographic){
	r.pos = canvas.e + r.dir + canvas.s*canvas.w;
	r.dir = -canvas.w;
      }
      r.dir.normalize();

      // debugging a single ray
      SbColor c;
      if (canvas.trace){
	DEBUG_STATE=1;
	c = (*this.*getRadiance)(r, canvas.flags, NULL);
	DEBUG_STATE=0;

	pthread_mutex_lock(&mutexes[LOCK_STDERR]);
	print("Sample color: ");
	printCol(c);
	print("\n");
	pthread_mutex_unlock(&mutexes[LOCK_STDERR]);
      }
      else
	c = (*this.*getRadiance)(r, canvas.flags, NULL);

      int x, y;
      x = (int)(du*gv->imgDim[X]);
      y = (int)(dv*gv->imgDim[Y]);

      int idx = gv->imgDim[X]*y+x;
      pix[idx] += c;
      pixDepth[idx] += 1.0;
    }
    
    goto renderTileDone;
  }


  
  //WWW''''''''''''''W''''''WWW'''''''''''''''
  //W  W             W      W  W              
  //WWW   W  W W  WW WW     WWW   WW  WWW  WWW
  //W  W W W W W W W W W    W    W W  WW   WW 
  //W  W  W   WW  WW W W    W     WW WWW  WWW 
  //               W                          
  //'''''''''''''WW'''''''''''''''''''''''''''
  // Rough pass to precalculate most irradiance values
  
  if (canvas.prepass && !canvas.trace){

    if (ctx->id==0)
      print("Precalculating indirect illumination (p)...\n");
    completion_stage[ctx->id] = 'p';
    
    for (j=lowerLeft[Y]+ctx->id; j<upperRight[Y]; j+=gv->numThreads){

      // completion meter
      completion[ctx->id] = j*100.0/gv->imgDim[Y];
      if (ctx->id==0){
	print("\r");
	for (int i=0; i<gv->numThreads; i++)
	  print("%6.2f%%%c ", completion[i], completion_stage[i]);
      }
      
      for (i=lowerLeft[X]; i<upperRight[X]; i++){

	// for removing randomness
	if (gv->nonRandomPixels)
	  ctx->mydrand48.setSeed((i+1)*(j+1));
	  
	// create Ray
	r.dir = SbVec3f(

	  // sideways component
	  ( -canvas.width/2.0 +
	    ( (double)i + 0.5) * canvas.width /
	    gv->imgDim[X] ) *
	  canvas.u +

	  // vertical component
	  ( -canvas.height/2.0 +
	    ( (double)j + 0.5) * canvas.height /
	    gv->imgDim[Y] ) *
	  canvas.v +

	  // distance ahead component
	  -canvas.s*canvas.w  
	  );

	// adjust for orthographic camera
	if (canvas.orthographic){
	  r.pos = canvas.e + r.dir + canvas.s*canvas.w;
	  r.dir = -canvas.w;
	}
	r.dir.normalize();
	

	// trace ray

	//debugging
	getRadiance1(r, C_SPECULAR|C_MDIFFUSE|C_PREPASS, NULL);
	//getRadiance1(ctx, r, 0, C_MDIFFUSE);

      } // end for (i=...
    } // end for (j=...

    // completion meter
    completion[ctx->id] = 100.0;
    if (ctx->id==0){
      print("\r");
      for (int i=0; i<gv->numThreads; i++)
	print("%6.2f%%%c ", completion[i], completion_stage[i]);
    }
    
  }

  if (ctx->id==0){
    print("\n%d stored points in irradiance cache (%d rejected).\n",
	  irradCache.numPoints(), rejectedICPoints);
  }

  //WWWW'm'''''''''WW'''''WWW'''''''''''''''''''
  //W               W      W                    
  //WWW  W WW   WW  W      W  WW W   WW  WW  WW 
  //W    W W W W W  W      W  W W W W W W W Wmmm
  //W    W W W  WW WWW    WWW W W W  WW  WW  WW 
  //                                      W     
  //''''''''''''''''''''''''''''''''''''WW''''''
  // Make final image

  if (ctx->id==0)
    print("Making final render pass (f)...\n");
  completion_stage[ctx->id] = 'f';
  
  for (j=lowerLeft[Y]+ctx->id; j<upperRight[Y]; j+=gv->numThreads){

    // completion meter
    completion[ctx->id] = j*100.0/gv->imgDim[Y];
    if (ctx->id==0){
      print("\r");
      for (int i=0; i<gv->numThreads; i++)
	print("%6.2f%%%c ", completion[i], completion_stage[i]);
    }

    int sampleSets=0;
    int maxTotalSamples = 0;
    int avgTotalSamples = 0;
      
    for (i=lowerLeft[X]; i<upperRight[X]; i++){

      // for removing randomness
      if (gv->nonRandomPixels)
	ctx->mydrand48.setSeed((i+1)*(j+1));

      // initialize color
      color = SbColor(0.0, 0.0, 0.0);
      
      // adaptive super sampling
      float sum=0.0, sum2=0.0, variance=0.0, last_variance=0.0;
      float mean, mean2;
      int sampleSet=0;

      int numSamples = gv->numSamples;
      int sqrtSamples = (int)sqrt((float)numSamples); //canvas.sqrt_samples;
      int totalSamples = 0;
      int quasiSamp=256;
      
      do {
	    
	totalSamples += sqrtSamples*sqrtSamples;
	if (totalSamples>maxTotalSamples)
	  maxTotalSamples = totalSamples;

      for (int sampx = 0; sampx<sqrtSamples; sampx++){
	for (int sampy = 0; sampy<sqrtSamples; sampy++){

	  // random
	  //sample[X] = drand48();
	  //sample[Y] = drand48();
      
	  // quasi random
	  //sobvect(2, samples);
	    
	  // jittered
	  //samples[X] = (sampx + ctx->mydrand48()) / sqrtSamples;
	  //samples[Y] = (sampy + ctx->mydrand48()) / sqrtSamples;

	  // Quasi random
	  //int n = sampx*sqrtSamples+sampy; // FIXME
	  //samples[X] = n/(float)(canvas.sqrt_samples*canvas.sqrt_samples);
	  //samples[Y] = RI_vdC(n+sampleSet*canvas.sqrt_samples*canvas.sqrt_samples);
	  samples[X] = RI_S(quasiSamp);
	  samples[Y] = RI_vdC(quasiSamp);
	  quasiSamp++;

	  //problem with quasiSamp = 260, -g747,169 on scene_jensen.iv

	  // box filter
	  //samples[X] -= 0.5;
	  //samples[Y] -= 0.5;

	  // tent filter from Realistic Ray Tracing 2nd Edition, p.60
	  // transforms [0,1) -> (-1,1)
	  if (samples[X]<0.5)
	    samples[X] = sqrt(2.0*samples[X]) - 1.0;
	  else
	    samples[X] = 1.0 - sqrt(2.0 - 2.0*samples[X]);
	  if (samples[Y]<0.5)
	    samples[Y] = sqrt(2.0*samples[Y]) - 1.0;
	  else
	    samples[Y] = 1.0 - sqrt(2.0 - 2.0*samples[Y]);


// 	  pthread_mutex_lock(&mutexes[LOCK_STDERR]);
// 	  cerr << "sample [-0.5, 0.5] is: " << samples[X] << " "
//             << samples[Y] << endl;
// 	  pthread_mutex_unlock(&mutexes[LOCK_STDERR]);

	  
	  // offset calc  [0, 1)       [-0.5, 0.5)
	  // transforms (-1,1)  ->   (-0.5,1.5)
	  // (traces pixel centers)
	  offset[X] = gv->numSamples>1?samples[X]+0.5:0.5;
	  offset[Y] = gv->numSamples>1?samples[Y]+0.5:0.5;

	  // create Ray
	  r.dir = SbVec3f(

	    // sideways component
	    ( -canvas.width/2.0 +
	      ( (double)i + offset[X]) * canvas.width /
	      gv->imgDim[X] ) *
	    canvas.u +

	    // vertical component
	    ( -canvas.height/2.0 +
	      ( (double)j + offset[Y] ) * canvas.height /
	      gv->imgDim[Y] ) *
	    canvas.v +
	    
	    // distance ahead component
	    -canvas.s*canvas.w  
	    );
	  // adjust for orthographic camera
	  if (canvas.orthographic){
	    r.pos = canvas.e + r.dir + canvas.s*canvas.w;
	    r.dir = -canvas.w;
	  }

	  // depth of field
	  if (gv->aperture != 0.0){
	    
	    // get point on unit disc
	    float e1 = ctx->mydrand48()*(1.0/(float)sqrtSamples)+
	      sampx*1.0/(float)sqrtSamples;
	    float e2 = ctx->mydrand48()*(1.0/(float)sqrtSamples)+
	      sampy*1.0/(float)sqrtSamples;
	    float p = sqrt(e1);
	    float t = 2.0*M_PI*e2;

	    // convert polar (p,t) to (x,y)
	    // t is angle from right 
	    float x = p*cos(t);
	    float y = p*sin(t);

	    // turn into lens offset
	    SbVec3f offset = x*canvas.u + y*canvas.v;

	    // aperture is degees off axis of lens edge, from focal plane
	    offset *= canvas.s*tan(gv->aperture/360.0*2.0*M_PI);

	    // move eye
	    r.pos = canvas.e + offset;

	    // adjust direction
	    r.dir -= offset;
	  }

	  // debugging
// 	  if (j==71&&i==129&&sampy==0&&sampx==2)
// 	    cout << "wtf\n";
	  
	  r.dir.normalize();	    
	  
	  //////////////////////////
	  // Get color along ray
	  //////////////////////////

	  // debugging a single ray
	  if (canvas.trace){
	    DEBUG_STATE=1;
	    SbColor scolor;
	    scolor = (*this.*getRadiance)(r, canvas.flags, NULL);
	    DEBUG_STATE=0;

	    pthread_mutex_lock(&mutexes[LOCK_STDERR]);
	    print("Sample color: ");
	    printCol(scolor);
	    print("\n");
	    pthread_mutex_unlock(&mutexes[LOCK_STDERR]);

	    color += scolor;
	  }

	  // normal calculation
	  else {
	    SbVec3f ncolor;

	    // DEBUG_STATE=1;
	    //ncolor = (*this.*getRadiance)(r, canvas.flags);
	    ncolor = (this->*getRadiance)(r, canvas.flags, NULL);
	    // DEBUG_STATE=0;

	    if (0&&!finite(ncolor[0])){
	      cerr << "Retrying pixel " << i << "," << j << "\n";
		DEBUG_STATE=1;
	      ncolor = (this->*getRadiance)(r, canvas.flags, NULL);
	      DEBUG_STATE=0;
	    }
	    

// 	    // shirley aliasing example, RRT 2nd Ed., p. 51
// 	    double x = (double)i + offset[X] + 1.0;
// 	    double y = (double)j + offset[Y] + 1.0;
// 	    double func1 = 0.5*(1.0 + sin((x*x+y*y)/100.0));
// 	    ncolor = SbColor(func1, func1, func1);

	    
	    color += ncolor;

	    // variance bookkeeping
	    float result = brightness(ncolor);
	    sum += result;
	    sum2 += result*result;
	      
	    if (STOP_RENDERING){
	      if (gv->numThreads != 1)
		pthread_exit(0);
	      else
		return;
	    }
	      
	  }
	    

	} // end for (int sampy...
      } // end for (int sampx...


      // variance tracking
      mean = sum / totalSamples;
      mean2 = sum2 / totalSamples;
      variance = (mean2 - mean*mean); // /totalSamples;
      sampleSet++;

      const float t_alpha = 1.96;
      float absError = gv->maxRelError * mean;
      float n = 0.0;
      if (absError != 0.0)
	n = t_alpha*t_alpha*variance/(absError*absError);
      numSamples = (int)n - totalSamples; // subtract those already done
      numSamples = clamp(numSamples, 0, 16*gv->numSamples);

//       cerr << "absError " << absError
// 	   << " variance " << variance
// 	   << " n " << n << endl;

      sqrtSamples = (int)sqrt((float)numSamples); //canvas.sqrt_samples;      

//       cerr << setprecision(10)
// 	   << "sum " << sum
// 	   << " sum2 " << sum2
// 	   << " totalSamples " << totalSamples << endl;
//       cerr << "variance " << variance << endl;
      
      } while (gv->maxRelError &&
	       totalSamples<numSamples &&
	       sampleSet<=3);

      sampleSets += sampleSet;
      avgTotalSamples += totalSamples;

      color /= totalSamples;

      //debugging
      //color = SbColor(variance/totalSamples, 0.0, 0.0);

      // if debugging a single pixel, say color
      if (canvas.trace){
	pthread_mutex_lock(&mutexes[LOCK_STDERR]);
	print("Avg color: ");
	printCol(color);
	print("\n");
	pthread_mutex_unlock(&mutexes[LOCK_STDERR]);

	// debugging
	//setPix(i, j, color, r.t);
      }
	
      // else draw color
      else {
	setPix(i, j, color, r.t);
      }


    } // end for (i=...

    if (gv->maxRelError && ctx->id==0){
      cerr << "avg passes this row: " <<
	(float)sampleSets/(upperRight[X]-lowerLeft[X])
	   << " avg samples/pixel: " <<
	(float)avgTotalSamples/(upperRight[X]-lowerLeft[X])
	   << " max: " << maxTotalSamples
	   << "     " << endl;
    }
  } // end for (j=...

  renderTileDone:

  // completion meter
  completion[ctx->id] = 100.0;
  pthread_mutex_lock(&mutexes[LOCK_STDERR]);
  print("\r");
  for (int i=0; i<gv->numThreads; i++)
    print("%6.2f%%%c ", completion[i], completion_stage[i]);
  pthread_mutex_unlock(&mutexes[LOCK_STDERR]);

  // quit threads
  if (gv->numThreads != 1){

    // time statistics
    struct tms renderTimes;
    times(&renderTimes);
    ctx->ticks += renderTimes.tms_utime + renderTimes.tms_cutime;
    
    pthread_exit(0);
  }
  
  return;
}

//'WW''''''''''''''''''WW''''WW''''''''''
// WW                  WWW  WWW          
//WWWW  WW   WWW   WW  W WWW WW  WWW WWW 
// WW  WW W  WW W WWWW W  W  WW W WW WW W
// WW  WW W  WW W WW   W     WW W WW WW W
// WW   WW   WW W  WWW W     WW  WWW WWW 
//                                   WW  
//'''''''''''''''''''''''''''''''''''WW''
//
// Tone maps the image, turning radiance
// values into illuminance ones, then finally
// into image pixel colors.
//
// The algorithm used is described in
// Peter Shirley's Realistic Ray Tracing 1st Ed. Ch. 18 p. 152.
// Radiance values are turned into illuminance values,
// dark values are de-saturated to simulate the human
// visual system's b&w night vision, and Greg Ward's
// contrast based scalefactor is used to compress the illuminance
// values into the displayable range.
//
// I tried my best to choose physically accurate primaries.
void Pane::toneMap(){

//   Values in pix[] are W/m^2-sr in three channels.
//
//   What spectra are these channels? Good question. First I take
//   the matrices for RGB->XYZ (and the inverse) from www.srgb.com.
  

  // Tone mapping algorithm is taken from
  // Realistic Ray Tracing by Peter Shirley page 152
       
  SbColor *c;
  double r, g, b;
    
  // tone mapping variables
  const double k = 683.0;  // k constant (lm/W) (Peter Shirley)
  double V;
  // const double Yw = 100.0; // maximum display luminance (cd/m^2) (PS)
  const double Yw = 110.5; // maximum display luminance (cd/m^2)
  // const double Yw = 200.0; // maximum display luminance (cd/m^2) (cinema)
  double Yi = 0.0;         // world adaptation luminance
  double cY;       // tristimulous values
  double x, y, W, s;       // misc
  //const double l[3] = { 620.0, 545.0, 465.0 }; // lambda (nm)
  const double dl[3] = { 1.0, 1.0, 1.0 }; // dlambda
  int numpix = 0;

  // From Chris R. Schoenenman
  // http://research.microsoft.com/~hollasch/cgindex/color/freq-rgb.html
  // Values for xbar, byar, zbar

//   const double xb[3] = { 0.8563, 0.4516, 0.2541 }; // xbar
//   const double yb[3] = { 0.3981, 0.9822, 0.1528 }; // ybar
//   const double zb[3] = { 0.0000, 0.0079, 1.5549 }; // zbar
//   const double vb[3] = { 0.0332, 0.4810, 0.3281 }; // vbarn



//   // from apple cinema display numbers
//   // www.cis.rit.edu/research/mcsl/research/PDFs/LCD.pdf
//   // computed by dividing matrix given by highest element
//   // so that components are relative.
//   const double l[3] = { 618.0, 555.0, 485.0 }; // lambda (nm)
//   const double xb[3] = { 0.7863, 0.4772, 0.2265 }; // xbar
//   const double yb[3] = { 0.4103, 1.0000, 0.1638 }; // ybar
//   const double zb[3] = { 0.0142, 0.1197, 1.2792 }; // zbar
//   const double vb[3] = { 0.00862, 0.402, 0.851 }; // vbar

  
  // sRGB standard, www.srgb.com
  // these are not real single-wavelength primaries
  // however, they have similar photopic responses
  // to real wavelengths.
  // i.e., red causes a red tristimulous response
  // with the same luminance (Y) as a wavelength of 635nm,
  // but with different chromacity (X, Z).

  // Each of "r", "g", "b" do not have a single wavelength
  // equivalent. Instead, each causes a response
  // according to the table below, yanked from www.srgb.com.
  // While not exactly realistic, the red channel causes
  // a red response, and similarly for green and blue.
  // Additionally, RGB(1,1,1) produces XYZ(.95,1,1.089)

  // In our RGB space, the r coordinate, expressed
  // in W/m^2/sr, is not the power of a single (red)
  // wavelength of light, but rather the power of a
  // distribution evoking a red response. For example,
  // a value of 1.0 evokes a response of XYZ(.4124,.2126,.0193),
  // or, said another way, .4124*(the X stimuli,


//  const double l[3] = { 635.0, 520.0, 466.0 }; // lambda (nm)


//   const double xb[3] = { 0.4124, 0.3576, 0.1805 }; // xbar
//   const double yb[3] = { 0.2126, 0.7152, 0.0722 }; // ybar
//   const double zb[3] = { 0.0193, 0.1192, 0.9505 }; // zbar
//   const double vb[3] = { 0.00862, 0.402, 0.851 }; // vbar


  // dominant wave lengths
  // const double l[3] = { 635.0, 520.0, 466.0 }; // lambda (nm)
  
  // primary spectra
  //const double l[3] = { 611.0, 549.0, 460.0 }; // lambda (nm)

  // xyz functions scaled by (0.595, 1.0, 0.85) to make white
//   const double xb[3] = { 0.5899, 0.4183, 0.2472 }; // xbar
//   const double yb[3] = { 0.2918, 0.9928, 0.0510 }; // ybar
//   const double zb[3] = { 0.0002, 0.0095, 1.4188 }; // zbar
//   const double vb[3] = { 0.00862, 0.402, 0.851 }; // vbar


  const double l[3] = { 611.0, 549.0, 467.0 }; // lambda (nm)
  
  // note: these multipliers highly dependent on the white point (xw, yw)!!
//   // this is centered about the sRGB white point but it
//   // desaturates the blue wall of jensen's cornell box
//   const double xw=0.3127, yw=0.3290;
//   const double d = 0.825;
//   const double e = 1.0;
//   const double f = 0.985;

//   // no white-point centering
//   const double xw=0.3333, yw=0.3333;
//   const double d = 0.65;
//   const double e = 0.94;
//   const double f = 1.0;

//   // responses taken from interpolated CIE 1931 xb yb zb data (wyszecki stiles)
//   const double xb[3] = { 0.9914*d, 0.4183*e,    0.2408*f }; // xbar
//   const double yb[3] = { 0.4905*d, 0.9928*e,    0.0770*f }; // ybar
//   const double zb[3] = { 0.0003*d, 0.0095*e,    1.4861*f }; // zbar 466
//   const double vb[3] = { 0.01477, 0.497, 0.631 }; // vbar


  //const double xw=0.3127, yw=0.3290;
  const double xw=0.3333, yw=0.3333;
  // scale fake primaries to match brightness (luminance) of
  // balanced real primaries: xb,yb,zb *= sum(real_yb)/sum(fake_yb)
  const double d = (0.4905*0.65 + 0.9928*0.94 + 0.0770*1.0) /
                   (0.2126+0.7152 +0.0722);  // 1.3291
  const double e = d;
  const double f = d;
  
  const double xb[3] = { 0.4124*d, 0.3576*e, 0.1805*f }; // xbar
  const double yb[3] = { 0.2126*d, 0.7152*e, 0.0722*f }; // ybar
  const double zb[3] = { 0.0193*d, 0.1192*e, 0.9505*f }; // zbar
  const double vb[3] = { 0.01477, 0.497, 0.631 }; // vbar

  // use to convert XYZ -> to my RGB (i.e., inv([xb,yb,zy]*d))
  const double inv[3][3] =
    {
      {  2.4383,   -1.1566,   -0.3752 },
      { -0.7290,    1.4113,    0.0312 },
      {  0.0419,   -0.1535,    0.7953 }
    };

//   const double xb[3] = { 0.6742, 0.4183, 0.2297 }; // xbar
//   const double yb[3] = { 0.3335, 0.9928, 0.0474 }; // ybar
//   const double zb[3] = { 0.0002, 0.0095, 1.3187 }; // zbar
//   const double vb[3] = { 0.00862, 0.402, 0.851 }; // vbar

  // Chromaticity values from Peter Shirley
  //const double xw = 0.283, yw = 0.298; // (xw, yw) for trinitron
  //const double xw=0.3328, yw=0.3516; // cinema display
  // my own chromacity values
  //const double xw = 0.34, yw = 0.37; // (xw, yw) for trinitron
  //const double xw = 0.284, yw = 0.3; // (xw, yw) for trinitron
  

  // convert r g b to X Y Z
  // calculate Yi, log average
  for (int i=0; i<gv->imgDim[X]*gv->imgDim[Y]; i++){

    SbColor c = pix[i];
    
      // make sure pixel colors are not inf or nan
      if ( !finite(c) ){
	int j = i/gv->imgDim[X];
	int ii = i%gv->imgDim[X];
	cerr << "Pixel at (" << ii << ", " << j << ") is invalid: " << c << endl;

	if (!finite(c[R]))
	  c[R] = 0.0;
	if (!finite(c[G]))
	  c[G] = 0.0;
	if (!finite(c[B]))
	  c[B] = 0.0;
      }	  

      cY = k * ( yb[R]*c[R]*dl[R] + yb[G]*c[G]*dl[G] + yb[B]*c[B]*dl[B] );

      // minimum value of cY taken from Tumblin Rushmeier '93
      if (cY>3.18e-7 && pixDepth[i]>0.0){
	Yi += log10(cY);
	numpix++;
      }
  }

  // Log average of Y idea of Jack Tumblin and Holly Rushmeier
  // finish calculating Yi
  if (numpix)
    Yi /= numpix;
  Yi += 0.84;           // from Tumblin Rushmeier '93
                        // log10(Lw) = E{log10(Lin)}+0.84
  Yi = pow(10.0, Yi);

  // use user specified Yi
  if (gv->luminanceMean>0.0)
    Yi = gv->luminanceMean;
  
  lastImageYi = Yi;

  cerr << "Log average luminance = " << Yi << " cd/m^2\n";

  // This scalefactor maps Yi to Yw/2.0, then divides by Yw
  // to map Yw to 1.0. This works in conjunction with
  // the XYZ->RGB matrix later, which maps
  // a luminance of 1.0 to a physical luminance of Yw.
  //
  // Example input: Yi -> Yw/2 -> 0.5 -> Yw/2 physical output
  double sf = 1.0/Yw * pow( ( 1.219 + pow(Yw/2.0, 0.4) ) /
			    ( 1.219 + pow(Yi, 0.4) ), 2.5);
  cerr << "Contrast-based scalefactor = " << sf << endl;

  for (int i=0; i<gv->imgDim[X]*gv->imgDim[Y]; i++){

    SbColor c = pix[i]*gv->exposure;
    SbColor L = c;

      // do tone mapping
      if (gv->toneMapping){

	// convert rgb to XYZ
	L[X] = k * ( xb[R]*c[R]*dl[R] + xb[G]*c[G]*dl[G] + xb[B]*c[B]*dl[B] );
	L[Y] = k * ( yb[R]*c[R]*dl[R] + yb[G]*c[G]*dl[G] + yb[B]*c[B]*dl[B] );
	L[Z] = k * ( zb[R]*c[R]*dl[R] + zb[G]*c[G]*dl[G] + zb[B]*c[B]*dl[B] );
      
	V = 1700.0 * ( vb[R]*c[R]*dl[R] + vb[G]*c[G]*dl[G] + vb[B]*c[B]*dl[B] );
      
	if (L[Y]<3.18e-7)
	  L[Y] = 3.18e-7;

	// choose desaturation depending on brightness (human visual system modeling)
	if (log10(L[Y]) < -2.0)
	  s = 0.0;
	else if (log10(L[Y]) < 0.6){
	  double f = (log10(L[Y])+2.0)/2.6;
	  s = 3.0*f*f -  2.0*f*f*f;
	}
	else
	  s = 1.0;

	// convert to Yxy
	W = L[Y] + L[X] + L[Z];
	x = L[X]/W;
	y = L[Y]/W;

	// desaturate
	x = (1.0-s)*xw + s*(x+xw-0.3333);
	y = (1.0-s)*yw + s*(y+yw-0.3333);
	L[Y] = 0.4468*(1.0-s)*V + s*L[Y];
	
	// scale luminance
	// after this scaling, 1.0 is really Yw (the real monitor brightness) in real life
	L[Y] *= sf;

	// store transformed colors (Yxy)
	pix[i][X] = x;
	pix[i][Y] = L[Y];
	pix[i][Z] = y;
      }
  }

  // convert Yxy ->rgb
  // two methods:
  // glare, and no glare

  // with glare  
  if (gv->useGlare)
    // glareFilter convolves a model of the eye's point-spread function
    // against all above a given thresh hold.
    // in this case, the threshhold is the monitor's maximum luminance,
    // after scaling, which is Yw*sf.
    //
    // This is really convoluted:
    // rgb->XYZ->Yxy->scale (tonemap)->
    //        glare filter->rgb->combine with glare->gamma correct->rgb
    glareFilter(Yi, Yw*sf); 
  
  // without glare
  else {

    for (int i=0; i<gv->imgDim[X]*gv->imgDim[Y]; i++){
  
      SbColor L = pix[i];

      // transform Yxy -> XYZ
      float x=L[X];
      float y=L[Z];
      L[X] = x*L[Y]/y;
      L[Z] = (1.0-x-y)*L[Y]/y;

      // sRGB standard, http://www.w3.org/Graphics/Color/sRGB
      // physical sRGB standard, based on Yw=110.5
      pix[i][R] =  3.2406*L[X] +  -1.5372*L[Y] +  -0.4986*L[Z];
      pix[i][G] = -0.9689*L[X] +   1.8758*L[Y] +   0.0415*L[Z];
      pix[i][B] =  0.0557*L[X] +  -0.2040*L[Y] +   1.0570*L[Z];

      // Technically this matrix is off by a factor of 1/Yw.
      // The given matrix expects a maximum luminance (Y)
      // of 1.0. This is fine since our scale factor
      // scales everything so the maximum luminance is 1.
      //
      // If we weren't using a scalefactor to scale our
      // luminances, and we wanted to convert (X,Y,Z)
      // to the actual (r,g,b) signal that produces
      // that exact stimuli, then you need to use
      // the above matrix and then divide each of
      // the r,g,b by Yw (or divide matrix by Yw).
      //
      // For example:
      //
      // r /= Yw;
      // g /= Yw;
      // b /= Yw;      
    }
  }

  // gamma correct
  for (int i=0; i<gv->imgDim[X]*gv->imgDim[Y]; i++){

      r = pix[i][R];
      g = pix[i][G];
      b = pix[i][B];

      // gamma correct
      r = clamp(r, 0.0, 1.0);
      g = clamp(g, 0.0, 1.0);
      b = clamp(b, 0.0, 1.0);

      if (fabs(gv->gamma-2.2)>0.001){

	// regular gamma correction
	r = pow(r, 1.0/gv->gamma);
	g = pow(g, 1.0/gv->gamma);
	b = pow(b, 1.0/gv->gamma);
      }
      else if (fabs(gv->gamma-1.0)>0.001){

	// sRGB standard gamma correction (similar to regular w/2.2)
	r = r<=0.0031308 ? 12.92*r : 1.055*pow(r, 1.0/2.4) - 0.055;
	g = g<=0.0031308 ? 12.92*g : 1.055*pow(g, 1.0/2.4) - 0.055;
	b = b<=0.0031308 ? 12.92*b : 1.055*pow(b, 1.0/2.4) - 0.055;

      }

      // store transformed colors
      pix[i].setValue(r,g,b);
  }

  toneMapped = 1;
  return;
}



//'''''''''''''WW''WW'''''''WW''W'WWWW''WWWW'
//                 WW       WW  W WW  W WW  W
//WW    W W WW WW WWWW  WW  WW  W WW  W WW  W
//WW WW W WW   WW  WW  WWWW WWWWW WW  W WWWW 
//WW WW W WW   WW  WW  WW   WW  W WW  W WW  W
// WW WW  WW   WW  WW   WWW WW  W WWWW  WW  W
//                                           
//'''''''''''''''''''''''''''''''''''''''''''

void Pane::writeHDR(){

  string comment = createComment();

  char *fn = new char[strlen(gv->imgFile)+100];
  char *lastPeriod = strrchr(gv->imgFile, '.');
  if (lastPeriod){
    int len = lastPeriod - gv->imgFile;
    strncpy(fn, gv->imgFile, len);
    sprintf(fn+len, ".hdr");
  }

  // or afterwards if no extension found
  else 
    sprintf(fn, "%s.hdr", gv->imgFile);

  checkForOverwrite(fn);

  // write .hdr
  FILE *f = fopen(fn, "wb");
  RGBE_WriteHeader(f, gv->imgDim[X], gv->imgDim[Y], NULL);
  for (int j=gv->imgDim[Y]-1; j>=0; j--)
    for (int i=0; i<gv->imgDim[X]; i++)
      RGBE_WritePixels(f, pix[j*gv->imgDim[X]+i].getValue(), 1);
  fclose(f);
}



//'''''''''''''''''WW'WW''W'WWWW''WWWW'
//                 WW WW  W WW  W WW  W
//W WW  WW   WWW  WWW WW  W WW  W WW  W
//WW   WWWW W WW W WW WWWWW WW  W WWWW 
//WW   WW   W WW W WW WW  W WW  W WW  W
//WW    WWW  WWW  WWW WW  W WWWW  WW  W
//                                     
//'''''''''''''''''''''''''''''''''''''

void Pane::readHDR(){

  cerr << "Reading " << gv->inImgFile << " into framebuffer.\n";
  
  // read file
  FILE *f = fopen(gv->inImgFile, "rb");
  if (!f){
    cerr << "Error reading " << gv->inImgFile << ": " << strerror(errno) << endl;
    return;
  }
    
  RGBE_ReadHeader(f, &gv->imgDim[X], &gv->imgDim[Y], NULL);
  float *image = (float *)malloc(sizeof(float)*3*gv->imgDim[X]*gv->imgDim[Y]);
  RGBE_ReadPixels_RLE(f, image, gv->imgDim[X], gv->imgDim[Y]);
  fclose(f);

  // allocate new pix
  delete [] pix;
  delete [] pixDepth;
  pix = new SbColor[gv->imgDim[X]*gv->imgDim[Y]]; // should i initialize this?
  pixDepth = new float[gv->imgDim[X]*gv->imgDim[Y]];
  
  // copy into frame buffer
  int k=0;
  for (int j=gv->imgDim[Y]-1; j>=0; j--)
    for (int i=0; i<gv->imgDim[X]; i++){
      pix[j*gv->imgDim[X]+i].setValue(&image[(k++)*3]);
      pixDepth[j*gv->imgDim[X]+i] = 0.1;
    }

  delete image;
}

// read filter image
int Pane::readGlareFilter(char *fn, int *fdim){

  //char fn[] = "filter.hdr";

  cerr << "Reading " << fn << " into memory.\n";
  
  // read file
  FILE *f = fopen(fn, "rb");
  if (!f){
    cerr << "Error reading " << fn << ": " << strerror(errno) << endl;
    return(0);
  }
    
  RGBE_ReadHeader(f, &(fdim[X]), &(fdim[Y]), NULL);
  float *image = (float *)malloc(sizeof(float)*3*fdim[X]*fdim[Y]);
  RGBE_ReadPixels_RLE(f, image, fdim[X], fdim[Y]);
  fclose(f);

  // allocate new pix
  filtpix = new SbColor[fdim[X]*fdim[Y]]; // should i initialize this?
  
  // copy into frame buffer
  int k=0;
  for (int j=fdim[Y]-1; j>=0; j--)
    for (int i=0; i<fdim[X]; i++)
      filtpix[j*fdim[X]+i].setValue(&image[(k++)*3]);

  return(1);
}



//'''''WW'''''''''''''WW''''WWWWW'''''''''''''WWW'''''''''''''''''''''''''''''
//     WW             WW    WW               WW  W                            
// WWW WWW   WW   WWW WW W  WWWW   WW   W WW WW  W WW W  WW  W WW WW    W W WW
//WW W WW W WWWW WW W WWW   WW    WW W  WW   WW  W WW W WWWW WW   WW WW W WW  
//WW   WW W WW   WW   WW W  WW    WW W  WW   WW  W WW W WW   WW   WW WW W WW  
// WWW WW W  WWW  WWW WW  W WW     WW   WW    WWW   WW   WWW WW    WW WW  WW  
//                                                                            
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

void Pane::checkForOverwrite(char *filename){

    // back up old file if it exists
  struct stat file_stat;
  if (stat(filename, &file_stat)==0 && gv->createBackup){

    // find unused filename
    char *fn = new char[strlen(filename)+100];
    int i;
    for (i=1;;i++){

      // insert number before extension
      char *lastPeriod = strrchr(filename, '.');
      if (lastPeriod){
	int len = lastPeriod - filename;
	strncpy(fn, filename, len);
	sprintf(fn+len, ".%03d%s", i, lastPeriod);
      }

      // or afterwards if no extension found
      else 
	sprintf(fn, "%s.%03d", filename, i);
      
      if (stat(fn, &file_stat))
	break;
    }

    // warn if many backups
    if (i>50)
      cerr << "Warning: excessive backups. You should probably delete some."
	   << endl;

    // copy file
    char *f = new char[strlen(filename)*2+60];
    sprintf(f, "mv %s %s", filename, fn);
    cerr << "Backing up old file..." << flush;
    cerr << f << endl;
    system(f);
      
    delete [] fn;
    delete [] f;
  }
}



//'''''''''''''''''''''WW''''''''WWW'''''''''''''''''''''''''''''''''''WW'
//                     WW       WW  W                                  WW 
// WWW W WW  WW   WWW WWWW  WW  WW     WW   WWWW W  WWWW W   WW  WWW  WWWW
//WW W WW   WWWW W WW  WW  WWWW WW    WW W  WW WW W WW WW W WWWW WW W  WW 
//WW   WW   WW   W WW  WW  WW   WW  W WW W  WW WW W WW WW W WW   WW W  WW 
// WWW WW    WWW  WWW  WW   WWW  WWW   WW   WW WW W WW WW W  WWW WW W  WW 
//                                                                        
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

string Pane::createComment(){

  // get hostname
  char machinename[100];
  gethostname(machinename, 100);
  
  // revision info
  const char *revision = getVersion();
  
  // create comment
  ostringstream comment;
  
  int totalRenderTime = gv->times.octreeCreation + gv->times.photonShooting +
    gv->times.render + gv->times.photonBalancing + gv->times.photonIrradiance;
  float renderCPUMs = gv->times.renderTicks / sysconf(_SC_CLK_TCK) * 1000.0;
  comment << "# Pane v" << revision << " by Kevin Beason\n";
  comment << "# http://www.csit.fsu.edu/~beason/pane/\n";
  comment << "# " << gv->numSamples << " samples, "
	  << gv->lightSamples << " light samples, "
	  << gv->numThreads << " threads\n";
  comment << "# Octree time: "      << gv->times.octreeCreation << "ms "
	  << " Photon shooting: "   << gv->times.photonShooting << "ms "
	  << " Photon balancing: "  << gv->times.photonBalancing << "ms "
	  << " Photon irradiance: " << gv->times.photonIrradiance << "ms "
	  << " Rendering: "         << gv->times.render << "ms\n";
  comment << "# Rendering CPU time (CPU time): " << renderCPUMs << "ms / ";
  comment << renderCPUMs/1000.0 << "s / ";
  comment << renderCPUMs/1000.0/60.0 << "m / ";
  comment << renderCPUMs/1000.0/60.0/60.0 << "h, ";
  comment << "Ticks: " << gv->times.renderTicks << " ";
  comment << "Ticks/s: " << sysconf(_SC_CLK_TCK) << "\n";
  comment << "# Total render time (actual time): " << totalRenderTime << "ms / ";
  comment << totalRenderTime/1000.0 << "s / ";
  comment << totalRenderTime/1000.0/60.0 << "m / ";
  comment << totalRenderTime/1000.0/60.0/60.0 << "h\n";
  comment << "# on " << machinename << "\n";
  if (gv->toneMapping){
    comment << "# Yi = " << lastImageYi
	<< " (Log average luminance used for tone mapping)\n";
  }
  else
    comment << "# (No tone mapping, though Yi = " << lastImageYi << ")\n";
  comment << "# comment: " << gv->comment << "\n";
  comment << "# " << gv->cmdLine;

  return(comment.str());
}

//'''''''''''''WW''WW'''''''WWWW'''''''''''''''WW'WW'
//                 WW        WW               WW   WW
//WW    W W WW WW WWWW  WW   WW  WWWW W   WWW WW   WW
//WW WW W WW   WW  WW  WWWW  WW  WW WW W W WW WW   WW
//WW WW W WW   WW  WW  WW    WW  WW WW W W WW WW   WW
// WW WW  WW   WW  WW   WWW WWWW WW WW W  WWW  WW WW 
//                                         WW        
//'''''''''''''''''''''''''''''''''''''''WWW'''''''''
//
// Writes image to disk in PNG format.
//
void Pane::writeImg(){

  // if configured, output iv file instead
  if (gv->renderIv){
    writeIv();
    return;
  }

  cerr << "\nWriting " << gv->imgFile << " ...";
  ofstream out;

  // for matting
  if (gv->makeMatte){
    for (int i=0; i<gv->imgDim[0]*gv->imgDim[1]; i++)
      if (pixDepth[i]<0.0)
	pix[i] = gv->worldBG;
  }
  
  // write hdr
  if (!gv->inImgFile)
    writeHDR();
  
  // back up old file if it exists
  checkForOverwrite(gv->imgFile);

  // tone map the image
  if (toneMapped == 0)
    toneMap();

  // create comment
  string comment = createComment();
  
  // Old PPM and PFM code
//   // open file
//   out.open(gv->imgFile);
//   if (!out){
//     cerr << "Error writing file " << gv->imgFile << endl;
//     return;
//   }
//  
//   // print PPM identifier
//   if (gv->floatMap)
//     out << "Pf\n";
//   else
//     out << "P6\n";
//   out << comment.str() << "\n";
//   out << gv->imgDim[X] << " " << gv->imgDim[Y] << "\n";
//   if (!gv->floatMap)
//     out << 255 << "\n";
//  
//   // english reading order output
//   SbColor *c;
//   double r, g, b;
//   for (int j=gv->imgDim[Y]-1; j>=0; j--){
//     for (int i=0; i<gv->imgDim[X]; i++){
//       c = &pix[gv->imgDim[X]*j+i];
//       r = (*c)[R];
//       g = (*c)[G];
//       b = (*c)[B];
//
//       if (gv->floatMap)
// 	out << r << " "
// 	    << g << " "
// 	    << b << "\n";
//       else {
// 	unsigned char rc, gc, bc;
// 	rc = (unsigned char) clamp((int)(r*255.0), 0, 255);
// 	gc = (unsigned char) clamp((int)(g*255.0), 0, 255);
// 	bc = (unsigned char) clamp((int)(b*255.0), 0, 255);
//
// 	out.put(rc);
// 	out.put(gc);
// 	out.put(bc);
//       }
//      
//     }
//   }
//  
//   out.close();
//   cerr << "done." << endl;

  // write PNG
  // source from http://www.libpng.org/pub/png/libpng-1.2.5-manual.html#section-4

  // open file
  FILE *fp = fopen(gv->imgFile, "wb");
  if (!fp){
    cerr << "Error writing file " << gv->imgFile << endl;
    return;
  }

  // create png_struct and png_info
  png_structp png_ptr = png_create_write_struct
    (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr)
    return;

  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr){
    png_destroy_write_struct(&png_ptr,
			     (png_infopp)NULL);
    cerr << "Error creating PNG data structures." << endl;
    return;
  }

  // set up error handling. PNG will expect to longjmp() here
  if (setjmp(png_jmpbuf(png_ptr))){
       png_destroy_write_struct(&png_ptr, &info_ptr);
       fclose(fp);
    cerr << "PNG creation error." << endl;
    return;
  }

  // set up png output
  png_init_io(png_ptr, fp);
  png_set_IHDR(png_ptr, info_ptr, gv->imgDim[X],
	       gv->imgDim[Y], 8, PNG_COLOR_TYPE_RGB,
	       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
	       PNG_FILTER_TYPE_DEFAULT);

  // convert comment to c style string
  int comment_size = comment.size();
  char commentstr[comment_size+1];
  strcpy(commentstr, comment.c_str());
  
  // create comment
  png_text text_ptr[1];
  text_ptr[0].key = "Comment";
  text_ptr[0].text = commentstr;
  text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
  png_set_text(png_ptr, info_ptr, text_ptr, 1);

  // set rendering intent
  //png_set_sRGB(png_ptr, info_ptr, PNG_sRGB_INTENT_ABSOLUTE);
  
  // set gamma
  png_set_gAMA(png_ptr, info_ptr, 1.0/gv->gamma);

  // set chromacity info (values from srgb.com)
  png_set_cHRM(png_ptr, info_ptr,
	       .3127, .3290,        // white point x, y
	       .64, .33,            // r 
	       .15, .06,            // g
	       .3127, .3290);       // b

  // create array of bytes
  png_byte *image = new png_byte[gv->imgDim[Y]*gv->imgDim[X]*3];
  png_bytep row_pointers[gv->imgDim[Y]];
  SbColor *c;
  double r, g, b;
  for (int j=0; j<gv->imgDim[Y]; j++){
    for (int i=0; i<gv->imgDim[X]; i++){
      c = &pix[gv->imgDim[X]*j+i];
      
      r = (*c)[R];
      g = (*c)[G];
      b = (*c)[B];
      
      unsigned char rc, gc, bc;
      rc = (unsigned char) clamp((int)(r*255.0), 0, 255);
      gc = (unsigned char) clamp((int)(g*255.0), 0, 255);
      bc = (unsigned char) clamp((int)(b*255.0), 0, 255);

      image[j*(gv->imgDim[X]*3)+i*3+0] = rc;
      image[j*(gv->imgDim[X]*3)+i*3+1] = gc;
      image[j*(gv->imgDim[X]*3)+i*3+2] = bc;
    }
  }
  for (int k=0; k<gv->imgDim[Y]; k++)
    row_pointers[k] = &image[(gv->imgDim[Y]-k-1)*(gv->imgDim[X]*3)];

  // write image
  png_write_info(png_ptr, info_ptr);
  png_write_image(png_ptr, row_pointers);
  png_write_end(png_ptr, info_ptr);

  // clean up
  png_destroy_write_struct(&png_ptr, &info_ptr);
  fclose(fp);
  cerr << "done.\n";
  delete [] image;
  
  if (!gv->renderQuit){
    char *f = new char[strlen(gv->imgFile)+20];
    f[0] = '\0';

    // choose appropriate viewer
    if (system("/usr/bin/which xv >& /dev/null") == 0)
      strcat(f, "xv ");
    else if (system("/usr/bin/which display >& /dev/null") == 0)
      strcat(f, "display ");
    else {
      cerr << "No appropriate viewer found!\n"
	   << "Try copying ~beason/linux/bin/xv into your path.\n";
      delete f;
      return;
    }
    
    strcat(f, gv->imgFile);
    strcat(f, " &");
    system(f);
    delete f;
  }
}


//''''''WW''''''''''''''''WW''WWWW''WW''''''''''WW''''''''''''''''''''WW'WW'
//      WW                WW  WW  W WW          WW                   WW   WW
// WWWW WWW   WW    WW   WWWW WW  W WWW   WW   WWWW  WW   WWW   WWWW WW   WW
//WWW   WW W WW W  WW W   WW  WWWW  WW W WW W   WW  WW W  WW W WWW   WW   WW
//  WWW WW W WW W  WW W   WW  WW    WW W WW W   WW  WW W  WW W   WWW WW   WW
//WWWW  WW W  WW    WW    WW  WW    WW W  WW    WW   WW   WW W WWWW   WW WW 
//                                                                          
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Starts threads to shoot photons from the lights into the scene.
// Each thread calls emitPhotons(), which in turn calls bouncePhoton().
//
void Pane::shootPhotons(){

  if (!gv->sceneHasLights){
    cerr << "Scene has no lights: not shooting any photons.\n";
    return;
  }
  
  int logNumToFire = -1;
  float numToFire;
  photon_t phot_type = GLOBAL;
  MemberFunction *memberFunc = (MemberFunction *)alloca(
    sizeof(MemberFunction)*gv->maxThreads); // used to pass object pointer
  thread_context *ctx = (thread_context *)alloca(
    sizeof(thread_context)*gv->maxThreads);   // random number generators
  
  // start timer
  struct timeval start, end;
  cerr << "Shooting photons...\n" << flush;
  gettimeofday(&start, NULL); // start delay measure

  // adjust max emissive avg for environment map
  if (envMap){
    double emisAvg = (envMap->Lavg[0] + envMap->Lavg[1] + envMap->Lavg[2])/3.0
      * M_PI * gv->meters2(tritree.radius*tritree.radius) *4.0*M_PI;
    totalEmisAvg += emisAvg;
    if (emisAvg > maxEmisAvg)
      maxEmisAvg = emisAvg;
  }

  // adjust max emissive avg for point lights
  vector<SoPointLight *>::iterator pI = litPnts.begin();
    
  if (gv->headLightPos == HL_OFF)
    pI++;
      
  for (; pI != litPnts.end(); pI++){

    // get emission
    double pntEmisAvg = 0.0;
    SbColor emission = (*pI)->color.getValue()*(*pI)->intensity.getValue();
    for (int i=0; i<3; i++)
      pntEmisAvg += emission[i];
    pntEmisAvg /= 3.0;

    totalEmisAvg += pntEmisAvg;
    if (pntEmisAvg > maxEmisAvg)
      maxEmisAvg = pntEmisAvg;
  }
    
  
  // position headlight
  findCamera();
  configureHeadlight();

  // set up canvas
  canvasSetup();
  
  // pack object pointer, function pointer, and argument into struct
  for (int i=0; i<gv->numThreads; i++){
    memberFunc[i].object = this;
    memberFunc[i].pmf2 = &Pane::emitPhotons;
    memberFunc[i].ctx = &ctx[i];
    ctx[i].mydrand48 = RNG(lrand48());
    ctx[i].id = i;
    ctx[i].rays = 0;
    ctx[i].ticks = 0;
  }

  // dont bother with caustic photons if rendering an approximation
  if (gv->approximate)
    gv->numCPhotons = 0;
  else if (gv->algorithm != 1 ||
           gv->renderFlags && (gv->renderFlags & C_MDIFFUSE) == 0){
    gv->numPhotons = 0;
  }

  int importance_done = 0;
  int global_done = 0;
  while (gPhotonMap->get_stored_photons()<gv->numPhotons ||
         cPhotonMap->get_stored_photons()<gv->numCPhotons ||
	 importanceMap->get_stored_photons()<gv->numIPhotons ){

    if (!importance_done && importanceMap->get_stored_photons()>=gv->numIPhotons){
      cerr << "\n" << importanceMap->get_stored_photons()
	   << " photons stored in importance map.\n";
      cerr << "Importance map done. Continuing to other maps...\n";
      logNumToFire = -1;

      // scale photon power
      float proj_area = M_PI * gv->meters2(tritree.radius*tritree.radius);      
      float scale = proj_area/importanceMap->get_stored_photons();
      importanceMap->scale_photon_power(scale);

      // balance
      cerr << "Balancing importance photon map...";
      TIME(importanceMap->balance(gv->iPhotonEstimate));

      // precompute importance
//       cerr << "Precomputing importance for importance photon map...\n";
//       gettimeofday(&start, NULL); // start delay measure
//       importanceMap->precompute_irradiance(gv->iPhotonEstimate,
// 					   gv->numThreads,
// 					   gv->showPhotons);
      
      importance_done = 1;
    }
    
    if (!global_done && gPhotonMap->get_stored_photons()>=gv->numPhotons){
      cerr << "\n" << gPhotonMap->get_stored_photons()
	   << " photons stored in global map.\n";
      cerr << "Global map done. Continuing to build caustic map...\n";
      global_done = 1;
    }

    if (!importance_done)
      phot_type = IMPORTANCE;
    else {
      
      // cut power in map in half
      cPhotonMap->scale_photon_power(0.5);
      
      if (global_done)
	phot_type = CAUSTIC;
      else {
	phot_type = GLOBAL;
	gPhotonMap->scale_photon_power(0.5);
      }
      totalPower /= 2.0;
    }

    logNumToFire++;
    numToFire = powf(2.0, (float)logNumToFire);
    //cerr << "numToFire=" << numToFire << "\n";
    
    // launch threads
    for (int i=0; i<gv->numThreads; i++){

      // call wrapper function with struct
      memberFunc[i].numToFire = numToFire;
      memberFunc[i].phot_type = phot_type;
      pthread_create(&gv->threads[i], NULL, Pane::emitPhotonsWrapper,
		     (void*)&memberFunc[i]);
    }

    // wait for threads to finish and flush buffers
    for (int i=0; i<gv->numThreads; i++){
      pthread_join(gv->threads[i], NULL);
      photonRaysFired += ctx[i].rays;
      gv->times.totalTicks += ctx[i].ticks;

      cPhotonMap->buffered_store(NULL, NULL, NULL, NULL, 0, 0, i, 1);
      gPhotonMap->buffered_store(NULL, NULL, NULL, NULL, 0, 0, i, 1);
      importanceMap->buffered_store(NULL, NULL, NULL, NULL, 0, 0, i, 1);
    }

  }

  gettimeofday(&end, NULL);   // end delay measure
  gv->times.photonShooting = delay(start, end);
  cerr << "\n" << firedPhotons << " total photons fired, for "
       << setiosflags(ios::showpoint | ios::fixed) << setprecision(5)
       << totalPower << "W (" << delay(start, end) << "ms)" << endl;
  cerr << gPhotonMap->get_stored_photons()
       << " photons stored in global map.\n";
  cerr << cPhotonMap->get_stored_photons()
       << " photons stored in caustic map.\n";
  if (gPhotonMap->get_stored_photons()>=max_photons)
    cerr << " WARNING: photon map full, some photons probably not stored!\n";
  cerr << flush;
}


//'''''''''''''WW''WW''WWWW''WW''''''''''WW''''''''''''''''''
//                 WW  WW  W WW          WW                  
// WW  WWWW W  WW WWWW WW  W WWW   WW   WWWW  WW   WWW   WWWW
//WWWW WW WW W WW  WW  WWWW  WW W WW W   WW  WW W  WW W WWW  
//WW   WW WW W WW  WW  WW    WW W WW W   WW  WW W  WW W   WWW
// WWW WW WW W WW  WW  WW    WW W  WW    WW   WW   WW W WWWW 
//                                                           
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Loops over lights, shooting out photons from each.
// Uses an advanced Quasi-Monte-Carlo technique called
// Trajectory Splitting, described by Kollig and Keller
// in their paper Efficient Multidimensional Sampling.
//
void Pane::emitPhotons(thread_context *ctx, float numToFire,
		       photon_t phot_type){
  
  int i, j;
  int approxNum, targetNum, addNum;
  int numPhots, numCPhots, maxPhots;
  double cA, cB, cC;
  SbVec3f x;
  SbVec3f hemiDir, diffuseDir;
  SbVec3f bc;
  SbRotation sbr;
  double phi, theta, rho;
  //MultiHalton h;
  //sobinit(2);
  //double eps[4];


  //WWW'''''''''''''''''''''''''''''''''''
  // W                     W              
  // W  WW W  WW   W   WW WWW  W  WW   WWW
  // W  W W W W W W W W    W  W W W W  WW 
  //WWW W W W WW   W  W    W   W  W W WWW 
  //''''''''''W'''''''''''''''''''''''''''
  //          W                           

  if (phot_type == IMPORTANCE){

    Ray photon;
    photon.ctx = ctx;
    photon.pos = canvas.e;
    int sqrt_phots = (int)sqrt((float)numToFire);
    int total_phots = sqrt_phots*sqrt_phots;

    // one thread only calculations
    if (ctx->id == 0){

      firedPhotons += total_phots;
      
      // status report
      cerr << "\r" << total_phots
	   << " firing, " << importanceMap->get_stored_photons()
	   << " stored in importance map."
	   << flush;
    }

    for (j=0+ctx->id; j<sqrt_phots; j+=gv->numThreads){

      for (i=0; i<sqrt_phots; i++){

	double eps1, eps2;
	int n = j*sqrt_phots+i;
	eps1 = (double)n/total_phots;
	eps2 = RI_LP(n);
	
	// create Ray
	photon.dir = SbVec3f(

	  // sideways component
	  ( -canvas.width/2.0 + eps1 * canvas.width ) * canvas.u +

	  // vertical component
	  ( -canvas.height/2.0 + eps2 * canvas.height ) * canvas.v +

	  // distance ahead component
	  -canvas.s*canvas.w  
	  );

	// adjust for orthographic camera
	if (canvas.orthographic){
	  photon.pos = canvas.e + photon.dir + canvas.s*canvas.w;
	  photon.dir = -canvas.w;
	}
	photon.dir.normalize();

	// send photon
	bouncePhoton(ctx, photon, SbColor(1.0, 1.0, 1.0), phot_type, 1.0,
		     1, NULL);

      } // end for (i=...
    } // end for (j=...
    
    pthread_exit(0);      
  }
  
  float maxPhotonComponent = 0.0;
  int raysPerPoint = 4;

//  numToFire =1000;
  
    //'WW''''''''''''''''''WW''m'''''W'''''''''''
    //W  W                  W        W    W      
    //WWWW  WW  WW   WW     W  W  WW WW  WWW  WWW
    //W  W W   Wmmm W W     W  W W W W W  W   WW 
    //W  W W    WW   WW    WWW W  WW W W  W  WWW 
    //'''''''''''''''''''''''''''''W'''''''''''''
    //                           WW              
    vector<Triangle *>::iterator aI = litTris.begin();
    for (; aI != litTris.end(); aI++){

      // calculate average emission (power)
      double emisAvg = (*aI)->emisAvg;
      if (emisAvg == 0.0)
	continue;

      // photon power
      SbColor emission = (*aI)->emission;
      
      // emisAvg is in W
      // totalPower is in W
      float approxNum = emisAvg/totalEmisAvg *    // ~emisAvg but in [0,1]
	                numToFire;

      int points = (int)(approxNum/raysPerPoint+0.5);
      addNum = points*raysPerPoint;
      float scale = 1.0/(addNum*2.0);

      int index_p, index_r;

      // debugging
//       emission[R] = emission[R]+emission[G]+emission[B];
//       emission[G] = emission[B] = 0.0;
      
      // choose emission point
      emission *= scale; // scale photon power now
      
      // debugging
//       cerr << "Average area photon emission: " << emission << "\n";
//       cerr << "Average area light emission: " << emisAvg << "\n";

      // one thread only calculations
      if (ctx->id == 0){

	firedPhotons += addNum;
	totalPower += addNum*emisAvg*3.0*scale;

	// status report
	cerr << "\r" << addNum
	     << " firing, " << gPhotonMap->get_stored_photons()
	     << "g/" << cPhotonMap->get_stored_photons()
	     << "c stored          " << flush;

	// photon power
	for (int i=0; i<3; i++)
	  if (emission[i] > maxPhotonComponent)
	    maxPhotonComponent = emission[i];
      }

      //SbVec3f origEmission = emission;
      
      // targetNum 1..16
      // points (8+) 1..8
      // real 8*4..16*4
      for (int n=ctx->id; n<points; n+=gv->numThreads){

	// calculate index into quasi random sequence.
	// we have already fired addNum.
	// we are now shooting the second half.
	// therefore, the index starts at addNum.
	index_p = addNum+n;
    
	double eps1, eps2;
	//eps1 = h.next(0); eps2 = h.next(1);
	//eps1 = eps[0]; eps2 = eps[1];
	//eps1 = 0.5; eps2 = 0.5;
	if (gv->quasiRandom==0){
	  eps1 = ctx->mydrand48();
	  eps2 = ctx->mydrand48();
	}
	else {
	  eps1 = (double)n/points;//points==1?0.5:(double)n/(points-1);
	  eps2 = RI_LP(index_p);
	  //eps1 = RI_vdC(n, r1);
	  //eps2 = RI_S(n, r2);
	}
	
	// choose random position inside light
	cA = 1.0-sqrt(eps1);
	cB = (1.0-cA)*eps2;
	cC = 1.0-cA-cB;

	// debugging
	//print("cA, cB, cC=%.5e, %.5e, %.5e\n", cA, cB, cC);
	//print("eps1=%.5e, sqrt(eps1)=%.5e\n", eps1, sqrt(eps1));

// 	// do a texture lookup
// 	SbColor diffCol = (*aI)->getIntDiffColor(SbVec3f(cA, cB, cC));
//         // cerr << diffCol << "\n";
//         // cerr << emission[0] << " a:";
// 	emission[0] = origEmission[0] * diffCol[0];
// 	emission[1] = origEmission[1] * diffCol[1];
// 	emission[2] = origEmission[2] * diffCol[2];
// 	if (emission[0]<0.000001)
// 	  continue;

	
	// emit photon diffusely
	SbVec3f ng = (*aI)->getNormal();
	sbr.setValue(SbVec3f(0.0, 1.0, 0.0), ng);

	// 2d adjustment
	if (gv->useDimensions<3 && ng[1]<1.0){
	  ng[1] = 0.0;
	  ng.normalize();
	}

	// create randomized realization of (t,m,s)-net using
	// random digit scrambling
	// (Efficient Multidimensional Sampling - Kollig and Keller)
	int r3 = (int)(ctx->mydrand48()*INT_MAX);
	int r4 = (int)(ctx->mydrand48()*INT_MAX);

	for (int m=0; m<raysPerPoint; m++){
	  // choose random value
	  double eps1, eps2;
	  //eps1 = h.next(2); eps2 = h.next(3);
	  //eps1 = h.next(0); eps2 = h.next(1);
	  //sobvect(2, eps);
	  //eps1 = eps[0]; eps2 = eps[1];
	  //eps1 = (double)n/maxPhots;
	  //eps2 = RI_vdC(n);
	  //eps1 = actualPhotons==1?0.5:(double)n/(actualPhotons-1);
	  //eps2 = RI_LP(maxPhots/2+n);
	  //eps1 = RI_vdC(maxPhots/2+n, r2);
	  //eps2 = RI_S(maxPhots/2+n, r2);
	  index_r = raysPerPoint*index_p+m;
	
	  if (gv->quasiRandom==0){
	    eps1 = ctx->mydrand48();
	    eps2 = ctx->mydrand48();
	  }
	  else {
	    eps1 = RI_vdC(index_r, r3);
	    eps2 = RI_S(index_r, r4);
	  }
	  //eps1 = eps2 = 1.0;
	
	  phi = eps1 * 2.0 * M_PI;   // rotation about normal
	  theta = acos(sqrt(eps2));  // angle from normal
	  hemiDir.setValue(sin(theta)*cos(phi),
			   cos(theta),
			   sin(theta)*sin(phi));    
    
	  sbr.multVec(hemiDir, diffuseDir);
	  
	  // 2d adjustment
	  if (gv->useDimensions<3){
	    SbVec3f ng_perp = SbVec3f(ng[2], 0.0, -ng[0]);
	    theta = asin(2.0*eps1-1.0);
	    diffuseDir = ng*cos(theta) + ng_perp*sin(theta);
	  }
	  
	  diffuseDir.normalize();
	  x = cA*(*aI)->vert[0] + cB*(*aI)->vert[1] + cC*(*aI)->vert[2];

	  // emit photon
	  Ray photon(x, diffuseDir, ctx);
	  bouncePhoton(ctx, photon, emission, phot_type, maxPhotonComponent,
		       1, *aI);

	} // end for (int m=0; m<rays...
      } // end for (n=0; n<pnts...

    } // end for (; aI != litTris.end(); aI++){    light loop



    //WWW''''''m''''''''''''WW''m'''''W'''''''''''
    //W  W            W      W        W    W      
    //WWW   W  W WW  WWW     W  W  WW WW  WWW  WWW
    //W    W W W W W  W      W  W W W W W  W   WW 
    //W     W  W W W  W     WWW W  WW W W  W  WWW 
    //'''''''''''''''''''''''''''''W''''''''''''''
    //                            WW
    vector<SoPointLight *>::iterator pI = litPnts.begin();
    
    if (gv->headLightPos == HL_OFF)
      pI++;
      
    for (; pI != litPnts.end(); pI++){

      // get emission
      double pntEmisAvg = 0.0;
      SbColor emission = (*pI)->color.getValue()*(*pI)->intensity.getValue();
      for (i=0; i<3; i++)
	pntEmisAvg += emission[i];
      pntEmisAvg /= 3.0;

      float approxNum = pntEmisAvg/totalEmisAvg *    // ~emisAvg but in [0,1]
	                numToFire;

      addNum = (int)(approxNum+0.5);
      float scale = 1.0/(addNum*2.0);

      firedPhotons += addNum;
      totalPower += addNum*pntEmisAvg*3.0*scale;

      emission *= scale;
      
//       if (ctx->id == 0)
// 	cerr << "\r" << targetNum << flush;

      // one thread only calculations
      if (ctx->id == 0){
      
	firedPhotons += addNum;
	totalPower += addNum*pntEmisAvg*3.0*scale;
      
	// status report
	cerr << "\r" << addNum
	     << " firing, " << gPhotonMap->get_stored_photons()
	     << "g/" << cPhotonMap->get_stored_photons()
	     << "c stored          " << flush;
      
	// photon power
	for (int i=0; i<3; i++)
	  if (emission[i] > maxPhotonComponent)
	    maxPhotonComponent = emission[i];
      }
    
      // compute light location
      x = (*pI)->location.getValue();
      
      // emit numPhots photons
      for (int n=ctx->id; n<addNum; n+=gv->numThreads){

	// find emission point with rejection sampling
	do {
	  phi = ctx->mydrand48()*2.0-1.0;
	  theta = ctx->mydrand48()*2.0-1.0;
	  rho = ctx->mydrand48()*2.0-1.0;
	} while ((phi*phi+theta*theta+rho*rho)>1.0);
	diffuseDir.setValue(phi, theta, rho);

	// emit photon
	Ray photon(x, diffuseDir, ctx);
      
	// global photon
	bouncePhoton(ctx, photon, emission, phot_type, maxPhotonComponent);

      } // end for (n=0; n<numPhots; n++){   photon loop

    } // end for (; pI != litPnts.end(); aI++){    light loop




    //WWWW'''''''''W''''W''''''''''''WW''m'''''W'''''''''''
    //W            WW  WW             W        W    W      
    //WWW  WW  W W W WW W  WW WW      W  W  WW WW  WWW  WWW
    //W    W W W W W    W W W W W     W  W W W W W  W   WW 
    //WWWW W W  W  W    W  WW WW     WWW W  WW W W  W  WWW 
    //''''''''''''''''''''''''W''''''''''''''W'''''''''''''
    //                        W            WW              

    // calculate average emission (power)
    if (envMap){


      // statistics
      SbColor emissionAvg(0,0,0);
      int emissionCnt = 0;
      float pdfavg = 0.0;
      
      
      SbVec3f center;
      double radius, proj_area;

      if (phot_type == CAUSTIC){
	center = tritree.shiny_center;
	radius = tritree.shiny_radius;
      }
      else {
	center = tritree.center;
	radius = tritree.radius;
      }
      proj_area = M_PI * gv->meters2(radius*radius);
      float proj_area_fs = M_PI * gv->meters2(tritree.radius*tritree.radius);
      
      double emisAvg = (envMap->Lavg[0] + envMap->Lavg[1] + envMap->Lavg[2])/3.0
	* proj_area *4.0*M_PI;
    
      if (emisAvg > 0.0){

	// emisAvg is in W
	// totalPower is in W
	float approxNum = emisAvg/totalEmisAvg *    // ~emisAvg but in [0,1]
	                  numToFire;

	int points = (int)(approxNum/raysPerPoint+0.5);
	addNum = points*raysPerPoint;
	float scale = 1.0/(addNum*2.0);

	// debugging
// 	cerr << "\nEnvironment map: "
// 	     << " emisAvg = " << emisAvg
// 	     << " approxNum = " << approxNum
// 	     << "\n";

	int index_p, index_r;

	// one thread only calculations
	if (ctx->id == 0){

	  firedPhotons += addNum;
	  totalPower += addNum*emisAvg*3.0*scale;

	  // status report
	  cerr << "\r" << addNum
	       << " firing, " << gPhotonMap->get_stored_photons()
	       << "g/" << cPhotonMap->get_stored_photons()
	       << "c stored          " << flush;

	}


	// choose points spots
	for (int n=ctx->id; n<points; n+=gv->numThreads){

	  // calculate index into quasi random sequence.
	  // we have already fired half of addNum.
	  // we are now shooting the second half.
	  // therefore, the index starts at addNum.
	  index_p = points+n;
	  
	  double eps1, eps2;
	  if (gv->quasiRandom==0){
	    eps1 = ctx->mydrand48();
	    eps2 = ctx->mydrand48();
	  }
	  else {
	    eps1 = (double)n/points;//points==1?0.5:(double)n/(points-1);
	    eps2 = RI_LP(index_p);
	  }


	  // choose emission point
	  SbColor emission;
	  float pdf = envMap->getSample(eps1, eps2, hemiDir, emission);
	  emission *= proj_area/pdf;

	  // debugging
// 	  emission[B] = emission[R]+emission[G]+emission[B];
// 	  emission[R] = emission[G] = 0.0;
	  
	  emission *= scale; // scale photon power now

	  // statistics
	  emissionAvg += emission;
	  pdfavg += pdf;
	  emissionCnt++;

	  // debugging
// 	  if (ctx->mydrand48()<0.01)
// 	  cerr << "Env photon emission: " << emission
// 	       << " brightness=" << brightness(emission)
// 	       << " pdf=" << pdf
// 	       << "\n";
	  
	  // photon power
	  for (int i=0; i<3; i++)
	    if (emission[i] > maxPhotonComponent)
	      maxPhotonComponent = emission[i];
	
	  // store maximum photon emission component
	  
	  
	  hemiDir = -hemiDir; // flip direction around

	  // create randomized realization of (t,m,s)-net using
	  // random digit scrambling
	  // (Efficient Multidimensional Sampling - Kollig and Keller)
	  int r3 = (int)(ctx->mydrand48()*INT_MAX);
	  int r4 = (int)(ctx->mydrand48()*INT_MAX);

	  // chose points
	  for (int m=0; m<raysPerPoint; m++){
	    
	    // choose random value
	    double eps1, eps2;
	    index_r = raysPerPoint*index_p+m;
	
	    if (gv->quasiRandom==0){
	      eps1 = ctx->mydrand48();
	      eps2 = ctx->mydrand48();
	    }
	    else {
	      eps1 = RI_vdC(index_r, r3);
	      eps2 = RI_S(index_r, r4);
	    }


	    // choose point on unit disc
	    double r = sqrt(eps1);
	    theta = 2.0*M_PI*eps2;

	    // scale to radius of bounding sphere
	    r *= radius;

	    x[X] = r*cos(theta);
	    x[Y] = 0.0;
	    x[Z] = r*sin(theta);

	    // rotate disc normal to l
	    sbr.setValue(SbVec3f(0.0, 1.0, 0.0), hemiDir);
	    sbr.multVec(x, x);

	    // position outside bounding sphere
	    x += center;
	    x -= hemiDir*radius;
	  
	    // emit photon
	    Ray photon(x, hemiDir, ctx);
	    bouncePhoton(ctx, photon, emission, phot_type, maxPhotonComponent,
			 1, NULL);

	  } // end for (int m=0; m<rays...
	} // end for (n=0; n<pnts...

	// debugging
// 	cerr << "Average env light emission: " << emisAvg << "\n";
// 	cerr << "Average env photon emission: " << emissionAvg/emissionCnt << "\n";
// 	cerr << "Average pdf: " << pdfavg/emissionCnt << "\n";
	
      }
    }
    
    // time statistics
    struct tms renderTimes;
    times(&renderTimes);
    ctx->ticks += renderTimes.tms_utime + renderTimes.tms_cutime;
    
    pthread_exit(0);
}


//WW'''''''''''''''''''''''''''''WWWW''WW''''''''''WW''''''''''''''WW'WW'
//WW                             WW  W WW          WW             WW   WW
//WWW   WW   W WW WWW   WWW  WW  WW  W WWW   WW   WWWW  WW   WWW  WW   WW
//WW W WW W  W WW WW W WW W WWWW WWWW  WW W WW W   WW  WW W  WW W WW   WW
//WW W WW W  W WW WW W WW   WW   WW    WW W WW W   WW  WW W  WW W WW   WW
//WWW   WW    WWW WW W  WWW  WWW WW    WW W  WW    WW   WW   WW W  WW WW 
//                                                                       
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// bounce photon around until it absorbs, disappears, gets lost, etc.
//
// Light paths and the appropriate map:
// global map (D+S)*D
// caustic map: LS+D
//
void Pane::bouncePhoton(thread_context *ctx,
			Ray photon, SbColor emission, photon_t phot_type,
			float maxPhotonComponent,
			int light_id, Triangle *ignoreTri){

  // for tracking specular-ness of photons for caustic map gen.      
  int hit_specular = 0;
  int hit_diffuse = 0;
  int bounce = -1;
  SbRotation sbr;
  double phi, theta;
  SbVec3f hemiDir, diffuseDir;


//   Reasoning
//   ==========
//
//
//   OLD WAY - caustics contains L(S|D)*SD photons
//   ---------------------------------------------------------------------
//
//   Photons that hit a diffuse after a specular surface are stored
//   in the caustic map.
//
//   Because of this, when doing a final gather for indirect illumination,
//   gather rays should not follow specular bounces because this light
//   contribution will come from the caustic photon map.
//
//   Similarly, when importance sampling for indirect illumination,
//   we would like to find areas where indirect and non-specular lighting
//   is coming from, in other words, areas that contain diffuse surfaces.
//
//   Therefore, we store with a photon whether or not it came from a
//   diffuse surface, or in other words, from an area we might want
//   to sample from in the final gather.
//
//   We also store the light_id if it's directly from a light.
//   Currently this is unused, but it may be used in the future
//   for some type of direct lighting test.
//
//
//   NEW WAY (actually, original way, again) - caustic photons are LS+D
//   ---------------------------------------------------------------------
//
//   Photons that hit one or more specular surfaces and then a diffuse
//   surface are stored in the caustic map.
//
//   Final gather rays are then responsible for following specular paths
//   until they hit a diffuse surface, to account for L(S|D)*DS+DE paths.
//
//   Therefore, we tag flags with from_diffuse=1 after they hit at least
//   one diffuse surface.
//
//   Photons that come directly from a light are stored with a
//   non-zero light_id.


  // debugging
  int tir_bounces = 0;

  // Figure purposes - show emitted photons
  int showIt = 0;
//   const int numang = 40;
//   static int angles[numang] = {0};
//   const double targetHeight = 29.0;
  
  // if (ctx->mydrand48()<15.0/gv->numPhotons*8.0*1.0)
//   if (gv->figureMode && ctx->id==0 && ABSF(photon.pos[1]-targetHeight)<8.0
//       && gv->tempInt!=30){
    
//     float angle = atan2(-photon.dir[2], photon.dir[0]);
//     int iangle = (int)(angle*M_1_PI*numang);
//     if (angle>=0){
//       iangle = clamp(iangle, 0, numang);
//       if ((iangle+4)%7==0 && angles[iangle]==0 && ctx->mydrand48()<0.3){
// 	cerr << "SHOWING ANGLE: " << iangle << "(" << angle << ")" << endl;
// 	photon.pos[0] = 50.0+(angle*M_1_PI-0.5)*-20.0;
// 	photon.pos[2] -= 0.001;
// 	angles[iangle]=1;
// 	showIt = 1;
//       }
//     }
//     photon.pos[1] = targetHeight;
//   }
  
  while (++bounce<1000){

    // debugging
//     if (0&&bounce>=50){
//       if (bounce == 50 || bounce == 100 || bounce == 200 ||
// 	  bounce == 400 || bounce == 800){
// 	cerr << "Warning: " << bounce << " bounces.\n";
// 	printVec(photon.pos); cerr << "\n";
// 	showPoint(photon.pos, photon.t);
// 	photon.tri->show();
//       }
//     }

    // don't bother with dark photons (ones that bounce off red then green wall)
    if ((emission[0]+emission[1]+emission[2])<=0.0)
      break;

    // shoot photon
    if (! (tritree.intersect_ray(photon, ignoreTri)) ){
      
      // show miss
      if (showIt && emission[0]>0.0)
	photon.show(getMaxDist()/20.0, emission*gv->numPhotons/8.0);
      
      break; // does not hit anything
    }
    ignoreTri = photon.tri;

    // Super surfaces (double surfaces)
    //
    // Sometimes surfaces are on top of another.
    // this is a difficult situation, and should be avoided.
    // it not obvious what to do in this case, but if
    // we let photons bounce off of overlapping surfaces
    // as many times as they want, we will have a thousand
    // photons in one spot.
    //
    // To prevent this, we effectively police a mininum hit_t
    // by pushing the photon epsilon distance along the ray, and then let
    // the new minimum hit_t be zero.

    const float epsilon = 1.5e-4f;
    if (photon.t<epsilon){
      photon.pos += epsilon*photon.dir; // barely over overlapping edge tolerance

      //cerr << "close hit! t = " << photon.t << "\n\n";
      //showPoint(x, 15.0);
      
      // shoot photon
      if (! (tritree.intersect_ray(photon, ignoreTri)) )
	break; // does not hit anything

      //if (photon.t<epsilon)
      //cerr << "STILL close hit! t = " << photon.t << "\n\n";
      
      ignoreTri = photon.tri;
    }


    // Figure purposes - show photon paths
    if (showIt){
      // if (emission[0]==0.0)
      // showIt = 0;
      // else
      photon.show(photon.t, emission*gv->numPhotons/8.0);
      // if (ctx->mydrand48()<0.4)
      // showIt = 0;
    }

    // compute point of intersection
    SbVec3f x = photon.pos + photon.t*photon.dir;

    // do Beer's law attenuation
    if (photon.atten){

      SbColor att(exp(gv->attenuation[R]*photon.t),
		  exp(gv->attenuation[G]*photon.t),
		  exp(gv->attenuation[B]*photon.t));
      
      emission[R] *= att[R];
      emission[G] *= att[G];
      emission[B] *= att[B];

      // russian roulette
      if (emission[0]<maxPhotonComponent &&
	  emission[1]<maxPhotonComponent &&
	  emission[2]<maxPhotonComponent){

	float attAvg = (att[0]+att[1]+att[2])*1.0/3.0;
	if (ctx->mydrand48() < attAvg){
	  float inv_attAvg = 1.0/attAvg;
	  emission[R] *= inv_attAvg;
	  emission[G] *= inv_attAvg;
	  emission[B] *= inv_attAvg;
	}
	else
	  return;
      }
    }
    
    // calculate surface normal
    SbVec3f ngl = photon.tri->getNormal();
    int flipNormal = 0;
    if (photon.dir.dot(ngl)>0.0){      // outward ng
      ngl.negate();
      flipNormal = 1;
    }
      
    float shiny_comp = photon.tri->shiny?photon.tri->shininess:0.0;
    float diff_comp = 1.0 - shiny_comp;

    // temporary test
    SbColor tempCol = photon.tri->getIntDiffColor(photon.bc);
    float totalRefl = tempCol[0]+tempCol[1]+tempCol[2];
    
    // store photon
    if (diff_comp > 0.01 && totalRefl > 0.0001){

      if (phot_type == IMPORTANCE){
	
	// store in importance map
	if (gv->showPhotons)
	  showPhoton(x, SbColor(1.0, 1.0, 0.0));
	
	importanceMap->buffered_store(emission.getValue(),
				      x.getValue(),
				      photon.dir.getValue(),
				      ngl.getValue(), light_id, hit_diffuse,
				      ctx->id);

	if (bounce >= 1)
	  return;
      }
      // before doing importance check, verify photon may possibly be stored
      else if (phot_type == GLOBAL || hit_specular) {

	// do importance sampling
	// (throw away unimportant photons)
	int store_okay = 1;
	if (gv->numIPhotons){
	  
	  store_okay = 0;
	  
	  // is it even close ?
	  SbVec3f diff = x-importanceMap->bbox_cen;
	  if (diff.dot(diff) < importanceMap->bound_rad2){
	    
	    float irrad[3];
 	    importanceMap->irradiance_estimate(
 	      irrad, x.getValue(), ngl.getValue(),
 	      gv->iPhotonEstimate);
// 	    importanceMap->irradiance_estimate_pre(
// 	      irrad, x.getValue(), ngl.getValue());

	    if (irrad[0]>0.0)
	      store_okay = 1;
	  }
	}

	if (store_okay) {
      
	  // store in caustic map if hit specular surface
	  int shown = 0;
	  if (hit_specular && !hit_diffuse){  
	    if (gv->showPhotons){
	      showPhoton(x, SbColor(1.0, 0.0, 0.0));
	      shown = 1;
	    }
	    cPhotonMap->buffered_store(emission.getValue(),
				       x.getValue(),
				       photon.dir.getValue(),
				       ngl.getValue(), light_id, hit_diffuse,
				       ctx->id);
	    // hit_specular = 0;
	  }

	  // stop here if we're only building caustics map
	  // if (caustic_only)
	  // return;

	  if (phot_type == GLOBAL){
	    // store in global map
	    if (gv->showPhotons && !shown)
	      showPhoton(x, SbColor(0.0, 0.0, 1.0));
	
	    gPhotonMap->buffered_store(emission.getValue(),
				       x.getValue(),
				       photon.dir.getValue(),
				       ngl.getValue(), light_id, hit_diffuse,
				       ctx->id);
	  }
	}
      }
    }

    if (gv->useGlassModel == 0){

      // transmit in straight line
      if (ctx->mydrand48() < photon.tri->transparency){
	photon.pos = x;

	hit_specular = 1;
	continue;
      }
    }
    
    float eps = ctx->mydrand48();

    // calculate interpolated normal
    SbVec3f n = photon.tri->getIntNormal(photon.bc);
    SbVec3f nl = n;
    if (flipNormal)
      nl.negate();
    
    // calculate specular reflectance
    SbColor pS = photon.tri->specular;

    // fresnel effect (schlick approx & H=n for specular directions)
    float fresnel = 1.0 + (photon.dir.dot(nl)); 
    fresnel = fresnel*fresnel*fresnel*fresnel*fresnel;
    pS += (SbColor(1.0,1.0,1.0)-pS)*fresnel;
    
    double pSavg = (pS[0]+pS[1]+pS[2])/3.0;

    // don't use russian roulette on high power photons
    if (emission[R]>maxPhotonComponent ||
	emission[G]>maxPhotonComponent ||
	emission[B]>maxPhotonComponent)
      pSavg = 1.0;
    
    ////////////////////////////
    // specular reflection
    ////////////////////////////

    if (eps < pSavg*shiny_comp){
      
      // track caustic photons
      hit_specular = 1;
      light_id = 0;      

      ////////////////////////////
      // transmit
      ////////////////////////////
      if ((photon.tri->transparency>0.0) && gv->useGlassModel &&
	  ctx->mydrand48() < photon.tri->transparency){

	SbVec3f tdir;
	double c;
	SbVec3f k;
	double R0;
	double Re;

	// From RRT by PS
	double nc, nt;
	//double ar, ag, ab;
	// convert a/m ->a/u
	//ar = ag = ab = log(photon.tri->transparency)*gv->meters(1.0); 
	int tir;

	if (n.dot(ngl)>0.0){  // into object
	  nc = N_AIR;         // index of refraction for vacuum/air
	  nt = gv->refractionIndex;       // index of refraction for glass
	  refract(tdir, tir, photon.dir, n, nc/nt);
	  k = SbVec3f(1.0, 1.0, 1.0);
	  c = -photon.dir.dot(n);
	}
	else {                // out of object
	  nc = gv->refractionIndex;
	  nt = N_AIR;
	  refract(tdir, tir, photon.dir, -n, nc/nt);
	  float t = gv->meters(photon.t);
	  k = SbVec3f(1.0, 1.0, 1.0);
// 	  k[R] = exp(gv->attenuation[R]*t);
// 	  k[G] = exp(gv->attenuation[G]*t);
// 	  k[B] = exp(gv->attenuation[B]*t);
	  //k = SbVec3f(exp(ar*photon.t), exp(ag*photon.t), exp(ab*photon.t));
	  c = tdir.dot(n);
	}
 
	//double attenuation = (k[R]+k[G]+k[B])/3.0;
	
	// total internal reflection
	if (tir){
	  tir_bounces++;
	  Re = 1.0f;

	  // use russian roulette to prevent infinite loops
	  if (tir_bounces>20){

	    // double power
	    const float extinction = 0.9;
	    if (ctx->mydrand48()<extinction)
	      emission /= extinction;

	    // or absorb
	    else
	      break;
	  }
	  
	}
	else {

	  // grazing angle problem (normal interpolation error)
	  // don't let transmitted ray be on same side
	  if (tdir.dot(ngl)>0.0){
	    tdir -= (tdir.dot(ngl)+0.0015) * ngl;
	    tdir.normalize();
	  }
    
	  // due to errors for grazing angles
	  // with normal interpolation, c may
	  // incorrectly be < 0.0. This clamps it.
	  if (c<0.0)
	    c = 0.0;
	  
	  R0 = (nt-nc)/(nt+nc);
	  R0 *= R0;
	  float temp = 1.0-c;
	  Re = R0+(1.0-R0)*temp*temp*temp*temp*temp;	  
	}

	// Re = 0.0; // UNDO ME
      
	// transmit
	if (ctx->mydrand48() > Re){
	      
	  // scale power
	  for (int i=0; i<3; i++)
	    emission[i] = emission[i]*pS[i]*k[i]/pSavg;

	  // bounce photon
	  photon.atten = 1-photon.atten;    // attenuate photon power (toggle bit)
	  photon.dir = tdir;
	  photon.pos = x;

	  continue;
	}

	// reflect
	else {
	      
	  // scale power
	  for (int i=0; i<3; i++)
	    emission[i] = emission[i]*pS[i]*k[i]/pSavg;

	  // normal correction (two sided lighting)
	  SbVec3f rdir = photon.dir-2.0*n.dot(photon.dir)*n;

	  // grazing angle problem
	  // don't let reflected ray go beneath surface
	  if (rdir.dot(ngl)<0.0){
	    rdir += (-rdir.dot(ngl)+0.0015) * ngl;
	    rdir.normalize();
	  }

	  // bounce photon
	  photon.dir = rdir;
	  photon.pos = x;

	  continue;
	}

      }

      ////////////////////////////
      // reflect
      ////////////////////////////
      else {

	// reflect specularly
	SbVec3f rdir = photon.dir-2.0*n.dot(photon.dir)*n;

	// grazing angle problem
	// don't let reflected photon go beneath surface
	if (rdir.dot(ngl)<0.0){
	  rdir += (-rdir.dot(ngl)+0.0015) * ngl;
	  rdir.normalize();
	}
	  
	// scale power
	for (int i=0; i<3; i++)
	  emission[i] = emission[i]*pS[i]/pSavg;
	    
	// bounce photon
	photon.dir = rdir;
	photon.pos = x;

	continue;
      }

    }

    // specular photons do not bounce diffusely
    if (phot_type == CAUSTIC)
      return;
    
    // calculate diffuse reflectance
    SbColor pD = SbColor(photon.tri->getIntDiffColor(photon.bc));
    double pDavg = (pD[0]+pD[1]+pD[2])/3.0;

    // don't use russian roulette on high power photons
    if (emission[R]>maxPhotonComponent ||
	emission[G]>maxPhotonComponent ||
	emission[B]>maxPhotonComponent)
      pDavg = 1.0;

    ////////////////////////////
    // diffuse reflection
    ////////////////////////////
    
    if (eps < ((pDavg*diff_comp)+(pSavg*shiny_comp))){
    
      // flip normal if transmitting
      int transp = (photon.tri->transparency>0.0) && gv->useGlassModel &&
	ctx->mydrand48() < photon.tri->transparency;
      if (transp)
	ngl = -ngl;

      // calculate new direction
      sbr.setValue(SbVec3f(0.0, 1.0, 0.0), ngl);

      double eps1 = ctx->mydrand48();
      double eps2 = ctx->mydrand48();
      phi = eps1 * 2.0 * M_PI;   // rotation about normal
      theta = acos(sqrt(eps2));  // angle from normal
      hemiDir.setValue(sin(theta)*cos(phi),
		       cos(theta),
		       sin(theta)*sin(phi));    
    
      sbr.multVec(hemiDir, diffuseDir);

      // 2d adjustment
      if (gv->useDimensions<3){
	ngl[1] = 0.0;
	ngl.normalize();
	SbVec3f ngl_perp = SbVec3f(ngl[2], 0.0, -ngl[0]);
	theta = asin(2.0*eps1-1.0);
	diffuseDir = ngl*cos(theta) + ngl_perp*sin(theta);
      }
      
      diffuseDir.normalize();

      // scale power
      for (int i=0; i<3; i++)
	emission[i] = emission[i]*pD[i]/pDavg;
	    
      // bounce photon
      photon.dir = diffuseDir;
      photon.pos = x;

      hit_diffuse = 1;
      light_id = 0;
      continue;
	  
    } // end trasmit/spec/diff/absorb decision

    ////////////////////////////
    // absorb
    ////////////////////////////

    return;
    
  } // end while() loop for bouncing   bounce loop
}



//'''''''''''''''''WW'''''''''''WWWW'''''''WW'WW'
//                 WW            WW       WW   WW
//W WW  WW  WWW   WWW  WW  W WW  WW  WW W WW   WW
//WW   WWWW WW W W WW WWWW WW    WW  WW W WW   WW
//WW   WW   WW W W WW WW   WW    WW  WW W WW   WW
//WW    WWW WW W  WWW  WWW WW   WWWW  WW   WW WW 
//                                               
//'''''''''''''''''''''''''''''''''''''''''''''''
// (renderIv)
//
// Like render(), but instead of rendering a 2D image
// renders a 3D Inventor scene,
// computing radiance values for each vertex
// in the scene.
// (It actally computes a radiance value for each face
// and distributes the color to the vertices.)
// The scene's color values is edited in place,
// inserting SoMaterial nodes when appropriate.
// 
// Unlike render(), renderIv() performs tone mapping
// before returning.
//
// This procedure is NOT multi-threaded.
//

void Pane::renderIv(thread_context *ctx){

  // tone mapping variables
  const double k = 683.0;  // k constant (cd/W) (Peter Shirley)
  double V;
  const double Yw = 110.5; // maximum display luminance (cd/m^2) (PS)
  double Yi = 0.0;         // world adaptation luminance
  double cX, cY, cZ;       // tristimulous values
  double x, y, W, s;       // misc
  const double l[3] = { 620.0, 545.0, 465.0 }; // lambda (nm)
  double r, g, b;
  int numpix = 0;

  const double xw=0.3333, yw=0.3333;
  const double d = (0.4905*0.65 + 0.9928*0.94 + 0.0770*1.0) /
                   (0.2126+0.7152 +0.0722);  // 1.3291
  const double e = d;
  const double f = d;
  const double xb[3] = { 0.4124*d, 0.3576*e, 0.1805*f }; // xbar
  const double yb[3] = { 0.2126*d, 0.7152*e, 0.0722*f }; // ybar
  const double zb[3] = { 0.0193*d, 0.1192*e, 0.9505*f }; // zbar
  const double vb[3] = { 0.01477, 0.497, 0.631 }; // vbar
  const double dlambda[3] = { 1.0, 1.0, 1.0 };

  // feedback
  struct timeval start, end;
  cerr << "Rendering...\n";
  gettimeofday(&start, NULL); // start delay measure
  
  // loop over surfaces
  int totalTris = 0;
  int totalVerts = 0;
  int j;

  // Search for indexed face sets
  SoCallbackAction callback;
  currMaterial = NULL;
  callback.addPreCallback(SoMaterial::getClassTypeId(), materialCBHandler, this);
  callback.addPreCallback(SoShapeHints::getClassTypeId(), shapeHintsCBHandler,
			  this);

  SoSearchAction sa;
  sa.setType(SoIndexedShape::getClassTypeId(), TRUE);
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.setFind(SoSearchAction::TYPE);
  sa.setInterest(SoSearchAction::ALL);
  sa.apply(gv->scene);
  SoPathList plist = sa.getPaths();
  int len = plist.getLength();
  cerr << len << " paths found\n";

  //W'''''''''''''''''''''''''''''''''''''''''''''''''''W'''''''
  //W                                                W  W       
  //W     W   W  WW      W  W W  WW   WW    WW   WW WWW WW   WWW
  //W    W W W W W W    W W W W Wmmm W      W W W W  W  W W  WW 
  //WWWW  W   W  WW      W   W   WW  W      WW   WW  W  W W WWW 
  //'''''''''''''W''''''''''''''''''''''''''W'''''''''''''''''''
  for (int i=0; i<len; i++){

    // insert lightmodel & material node
    // render one pixel for triangle
    // add color to vertex colors (3),
    //   add 1 to each vertex surface counter
    SoPath *path = (SoPath*) plist[i];
    SoIndexedShape *shape = (SoIndexedShape *)path->getTail();

    cerr << "indexed shape ";
    
    SoVertexProperty *vprop;
    vprop = (SoVertexProperty *)shape->vertexProperty.getValue();
    if (!vprop){
      cerr << "Vprop not found\n";
      continue;
    }

    // get group
    currGroup = (SoGroup *)path->getNodeFromTail(1);
    currMaterial = NULL;
    currShapeHints = NULL;
    callback.apply(path);

    // get shape hints
    int reverseOrdering = 0;
    if (currShapeHints){
      if (currShapeHints->vertexOrdering.getValue() == SoShapeHints::CLOCKWISE)
	reverseOrdering = 1;
    }
    
    // get transformation
    SoGetMatrixAction ma = SoGetMatrixAction(SbViewportRegion());
    ma.apply(path);
    SbMatrix matrix = ma.getMatrix();
    SbMatrix matInvTran = matrix.inverse().transpose();

    // fix orderedRGBA
    vprop->orderedRGBA.setNum(0);
    int numVerts = vprop->vertex.getNum();
    vprop->materialBinding.setValue(SoVertexProperty::PER_VERTEX_INDEXED);
    vprop->normalBinding.setValue(SoVertexProperty::PER_VERTEX_INDEXED);

    // fix material binding
    // shape->materialIndex.set1Value(j, shape->coordIndex[j]);
    int numCoordIdxs = shape->coordIndex.getNum();
    cerr << numCoordIdxs << " coord indexes\n";
    for (j=0; j<numCoordIdxs; j++)
      shape->materialIndex.set1Value(j, shape->coordIndex[j]);

    // currGroup->findChild(shape) is flawed,
    // as it assumes the group contains the shape.
    // the group might not contain the shape,
    // merely come before the shape (such as a switch)
    // and be empty or otherwise not contain the shape.
    
    // insert lightmodel
    SoLightModel *lightModel = new SoLightModel();
    lightModel->model = SoLightModel::BASE_COLOR;
    currGroup->insertChild(lightModel, currGroup->findChild(shape));

    // insert material
    SoMaterial *mat = new SoMaterial();
    if (currMaterial)
      mat->transparency = currMaterial->transparency;
    mat->diffuseColor.setNum(numVerts);
    for (j=0; j<numVerts; j++)
      mat->diffuseColor.set1Value(j, SbColor(0.0, 0.0, 0.0));
    currGroup->insertChild(mat, currGroup->findChild(shape));

    // insert material binding
    SoMaterialBinding *matBind = new SoMaterialBinding();
    matBind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
    currGroup->insertChild(matBind, currGroup->findChild(shape));

    // insert shininess stuff after last indexed shape
    if (gv->renderFlags==60 && i==(len-1)){


      // is this working right? - kevin 11/9/04

      // append light model
      SoLightModel *phongLightModel = new SoLightModel();
      phongLightModel->model = SoLightModel::PHONG;
      currGroup->insertChild(phongLightModel, currGroup->findChild(shape)+1);

      // insert material
      SoMaterial *shinyMat = new SoMaterial();

      // if .8 shiny, we make (1-.8) transparent and 100% shiny
      if (currMaterial)
	shinyMat->transparency.set1Value(0,
			MAX2(1.0 - currMaterial->shininess[0], 0.01));
      else
	shinyMat->transparency.setValue(0.5);
      for (j=0; j<numVerts; j++)
	shinyMat->diffuseColor.set1Value(j, SbColor(0.0, 0.0, 0.0));
      // shinyMat->diffuseColor.setValue(0.0, 0.0, 0.0);
      shinyMat->specularColor.setValue(1.0, 1.0, 1.0);
      shinyMat->shininess.setValue(1.0);
      currGroup->insertChild(shinyMat, currGroup->findChild(shape)+2);

      // re-add shape
      currGroup->insertChild(shape, currGroup->findChild(shape)+3);

      // create off switch
      SoSwitch *shinySwitch = new SoSwitch();
      shinySwitch->addChild(mat);
      shinySwitch->addChild(shape);
      
      // add switch (so blurring tool can find it, i.e. it is last)
      currGroup->insertChild(shinySwitch, currGroup->findChild(shape)+4);
    }

    // initialize flags
    int flags;
    if (gv->renderFlags){
      flags = gv->renderFlags |
	gv->approximate * Pane::C_APPROX;
    }
    else {
      flags =
	Pane::C_NORMAL |
	(gv->algorithm==1?1:0) * Pane::C_MDIFFUSE |
	gv->approximate * Pane::C_APPROX;
    }

    
    // keep record of colors stored per vertex
    float *okayNormals = new float[numVerts];
    for (j=0; j<numVerts; j++)
      okayNormals[j] = -1.0;


    //WW''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''WW''WW'
    //WW                                                                WW     
    //WW     WW    WW   WWW      WW   WW W  WW  W WW    WW W  WW  W WW WWWW WW 
    //WW    WW W  WW W  WW W    WW W  WW W WWWW WW      WW W WWWW WW    WW  WW 
    //WW    WW W  WW W  WW W    WW W  WW W WW   WW      WW W WW   WW    WW  WW 
    //WWWWW  WW    WW   WWW      WW    WW   WWW WW       WW   WWW WW    WW  WW 
    //                  WW                                                     
    //''''''''''''''''''WW'''''''''''''''''''''''''''''''''''''''''''''''''''''

    typedef struct renderVert_st {
      int numVerts;
      SoMaterial *mat;
      SoVertexProperty *vprop;
      SbMatrix *matrix;
      SbMatrix *matInvTran;
      float *okayNormals;
      int flags;
      SoIndexedShape *shape;
      int numCoordIdxs;
      int reverseOrdering;
    } renderVert_st_t;

    renderVert_st_t passingVals;
    passingVals.numVerts = numVerts;
    passingVals.mat = mat;
    passingVals.vprop = vprop;
    passingVals.matrix = &matrix;
    passingVals.matInvTran = &matInvTran;
    passingVals.okayNormals = okayNormals;
    passingVals.flags = flags;
    passingVals.shape = shape;
    passingVals.numCoordIdxs = numCoordIdxs;
    passingVals.reverseOrdering = reverseOrdering;

    // launch threads
    MemberFunction *memberFunc = (MemberFunction *)alloca(
      sizeof(MemberFunction)*gv->maxThreads); // used to pass object pointer
    thread_context *ctx = (thread_context *)alloca(
      sizeof(thread_context)*gv->maxThreads);   // random number generators

    // debugging
    int cp = irradCache.numPoints();
    
    // subscoping to eliminate i loop variable namespace conflict
    {
      // scramble random number generators
      for (int i=0; i<gv->numThreads; i++){
	ctx[i].mydrand48 = RNG(lrand48());
	ctx[i].id = i;
	ctx[i].rays = 0;
	ctx[i].ticks = 0;
	ctx[i].other = (void *) &passingVals;
      }

      if (gv->numThreads == 1){
	renderVerts(&ctx[0]);
      }
      else {

	// Launch Threads
	for (int i=0; i<gv->numThreads; i++){

	  // pack object pointer, function pointer, and argument into struct
	  memberFunc[i].object = this;
	  memberFunc[i].pmf = &Pane::renderVerts;
	  memberFunc[i].ctx = &ctx[i];

	  // call wrapper function with struct
	  pthread_create(&gv->threads[i], NULL, Pane::renderTileWrapper,
			 (void*)&memberFunc[i]);
	}
  
	// wait for threads to finish
	for (int i=0; i<gv->numThreads; i++){
	  pthread_join(gv->threads[i], NULL);
	  renderRaysFired += ctx[i].rays;
	  gv->times.renderTicks += ctx[i].ticks;
	  gv->times.totalTicks += ctx[i].ticks;
	}
      }

      cerr << "\n";

      // debugging
      if (STOP_RENDERING)
	return;
    }

    
    
    // debugging
    int cp2 = irradCache.numPoints();
    cerr << cp2 << " irrcache points (+"
	 << cp2-cp << ")\n";
    
    totalVerts += numVerts;

    // normalize colors and find Yi
    for (j=0; j<numVerts; j++){

      // skip bad vertices
      if (okayNormals[j]<0.0)
	continue;

      r = mat->diffuseColor[j][R];
      g = mat->diffuseColor[j][G];
      b = mat->diffuseColor[j][B];

      // make sure pixel colors are not inf or nan
      if ( !(finite(r) && finite(g) && finite(b)) ){
	cerr << "Vertex color at (" << i << ", " << j << ") is invalid: ";
	printCol(SbColor(r,g,b));
	cerr << endl;

	if (!finite(r))
	  r = 0.0;
	if (!finite(g))
	  g = 0.0;
	if (!finite(b))
	  b = 0.0;

	mat->diffuseColor.set1Value(j, SbColor(r,g,b));
      }

      cY = k * ( yb[R]*r*dlambda[R] + yb[G]*g*dlambda[G] +
		 yb[B]*b*dlambda[B] );

      // minimum value of cY taken from Tumblin Rushmeier '93
      if (cY>3.18e-7){
        //cY = 3.18e-7;
	Yi += log(cY);
	numpix++;
      }
      
    }

    delete [] okayNormals;

  } // end IndexedShape loop
  cerr << totalTris << " triangles, "
       << totalVerts << " verts found\n";



  //WWWWW'''''''''''''''''W''''W''''''''
  //  W                   WW  WW        
  //  W    W  WW   WW     W WW W  WW WW 
  //  W   W W W W Wmmm    W    W W W W W
  //  W    W  W W  WW     W    W  WW WW 
  //'''''''''''''''''''''''''''''''''W''
  
  // Log average of Y idea of Jack Tumblin and Holly Rushmeier
  // finish calculating Yi
  if (numpix)
    Yi /= numpix;
  Yi += 0.84;           // from Tumblin Rushmeier '93
  Yi = exp(Yi);


  // debugging
  if (gv->luminanceMean>0.0)
    Yi = gv->luminanceMean;
  
  lastImageYi = Yi;
  
  cerr << "Using Yi = " << Yi << endl;
  
  double sf = 1.0/Yw * pow( ( 1.219 + pow(Yw/2.0, 0.4) ) /
			    ( 1.219 + pow(Yi, 0.4) ), 2.5);
//  sf *= gv->exposure;

  
  // loop over paths again
  for (int i=0; i<len; i++){

    // insert lightmodel & material node
    // render one pixel for triangle
    // add color to vertex colors (3),
    //   add 1 to each vertex surface counter
    SoPath *path = (SoPath*) plist[i];
    SoIndexedShape *shape = (SoIndexedShape *)path->getTail();

    cerr << "indexed shape\n";
    
    SoVertexProperty *vprop;
    vprop = (SoVertexProperty *)shape->vertexProperty.getValue();
    if (!vprop){
      cerr << "Vprop not found\n";
      continue;
    }

    // get material
    callback.apply(path);
    if (!currMaterial){
      cerr << "material not found\n";
      continue;
    }
 
    // tone map vertex colors
    int numVerts = vprop->vertex.getNum();
    for (j=0; j<numVerts; j++){

      r = currMaterial->diffuseColor[j][R]*gv->exposure;
      g = currMaterial->diffuseColor[j][G]*gv->exposure;
      b = currMaterial->diffuseColor[j][B]*gv->exposure;

      // do tone mapping
      if (gv->toneMapping){

	cX = k * ( xb[R]*r*dlambda[R] + xb[G]*g*dlambda[G] +
		   xb[B]*b*dlambda[B] );
	cY = k * ( yb[R]*r*dlambda[R] + yb[G]*g*dlambda[G] +
		   yb[B]*b*dlambda[B] );
	cZ = k * ( zb[R]*r*dlambda[R] + zb[G]*g*dlambda[G] +
		   zb[B]*b*dlambda[B] );
      
	V = 1700.0 * ( vb[R]*r*dlambda[R] + vb[G]*g*dlambda[G] +
		       vb[B]*b*dlambda[B] );
      
	if (cY<3.18e-7)
	  cY = 3.18e-7;

	if (log10(cY) < -2.0)
	  s = 0.0;
	else if (log10(cY) < 0.6){
	  double f = (log10(cY)+2.0)/2.6;
	  s = 3.0*f*f -  2.0*f*f*f;
	}
	else
	  s = 1.0;

	W = cY + cX + cZ;
	x = cX/W;
	y = cY/W;

	x = (1.0-s)*xw + s*(x+xw-0.3333);
	y = (1.0-s)*yw + s*(y+yw-0.3333);
	cY = 0.4468*(1.0-s)*V + s*cY;
	
	// Scale luminance
	cY *= sf;
	cX = x*cY/y;
	cZ = (1.0-x-y)*cY/y;

	r =  3.2406*cX +  -1.5372*cY +  -0.4986*cZ;
	g = -0.9689*cX +   1.8758*cY +   0.0415*cZ;
	b =  0.0557*cX +  -0.2040*cY +   1.0570*cZ;
	
      }
      
      // gamma correct
      r = clamp(r, 0.0, 1.0);
      g = clamp(g, 0.0, 1.0);
      b = clamp(b, 0.0, 1.0);

      if (fabs(gv->gamma-2.2)>0.001){

	// regular gamma correction
	r = pow(r, 1.0/gv->gamma);
	g = pow(g, 1.0/gv->gamma);
	b = pow(b, 1.0/gv->gamma);
      }
      else if (fabs(gv->gamma-1.0)>0.001){
	// sRGB standard gamma correction (similar to regular w/2.2)
	r = r<=0.0031308 ? 12.92*r : 1.055*pow(r, 1.0/2.4) - 0.055;
	g = g<=0.0031308 ? 12.92*g : 1.055*pow(g, 1.0/2.4) - 0.055;
	b = b<=0.0031308 ? 12.92*b : 1.055*pow(b, 1.0/2.4) - 0.055;
      }
	
      currMaterial->diffuseColor.set1Value(j, SbColor(r,g,b));
      
    } // end verts loop
    // end tone mapping
  } // end paths loop

  // end delay measure
  gettimeofday(&end, NULL);
  int ms = delay(start, end);
  if (!gv->silent)
    print("\a");
  print("Done. (%d ms, %f s, %f m, %f h)\n", ms, ms/1000.0, ms/1000.0/60.0,
	ms/1000.0/60.0/60.0);
  gv->times.render = ms;
}



//'''''''''''''''''WW'''''''''''WW''W''''''''''''WW'''''''
//                 WW           WW  W            WW       
//W WW  WW  WWW   WWW  WW  W WW WW  W  WW  W WW WWWW  WWWW
//WW   WWWW WW W W WW WWWW WW   WW  W WWWW WW    WW  WWW  
//WW   WW   WW W W WW WW   WW   WW W  WW   WW    WW    WWW
//WW    WWW WW W  WWW  WWW WW    WW    WWW WW    WW  WWWW 
//                                                        
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''

void Pane::renderVerts(thread_context *ctx){

  typedef struct renderVert_st {
    int numVerts;
    SoMaterial *mat;
    SoVertexProperty *vprop;
    SbMatrix *matrix;
    SbMatrix *matInvTran;
    float *okayNormals;
    int flags;
    SoIndexedShape *shape;
    int numCoordIdxs;
    int reverseOrdering;
  } renderVert_st_t;

  renderVert_st_t *passingVals = (renderVert_st_t *) ctx->other;
  int numVerts = passingVals->numVerts;
  SoMaterial *mat = passingVals->mat;
  SoVertexProperty *vprop = passingVals->vprop;
  SbMatrix *matrix = passingVals->matrix;
  SbMatrix *matInvTran = passingVals->matInvTran;
  float *okayNormals = passingVals->okayNormals;
  int flags = passingVals->flags;
  SoIndexedShape *shape = passingVals->shape;
  int numCoordIdxs = passingVals->numCoordIdxs;
  int reverseOrdering = passingVals->reverseOrdering;

  // loop over triangles
  int v[3];
  v[0] = v[1] = v[2] = -1;
  int triCount = 0;
  int numTris = 0;
  int colinearVerts = 0;
  for (int k=0; k<numCoordIdxs; k++){
    v[0] = v[1];
    v[1] = v[2];
    v[2] = shape->coordIndex[k];
    if (v[0] >= 0 && v[1] >= 0 && v[2] >= 0){
      	
      // triangle found
      numTris++;
      triCount++;
      if (ctx->id==0 && numTris%1000==0)
	print("\r%6.2f%%    ", k/(float)(numCoordIdxs)*100.0);

      // detect same vertices
      if (v[0] == v[1] || v[0] == v[2] || v[1] == v[2]){
	if (colinearVerts++ == 0)
	  cerr << "Warning! Colinear vertices in triangle!\n";
	continue;
      }

      // get points
      SbVec3f vert[3];
      matrix->multVecMatrix(vprop->vertex[v[0]], vert[0]);
      matrix->multVecMatrix(vprop->vertex[v[1]], vert[1]);
      matrix->multVecMatrix(vprop->vertex[v[2]], vert[2]);


      // get normal and area
      SbVec3f normal;
      SbVec3f edge1 = vert[1] - vert[0];
      SbVec3f edge2 = vert[2] - vert[0];
      
      if (reverseOrdering == 0)
	normal = edge1.cross(edge2);
      else
	normal = edge2.cross(edge1);
      
      // flip odd triangles if indexedtrianglestripset
      if (shape->isOfType(SoIndexedTriangleStripSet::getClassTypeId()) &&
	  triCount%2==0)
	normal = -normal;
      
      float area2 = normal.normalize();  
      
      if (area2<0.00001){
//	cerr << "area too small, skipping.\n";
	continue;
      }

      // sample triangles vertices
      for (int i=0; i<3; i++){

	// skip points with no valid normal or already computed normal
	int j = v[i];
 	if (j%gv->numThreads != ctx->id)
 	  continue;
	
// 	if (okayNormals[j] != 1)
// 	  continue;
// 	okayNormals[j]++;
	if (okayNormals[j] >= 0.0  &&          // normal is defined
	    ! (area2 > 5.0*okayNormals[j] ))   // tri is not 5x bigger than last tri
	  continue;
	okayNormals[j] = area2;

	// set vert to 99% of one corner of triangle
	int i2 = (i+1)%3;
	int i3 = (i+2)%3;
	float w[3] = // { 1.0/3.0, 1.0/3.0, 1.0/3.0};
	  { 0.96, 0.02, 0.02 };
	SbVec3f cvert = vert[i]*w[0] + vert[i2]*w[1] + vert[i3]*w[2];

// 	SbVec3f normal2;
// 	matInvTran->multDirMatrix(vprop->normal[j], normal2);
// 	normal2.normalize();
	SbVec3f normal2 = normal;

	SbVec3f antinormal = -normal2;
	for (int c=0; c<3; c++){
	  if (normal2[c] == 0.0)
	    normal2[c] = 0.0;
	  if (antinormal[c] == 0.0)
	    antinormal[c] = 0.0;
	}
	Ray r(cvert+normal2*0.002, antinormal, ctx, 1);
// 	if (ctx->mydrand48()<0.01)
// 	  r.show(2.0);

	unsigned long long seed = ctx->mydrand48.seed;	
	SbColor col = getRadiance1(r, flags, NULL);

	if (STOP_RENDERING){
	  if (gv->numThreads > 1)
	    pthread_exit(0);
	  else
	    return;
	}

	/*
 	if (col[0]==0.0 && col[1]==0.0 && col[2]==0.0){
	  while (1){
	    cerr << "BLACK\n";
	    ctx->mydrand48.seed = seed;

	    showPhoton(vert[i], SbColor(0.5,0,0));
	    showPhoton(vert[i2], SbColor(0,0.5,0));
	    showPhoton(vert[i3], SbColor(0,0,0.5));
	    
	    DEBUG_STATE=1;
	    col = getRadiance1(r, flags, NULL);

	    STOP_RENDERING=1;
	    if (gv->numThreads > 1)
	      pthread_exit(0);
	    else
	      return;
	    
	  }
	}
	*/
	if (r.t>0.003){
	  cerr << "  MISSED: " << r.t << "\n";
	  continue;
// 	  showPoint(r.pos, 0.5, SbColor(1,0,0));
// 	  Ray(r.x,normal,r.ctx).show(1.0, SbColor(0.0, 1.0, 0.0));     // green
// 	  Ray(r.x,normal2,r.ctx).show(1.0, SbColor(0.0, 1.0, 1.0));    // cyan
	}

	mat->diffuseColor.set1Value(j, col);
      }
      
    }
    else
	triCount = 0;
  } // end coords/triangle loop
  
  
  // completion meter
  completion[ctx->id] = 100.0;
  print("\r");
  for (int i=0; i<gv->numThreads; i++)
    print("%6.2f%%%c ", completion[i], completion_stage[i]);

  // time statistics
  struct tms renderTimes;
  times(&renderTimes);
  ctx->ticks += renderTimes.tms_utime + renderTimes.tms_cutime;

  if (gv->numThreads > 1)
    pthread_exit(0);
  else
    return;
}




//''''''''''''''''''''''''''''''WW'''''''WW'''W'''''''''''''''''''''''''WWW''
//                              WW       WWW  W                          WW  
// WWW  WW   WWWW W  WWW  W WW WWWW  WW  W WW W  WW   W WW WWWW W   WWW  WW  
//WW W WW W  WW WW W WW W W WW  WW  WWWW W  WWW WW W  WW   WW WW W W WW  WW  
//WW   WW W  WW WW W WW W W WW  WW  WW   W   WW WW W  WW   WW WW W W WW  WW  
// WWW  WW   WW WW W WWW   WWW  WW   WWW W    W  WW   WW   WW WW W  WWW WWWW 
//                   WW                                                      
//'''''''''''''''''''WW''''''''''''''''''''''''''''''''''''''''''''''''''''''
 
void Pane::computeNormals(){

  int totalTris = 0;
  int debugLimit = 1000;
  
  // search for shapehints
  SoCallbackAction callback;
  callback.addPreCallback(SoShapeHints::getClassTypeId(), shapeHintsCBHandler,
			  this);

  SoSearchAction sa;
  sa.setType(SoIndexedShape::getClassTypeId(), TRUE);
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.setFind(SoSearchAction::TYPE);
  sa.setInterest(SoSearchAction::ALL);
  sa.apply(gv->scene);
  SoPathList plist = sa.getPaths();
  int len = plist.getLength();
  cerr << len << " shape paths found\n";

  //WW''''''''''''''''''''''''''''''''''''''''''W''''''''''''''''''''
  // W                                          W                    
  // W   W   W  WW      W  W W  WW   WW     WWW WW   WW WW   WW   WWW
  // W  W W W W W W    W W W W Wmmm W       WW  W W W W W W Wmmm  WW 
  //WWW  W   W  WW      W   W   WW  W      WWW  W W  WW WW   WW  WWW 
  //''''''''''''W'''''''''''''''''''''''''''''''''''''''W''''''''''''
  for (int i=0; i<len; i++){

    // insert lightmodel & material node
    // render one pixel for triangle
    // add color to vertex colors (3),
    //   add 1 to each vertex surface counter
    SoPath *path = (SoPath*) plist[i];
    SoIndexedShape *shape = (SoIndexedShape *)path->getTail();

    cerr << "indexed shape ";
    
    SoVertexProperty *vprop;
    vprop = (SoVertexProperty *)shape->vertexProperty.getValue();
    if (!vprop){
      cerr << "Vprop not found\n";
      continue;
    }
    vprop->normalBinding.setValue(SoVertexProperty::PER_VERTEX_INDEXED);
    shape->normalIndex.setValue(-1);

    // get shapehints
    currShapeHints = NULL;
    callback.apply(path);
    cerr << "\n";

    // get shape hints
    int reverseOrdering = 0;
    if (currShapeHints){
      if (currShapeHints->vertexOrdering.getValue() == SoShapeHints::CLOCKWISE)
	reverseOrdering = 1;
    }
    
    // get transformation
//     SoGetMatrixAction ma = SoGetMatrixAction(SbViewportRegion());
//     ma.apply(path);
//     SbMatrix matrix = ma.getMatrix();
//     SbMatrix matInvTran = matrix.inverse().transpose();

    // get number of vertices
    int numVerts = vprop->vertex.getNum();
    int numCoordIdxs = shape->coordIndex.getNum();
    
    // keep record of colors stored per vertex
//     float *okayNormals = new float[numVerts];
//     for (int j=0; j<numVerts; j++)
//       okayNormals[j] = -1.0;

    //W'''''''''''''''''''''''''''''''''''''''''''''''m'''''''''''''WW'''''''''''
    //W                                        W                     W           
    //W     W   W  WW      W  W W  WW   WW    WWW  WW W  WW WW   WW  W   WW   WWW
    //W    W W W W W W    W W W W Wmmm W       W  W   W W W W W W W  W  Wmmm  WW 
    //WWWW  W   W  WW      W   W   WW  W       W  W   W  WW W W  WW WWW  WW  WWW 
    //             W                                              W              
    //'''''''''''''W''''''''''''''''''''''''''''''''''''''''''''WW'''''''''''''''

    // data structures
//    SoMFVec3f *normals = new SoMFVec3f(); 
//    SoMFInt32 *vert_trilist[numVerts];     //the list of tris that use the vertex
    float *normalWeights = new float[numVerts];
    for(int j=0; j<numVerts; j++)
//      vert_trilist[j] = new SoMFInt32();
      normalWeights[j] = 0.0;
    
    int v[3];
    v[0] = v[1] = v[2] = -1;
    int triCount = 0;
    int numTris = 0;
    int smallAreas = 0;
    int colinearVerts = 0;
    for (int j=0; j<numCoordIdxs; j++){
      v[0] = v[1];
      v[1] = v[2];
      v[2] = shape->coordIndex[j];
      if (v[0] >= 0 && v[1] >= 0 && v[2] >= 0){

	// triangle found
	numTris++;
	triCount++;
	if (numTris%1000==0)
	  print("\r%6.2f%%    ", j/(float)(numCoordIdxs)*100.0);

	// detect same vertices
	if (v[0] == v[1] || v[0] == v[2] || v[1] == v[2]){
	  if (colinearVerts++ == 0)
	    cerr << "Warning! Colinear vertices in triangle!\n";
	  continue;
	}
	
	// get center
	SbVec3f vert[3];
	vert[0] = vprop->vertex[v[0]];
	vert[1] = vprop->vertex[v[1]];
	vert[2] = vprop->vertex[v[2]];
	
	// get normal
	SbVec3f normal;
	SbVec3f edge1 = vert[1] - vert[0];
	SbVec3f edge2 = vert[2] - vert[0];
	
	if (reverseOrdering == 0)
	  normal = edge1.cross(edge2);
	else
	  normal = edge2.cross(edge1);
	
	// flip odd triangles if indexedtrianglestripset
	if (shape->isOfType(SoIndexedTriangleStripSet::getClassTypeId()) &&
	    triCount%2==0)
	  normal = -normal;

	float area2 = normal.normalize();  

	// store normal for each tri
	//normals->set1Value(numTris-1, normal);

	if (area2<0.00001){
	  cerr << smallAreas++ << " areas too small, skipping.";
	  print("\r%6.2f%%    ", j/(float)(numCoordIdxs)*100.0);
	  continue;
	}

	float area = area2/2.0;
	
	// add tri to vert's tri lists
	for (int k=0; k<3; k++){
//	  vert_trilist[v[k]]->set1Value(vert_trilist[v[k]]->getNum(), numTris-1);

	  // initialize
	  if (normalWeights[v[k]] <= 0.0)
	    vprop->normal.set1Value(v[k], SbVec3f(0.0, 0.0, 0.0));

	  vprop->normal.set1Value(v[k], vprop->normal[v[k]] + normal*area);
	  normalWeights[v[k]] += area;
	}
	
      }
      else
	triCount = 0;
    } // end coords/triangle loop
    print("\r%6.2f%%    \n", 100.0);
    totalTris += numTris;

    //''''''''''''''''''''''''''''''''''''''''''''''''''''''''WW''''''
    //                       W                                 W      
    // WW  W  WW W  WW  W W WWW  WW     WW   W   WW WW W   WW  W   WWW
    //W   W W W W W W W W W  W  Wmmm    W W W W W   W W W W W  W   WW 
    // WW  W  W W W WW   WW  W   WW     W W  W  W   W W W  WW WWW WWW 
    //              W                                                 
    //              W                                                 
    //''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  
    //cerr<<" poly normal finished"<<endl;
    int computedNormals = 0;

    // foreach vert
    int badNormals = 0;
    for(int j=0; j<numVerts; j++){
//       SbVec3f normal = SbVec3f(0,0,0);

//       // average asscociated triangle normals
//       for(int k=0; k<vert_trilist[j]->getNum(); k++){
// 	int triNum = (*vert_trilist[j])[k];
// 	normal += (*normals)[triNum];
//       }
//       if (vert_trilist[j]->getNum()>0){
// 	normal /= vert_trilist[j]->getNum();
// 	normal.normalize();
//  	vprop->normal.set1Value(j, normal);
// 	computedNormals++;
//       }
      if (normalWeights[j]>0.0){
	SbVec3f normal = vprop->normal[j]/normalWeights[j];
	normal.normalize();
	vprop->normal.set1Value(j, normal);
 	computedNormals++;
      }
      else
	badNormals++;
    }
    cerr << computedNormals << " vertex normals computed ("
	 << badNormals << " bad)" << endl;
//     for(int j=0; j<numVerts; j++)
//       delete vert_trilist[j];
//     delete normals;
    delete [] normalWeights;
  }
}


//'''''''''''''WW''WW'''''''WWWW'''''''WW'WW'
//                 WW        WW       WW   WW
//WW    W W WW WW WWWW  WW   WW  WW W WW   WW
//WW WW W WW   WW  WW  WWWW  WW  WW W WW   WW
//WW WW W WW   WW  WW  WW    WW  WW W WW   WW
// WW WW  WW   WW  WW   WWW WWWW  WW   WW WW 
//                                           
//'''''''''''''''''''''''''''''''''''''''''''
//
// Writes Inventor scene to disk, inserting
// a comment.
//
void Pane::writeIv(){

  // back up old file if it exists
  checkForOverwrite(gv->ivOutFile);
  
  SoOutput *out = new SoOutput();
  out->setBinary(TRUE);
  if (!out->openFile(gv->ivOutFile)){
    cerr << "Error: Could not open " << gv->ivOutFile << " for writing!\n";
    exit(1);
  }

  cerr << "Writing " << gv->ivOutFile << " ..." << flush;

  SoWriteAction *wa = new SoWriteAction(out);

  // prepend command line to scene
  static int addedCmdLine = 0;
  if (addedCmdLine == 0){
    addedCmdLine = 1;

    string out = createComment();
    
    SoInfo *info = new SoInfo();
    info->string.setValue(SbString(out.c_str()));
    gv->scene->insertChild(info, 0);
  }

  if (gv->figureMode)
    wa->apply(scnRoot);        // arrows, photons
  else
    wa->apply(gv->scene);      // just scene
  
  out->closeFile();

  cerr << "done." << endl;
}

// glare filter that accounts for pupil diffraction
// uses two techniques:
//
// Physically-based glare effects for digital images
// by Spencer, Shirley, Zimmerman, Greenberg
// and
// A lighting model aiming at drive simulators
// by Nakamae, Kaneda, Okamoto, Nishita
//
// Input: pix[] array filled with Yxy luminance values,
// where pix[Y] is luminance and pix[x] pix[z] are x and y
// Yi is the adaptation luminance, and
// Yw is the maximum luminance of the display.
// toneMap() calls with Yw=1.0 because since it scales the monitor max luminance to 1
void Pane::glareFilter(float Yi, float Yw){

  cerr << "Performing simple glare filter...\n";
  
  // create glare buffer
  SbColor *bPix = new SbColor[gv->imgDim[X] * gv->imgDim[Y]];
  for (int i=0; i<gv->imgDim[X]*gv->imgDim[Y]; i++)
    bPix[i] = SbColor(0.0, 0.0, 0.0);

  // shirley
  int fdim[2];
  if (gv->useGlare == 1){

    int result = 0;
    char *type;
    
    if (log10(Yi) < -2.0)
      type = "scotopic";
    else if (log10(Yi) < 0.6)
      type = "mesopic";
    else
      type = "photopic";

    char fn[100];
    sprintf(fn, "%s.hdr", type);

    float degrees = 45.0;
    if (gv->cam->isOfType(SoPerspectiveCamera::getClassTypeId())){
      degrees = ((SoPerspectiveCamera *)gv->cam)->heightAngle.getValue();
      degrees *= 180.0/M_PI;
    }
    
    cerr << "Glare should have been created using:\n"
	 << "makeFilter  -d " << degrees << " -i " << gv->imgDim[Y]
	 << " " << type << "\n";
    
    result = readGlareFilter(fn, fdim);
    if (result != 1){

      cerr << "Error: Could not open filter image: " << fn << "\n"
	   << "Not applying any glare!\n"
	   << "Use above command to generate appropriate glare filter.\n";

      return;
     }
  }

  // loop over high intensity pixels in framebuffer
  for (int y=0; y<gv->imgDim[Y]; y++){

    // status meter
    fprintf(stderr, "\r%6.2f%%  ", y/(float)(gv->imgDim[Y]-1)*100.0);
    
    for (int x=0; x<gv->imgDim[X]; x++){

      int m = gv->imgDim[X]*y+x;
      SbColor L = pix[m];


      // transform Yxy -> XYZ
      {
      float x=L[X];
      float y=L[Z];
      L[X] = x*L[Y]/y;
      L[Z] = (1.0-x-y)*L[Y]/y;
      }

      SbColor c;

      // sRGB standard, www.srgb.com
      // physical sRGB standard, based on Yw=110.5
      c[R] =  3.2406*L[X] +  -1.5372*L[Y] +  -0.4986*L[Z];
      c[G] = -0.9689*L[X] +   1.8758*L[Y] +   0.0415*L[Z];
      c[B] =  0.0557*L[X] +  -0.2040*L[Y] +   1.0570*L[Z];

      // store tranformed color
      pix[m] = c;

      // only process glare for bright pixels
      if (L[Y]<=Yw)
	continue;

      //cerr << "Convolving filter with remaining pixel value: " << pixm << "\n";

      // mark high pixels
      if (gv->tempInt == 33){
	pix[m] = SbColor(1,0,0);
	continue;
      }

      // shirley
      if (gv->useGlare == 1){

	// convolve
	const int di = fdim[0]/2;
	for (int j=MAX2(0, y-di); j<MIN2(gv->imgDim[Y], y+di+1); j++){
	  for (int i=MAX2(0, x-di); i<MIN2(gv->imgDim[X], x+di+1); i++){
	    int l = i-x+fdim[0]/2;
	    int m = j-y+fdim[1]/2;
	    int f = fdim[0]*m+l;
	    int n = gv->imgDim[X]*j+i;

	    bPix[n][R] += filtpix[f][R]*c[R];
	    bPix[n][G] += filtpix[f][G]*c[G];
	    bPix[n][B] += filtpix[f][B]*c[B];
	  }
	}
      }

      // nakamae
      else {

	const float a = 2.5  * 1.0/1000.0; // slit diameter (pupil, meters)
	const float d = 1.5  * 1.0/100.0; // focusing distance (of eye, meters)
	const float h = 1.0  * 1.0/100.0; // image plane size (height, meters)
	const float w = h * gv->imgDim[X]/(float)gv->imgDim[Y]; // (width, meters)

	SbColor k = SbColor(
	  2.0*M_PI/(611e-9), // red
	  2.0*M_PI/(549e-9), // green
	  2.0*M_PI/(467e-9)  // blue
	  );
	
	const int de = (int)(.3*gv->imgDim[X]);
	for (int j=MAX2(0, y-de); j<MIN2(gv->imgDim[Y], y+de); j++){

	  int dyi = abs(j-y);
	  float db = 0.0;
	  if (dyi > 0){
	    float dy = dyi/(float)gv->imgDim[Y] * h;
	    db = atan2(dy, d);
	  }
	
	  for (int i=MAX2(0, x-de); i<MIN2(gv->imgDim[X], x+de); i++){
	    int n = gv->imgDim[X]*j+i;
	
	    int dxi = abs(i-x);
	    float da = 0.0;
	    if (dxi > 0){
	      float dx = dxi/(float)gv->imgDim[X] * w;
	      da = atan2(dx, d);
	    }

	    float theta = sqrt(da*da+db*db);
	    if (theta <= 0.0)
	      continue;

	    SbColor alpha = 0.5*k*a*sin(theta);

	    SbColor f0 = SbColor(
	      sin(alpha[R])/alpha[R],
	      sin(alpha[G])/alpha[G],
	      sin(alpha[B])/alpha[B]
	      );
	      
	    f0[R] *= f0[R];
	    f0[G] *= f0[G];
	    f0[B] *= f0[B];

	    bPix[n][R] += c[R]*f0[R];
	    bPix[n][G] += c[G]*f0[G];
	    bPix[n][B] += c[B]*f0[B];
	
	  }
	}
      }
    }
  }

  // combine buffers and free memory
  for (int i=0; i<gv->imgDim[X]*gv->imgDim[Y]; i++)
    pix[i] += bPix[i];
  delete [] bPix;
}
